@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -219,7 +215,7 @@ static int readBitstreamFromFile( std::ifstream *f, vvdecAccessUnit* pcAccessUni
   unsigned char* pBuf = pcAccessUnit->payload;
   pcAccessUnit->payloadUsedSize = 0;
 
-  int curfilpos = f->tellg();
+  auto curfilpos = f->tellg();
   if( curfilpos < 0 )
   {
     if( bLoop )
@@ -345,6 +341,55 @@ static int readBitstreamFromFile( std::ifstream *f, vvdecAccessUnit* pcAccessUni
   return len;
 }
 
+
+/**
+  This method writes the y4m header in front of every frame
+*/
+static int writeY4MHeader( std::ostream *f, vvdecFrame *frame )
+{
+  std::stringstream cssHeader;
+
+  if ( f == NULL || frame == NULL )
+  {
+    return -1;
+  }
+
+  if ( frame->sequenceNumber == 0 )
+  {
+    const char *cf = (frame->colorFormat == VVDEC_CF_YUV444_PLANAR) ? "444"
+                   : (frame->colorFormat == VVDEC_CF_YUV422_PLANAR) ? "422" : "420";
+
+    const char *bdepth = (frame->bitDepth == 10) ? "p10" : "";
+
+    int frameRate=50;
+    int frameScale=1;
+    if( frame->picAttributes && frame->picAttributes->hrd )
+    {
+      if( frame->picAttributes->hrd->numUnitsInTick && frame->picAttributes->hrd->timeScale )
+      {
+        frameRate  = frame->picAttributes->hrd->timeScale;
+        frameScale = frame->picAttributes->hrd->numUnitsInTick;
+
+        if( frame->picAttributes->olsHrd )
+        {
+          // if the OLS HRD is present and indicates a non-constant frame-rate, producing a Y4M is no possible
+          //if( !frame->picAttributes->olsHrd->fixedPicRateGeneralFlag && !frame->picAttributes->olsHrd->fixedPicRateWithinCvsFlag ) return -1;
+          frameScale *= frame->picAttributes->olsHrd->elementDurationInTc;
+        }
+      }
+    }
+
+    const char *interlacedMode = (frame->frameFormat == VVDEC_FF_TOP_FIELD || frame->frameFormat == VVDEC_FF_BOT_FIELD) ? " Im" : " Ip";
+    cssHeader << "YUV4MPEG2 W" << frame->width << " H" << frame->height << " F" << frameRate << ":" << frameScale << interlacedMode << " C" << cf << bdepth << "\n";
+  }
+
+  cssHeader << "FRAME\n";
+
+  f->write( cssHeader.str().c_str(),cssHeader.str().length() );
+
+  return 0;
+}
+
 /**
   This method writes and yuv planar picture into a file sink.
   \param[in]  *f file sink pointer
@@ -352,7 +397,7 @@ static int readBitstreamFromFile( std::ifstream *f, vvdecAccessUnit* pcAccessUni
   \retval     int  if non-zero an error occurred (see ErrorCodes), otherwise the return value indicates success VVC_DEC_OK
   \pre        The decoder must not be initialized.
 */
-static int writeYUVToFile( std::ostream *f, vvdecFrame *frame  )
+static int writeYUVToFile( std::ostream *f, vvdecFrame *frame, bool y4mFormat )
 {
  int ret;
  uint32_t c = 0;
@@ -361,6 +406,11 @@ static int writeYUVToFile( std::ostream *f, vvdecFrame *frame  )
 
  assert( f != NULL );
 
+ if ( y4mFormat )
+ {
+   writeY4MHeader( f, frame );
+ }
+
  for( c = 0; c < frame->numPlanes; c++ )
  {
    uiBytesPerSample = std::max( (uint32_t)frame->planes[c].bytesPerSample, uiBytesPerSample );
@@ -385,7 +435,7 @@ static int writeYUVToFile( std::ostream *f, vvdecFrame *frame  )
    \retval     int  if non-zero an error occurred (see ErrorCodes), otherwise the return value indicates success VVC_DEC_OK
    \pre        The decoder must not be initialized.
  */
-static int writeYUVToFileInterlaced( std::ostream *f, vvdecFrame *topField, vvdecFrame *botField )
+static int writeYUVToFileInterlaced( std::ostream *f, vvdecFrame *topField, vvdecFrame *botField, bool y4mFormat )
 {
   int ret;
   uint32_t c = 0;
@@ -397,6 +447,11 @@ static int writeYUVToFileInterlaced( std::ostream *f, vvdecFrame *topField, vvde
   assert( botField != NULL );
   assert( topField->numPlanes == botField->numPlanes );
 
+  if ( y4mFormat )
+  {
+    writeY4MHeader( f, topField );
+  }
+
   for( c = 0; c < topField->numPlanes; c++ )
   {
     uiBytesPerSample = std::max( (uint32_t)topField->planes[c].bytesPerSample, uiBytesPerSample );
@@ -412,7 +467,6 @@ static int writeYUVToFileInterlaced( std::ostream *f, vvdecFrame *topField, vvde
   return 0;
 }
 
-
 static inline std::string getTimePointAsString( )
 {
   std::chrono::system_clock::time_point _time_now = std::chrono::system_clock::now();
