@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -65,8 +61,8 @@ THE POSSIBILITY OF SUCH DAMAGE.
 #endif
 
 /*! Prototypes */
-int writeYUVToFile( std::ostream *f, vvdecFrame *frame );
-int writeYUVToFileInterlaced( std::ostream *f, vvdecFrame *topField, vvdecFrame *botField = nullptr );
+int writeYUVToFile( std::ostream *f, vvdecFrame *frame, bool y4mFormat = false );
+int writeYUVToFileInterlaced( std::ostream *f, vvdecFrame *topField, vvdecFrame *botField = nullptr, bool y4mFormat = false );
 
 void msgFnc( void *, int level, const char* fmt, va_list args )
 {
@@ -78,6 +74,327 @@ void msgFncErr( void *, int, const char* fmt, va_list args )
   vfprintf( stderr, fmt, args );
 }
 
+// this static variables are only needed for testing purposes, to check if external allocator is working properly
+static std::vector<void*> extAllocVec;
+static std::vector<void*> extFreeVec;
+static unsigned extRefCount;
+static unsigned extUnrefCount;
+
+class BufferStorage
+{
+public:
+BufferStorage()  = default;
+~BufferStorage()
+{
+  if ( _isAllocated ) destroy();
+}
+
+int create( size_t size, size_t alignment )
+{
+  if( size == 0 ){ return VVDEC_ERR_ALLOCATE; }
+
+#if  ( _WIN32 && ( _MSC_VER > 1300 ) ) || defined (__MINGW64_VERSION_MAJOR)
+  _ptr = (unsigned char *)_aligned_malloc( sizeof(unsigned char)*(size), alignment );
+#elif defined (__MINGW32__)
+  _ptr = (unsigned char *)__mingw_aligned_malloc( sizeof(unsigned char)*(size), alignment )
+#else
+  if( posix_memalign( (void**)&_ptr, alignment, sizeof(unsigned char)*(size) ) )
+  {
+    return -1;
+  }
+#endif
+
+  _size        = size;
+  _isAllocated = true;
+  return 0;
+}
+
+int buffer_ref()
+{
+  if( !_isAllocated) { return VVDEC_ERR_ALLOCATE; }
+  _usecnt++;
+  extRefCount++;
+  return 0;
+}
+
+int buffer_unref()
+{
+  if( !_isAllocated) { return VVDEC_ERR_ALLOCATE; }
+  if( 0 == _usecnt ) { return VVDEC_ERR_ALLOCATE; }
+  _usecnt--;
+  extUnrefCount++;
+
+  if ( _usecnt == 0 )
+  {
+    destroy();
+  }
+
+  return 0;
+}
+
+unsigned char * getStorage() { return _ptr; }
+size_t          size()       { return _size; }
+bool            isUsed()     { return _isAllocated; }
+
+private:
+int destroy()
+{
+  if( !_isAllocated) { return VVDEC_ERR_ALLOCATE; }
+
+#if     ( _WIN32 && ( _MSC_VER > 1300 ) ) || defined (__MINGW64_VERSION_MAJOR)
+_aligned_free  ( _ptr );
+#elif defined (__MINGW32__)
+__mingw_aligned_free( _ptr );
+#else
+  free (_ptr);
+#endif
+
+  _isAllocated = false;
+  _size        = 0;
+  return 0;
+}
+
+private:
+bool           _isAllocated = false;
+unsigned char *_ptr         = nullptr;     // pointer to plane buffer
+size_t         _size        = 0;
+uint32_t       _usecnt      = 0;
+
+};
+
+
+void* picBufferCreateFnc( void *, vvdecComponentType, uint32_t size, uint32_t alignment, void **application_data )
+{
+  BufferStorage* s = new BufferStorage();
+  //std::cout << "create(" << plane << ") " << size << " ptr " << (void*)s << std::endl;
+  extAllocVec.push_back( (void*)s  );
+
+  if( 0 != s->create(size, alignment) )
+  {
+    delete s;
+    return nullptr;
+  }
+  // add a buffer reference to the created buffer
+  s->buffer_ref();
+  *application_data = (void*)s;
+  return (void*)s->getStorage();
+}
+
+void picBufferUnrefFnc( void *, void *application_data )
+{
+  BufferStorage* s = (BufferStorage*)application_data;
+  //std::cout << "destroy ptr  " << (void*)s << std::endl;
+  extFreeVec.push_back( (void*)s  );
+
+  // remove buffer reference and free memory when all references are released
+  s->buffer_unref();
+  if( !s->isUsed() )
+    delete s;
+}
+
+
+int unrefBuf( void*& allocator )
+{
+  if( allocator )
+  {
+    BufferStorage* s = (BufferStorage*)allocator;
+    s->buffer_unref();
+    if( !s->isUsed() )
+      delete s;
+  }
+
+  return 0;
+}
+
+/* refFrame
+  create a new reference to all plane buffers
+*/
+void refFrame( vvdecFrame *frame )
+{
+  for( auto p : frame->planes )
+  {
+    if( p.allocator )
+    {
+      BufferStorage* s = (BufferStorage*)p.allocator ;
+      s->buffer_ref();
+    }
+  }
+}
+
+/* unrefFrame
+   unreference frame after it is rendered 
+   when all buffer references are release the buffer must be freed
+*/
+void unrefFrame( vvdecFrame *frame )
+{
+  for( auto p : frame->planes )
+  {
+    if( p.allocator )
+    {
+      BufferStorage* s = (BufferStorage*)p.allocator;
+      s->buffer_unref();
+      if( !s->isUsed() )
+        delete s;
+    }
+  }
+}
+
+void printSEI( vvdecDecoder *dec, vvdecFrame *frame, std::ostream * logStream )
+{
+  vvdecSEI *sei = vvdec_find_frame_sei( dec, VVDEC_BUFFERING_PERIOD, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI BUFFERING_PERIOD: " << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_PICTURE_TIMING, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI PICTURE_TIMING" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_FILLER_PAYLOAD, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI FILLER_PAYLOAD" << std::endl;
+  }
+  
+  sei = vvdec_find_frame_sei( dec, VVDEC_USER_DATA_REGISTERED_ITU_T_T35, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI USER_DATA_REGISTERED_ITU_T_T35" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_USER_DATA_UNREGISTERED, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI USER_DATA_UNREGISTERED" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_FILM_GRAIN_CHARACTERISTICS, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI FILM_GRAIN_CHARACTERISTICS" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_FRAME_PACKING, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI FRAME_PACKING" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_PARAMETER_SETS_INCLUSION_INDICATION, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI PARAMETER_SETS_INCLUSION_INDICATION" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_DECODING_UNIT_INFO, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI DECODING_UNIT_INFO" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_DECODED_PICTURE_HASH, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI DECODED_PICTURE_HASH " << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_SCALABLE_NESTING, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI SCALABLE_NESTING" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_MASTERING_DISPLAY_COLOUR_VOLUME, frame );
+  if( sei )
+  {
+    vvdecSEIMasteringDisplayColourVolume* p = reinterpret_cast<vvdecSEIMasteringDisplayColourVolume *>(sei->payload);
+    *logStream << "vvdecapp [detail]: SEI MASTERING_DISPLAY_COLOUR_VOLUME: lumaMin,Max " << p->minLuminance << "," << p->maxLuminance << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_DEPENDENT_RAP_INDICATION, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI DEPENDENT_RAP_INDICATION" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_EQUIRECTANGULAR_PROJECTION, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI EQUIRECTANGULAR_PROJECTION" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_SPHERE_ROTATION, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI SPHERE_ROTATION" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_REGION_WISE_PACKING, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI REGION_WISE_PACKING" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_OMNI_VIEWPORT, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI OMNI_VIEWPORT" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_GENERALIZED_CUBEMAP_PROJECTION, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI GENERALIZED_CUBEMAP_PROJECTION" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_FRAME_FIELD_INFO, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI FRAME_FIELD_INFO" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_SUBPICTURE_LEVEL_INFO, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI SUBPICTURE_LEVEL_INFO" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_SAMPLE_ASPECT_RATIO_INFO, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI SAMPLE_ASPECT_RATIO_INFO" << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_CONTENT_LIGHT_LEVEL_INFO, frame );
+  if( sei )
+  {
+    vvdecSEIContentLightLevelInfo* p = reinterpret_cast<vvdecSEIContentLightLevelInfo *>(sei->payload);
+    *logStream << "vvdecapp [detail]: SEI CONTENT_LIGHT_LEVEL_INFO: " << p->maxContentLightLevel << "," << p->maxPicAverageLightLevel << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_ALTERNATIVE_TRANSFER_CHARACTERISTICS, frame );
+  if( sei )
+  {
+    vvdecSEIAlternativeTransferCharacteristics* p = reinterpret_cast<vvdecSEIAlternativeTransferCharacteristics *>(sei->payload);
+    *logStream << "vvdecapp [detail]: SEI ALTERNATIVE_TRANSFER_CHARACTERISTICS: preferred_transfer_characteristics " << (int)p->preferred_transfer_characteristics << std::endl;
+  }
+
+  sei = vvdec_find_frame_sei( dec, VVDEC_AMBIENT_VIEWING_ENVIRONMENT, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI AMBIENT_VIEWING_ENVIRONMENT" << std::endl;
+  }
+  
+  sei = vvdec_find_frame_sei( dec, VVDEC_CONTENT_COLOUR_VOLUME, frame );
+  if( sei )
+  {
+    *logStream << "vvdecapp [detail]: SEI CONTENT_COLOUR_VOLUME" << std::endl;   
+  }
+}
+
 int main( int argc, char* argv[] )
 {
   std::string cAppname = argv[0];
@@ -99,7 +416,11 @@ int main( int argc, char* argv[] )
   std::string cOutputFile    = "";
   int         iMaxFrames     = -1;
   int         iLoopCount     = 1;
+  bool        y4mOutput      = false;
+  bool        externAllocator = false;
   std::string cExpectedYuvMD5;
+  std::string sTracingRule;
+  std::string sTracingFile;
   vvdecParams params;
   vvdec_params_default(&params);
 
@@ -107,21 +428,27 @@ int main( int argc, char* argv[] )
 
   if(  argc > 1 && (!strcmp( (const char*) argv[1], "--help" ) || !strcmp( (const char*) argv[1], "-help" )) )
   {
-    vvdecoderapp::CmdLineParser::print_usage( cAppname, params );
+    vvdecoderapp::CmdLineParser::print_usage( cAppname, params, false );
     return 0;
   }
 
-  int iRet = vvdecoderapp::CmdLineParser::parse_command_line(  argc, argv, params, cBitstreamFile, cOutputFile, iMaxFrames, iLoopCount, cExpectedYuvMD5 );
+
+  int iRet = vvdecoderapp::CmdLineParser::parse_command_line( argc, argv, params, cBitstreamFile, cOutputFile, iMaxFrames, iLoopCount, cExpectedYuvMD5, y4mOutput, externAllocator, sTracingFile, sTracingRule );
   if( iRet != 0 )
   {
     if( iRet == 2 )
     {
-      vvdecoderapp::CmdLineParser::print_usage( cAppname, params);
+      vvdecoderapp::CmdLineParser::print_usage( cAppname, params, false );
       return 0;
     }
     else if( iRet == 3 )
     {
-      std::cout << cAppname  << " version " << vvdec_get_version()<< std::endl;
+      vvdecoderapp::CmdLineParser::print_usage( cAppname, params, true );
+      return 0;
+    }
+    else if( iRet == 4 )
+    {
+      std::cout << cAppname << " version " << vvdec_get_version() << std::endl;
       return 0;
     }
 
@@ -163,6 +490,12 @@ int main( int argc, char* argv[] )
     }
     else
     {
+      const char * s = strrchr(cOutputFile.c_str(), '.');
+      if (s && !strcmp(s, ".y4m"))
+      {
+        y4mOutput = true;
+      }
+
       cRecFile.open( cOutputFile.c_str(), std::fstream::binary | std::fstream::out );
       if( !cRecFile )
       {
@@ -192,6 +525,17 @@ int main( int argc, char* argv[] )
   vvdecAccessUnit* accessUnit = vvdec_accessUnit_alloc();
   vvdec_accessUnit_alloc_payload( accessUnit, MAX_CODED_PICTURE_SIZE );
 
+#ifdef ENABLE_TRACING
+  if( sTracingFile.size() || sTracingRule.size() )
+  {
+    if( vvdec_set_tracing( sTracingFile.c_str(), sTracingRule.c_str() ) != VVDEC_OK )
+    {
+      *logStream << "vvdecapp [error]: failed to initialize tracing to file '" << sTracingFile << "' with rule '" << sTracingRule << "'" << std::endl;
+      return -1;
+    }
+  }
+#endif   // ENABLE_TRACING
+
   bool bOutputInfoWritten = false;
 
   std::chrono::steady_clock::time_point cTPStartRun;
@@ -211,7 +555,7 @@ int main( int argc, char* argv[] )
     vvdecFrame* pcFrame     = NULL;
     vvdecFrame* pcPrevField = NULL;
 
-    if( iLoopCount > 1 )
+    if( iLoopCount > 1 && params.logLevel >= VVDEC_INFO )
     {
       *logStream << "-------------------------------------" << std::endl;
       *logStream << "begin decoder loop #" << iLoop << std::endl;
@@ -219,17 +563,25 @@ int main( int argc, char* argv[] )
     }
 
     // initialize the decoder
-    dec = vvdec_decoder_open( &params );
+    if( externAllocator )
+    {
+      dec = vvdec_decoder_open_with_allocator( &params, picBufferCreateFnc, picBufferUnrefFnc );
+    }
+    else
+    {
+      dec = vvdec_decoder_open( &params );
+    }
+
     if( nullptr == dec )
     {
-      *logStream << "cannot init decoder" << std::endl;
+      *logStream << "vvdecapp [error]: cannot init decoder" << std::endl;
       vvdec_accessUnit_free( accessUnit );
       return -1;
     }
 
     vvdec_set_logging_callback( dec, writeStdout ? msgFncErr : msgFnc );
 
-    if( iLoop == 0 )
+    if( iLoop == 0 && params.logLevel >= VVDEC_INFO )
     {
       *logStream << vvdec_get_dec_information( dec ) << std::endl;
     }
@@ -262,7 +614,7 @@ int main( int argc, char* argv[] )
         if( params.logLevel == VVDEC_DETAILS )
         {
           std::string cNal = getNalUnitTypeAsString( eNalType );
-          *logStream << "  read nal " <<  cNal << " size " << accessUnit->payloadUsedSize << std::endl;
+          *logStream << "vvdecapp [details]:  read nal " <<  cNal << " size " << accessUnit->payloadUsedSize << std::endl;
         }
 
         if( eNalType == VVC_NAL_UNIT_PH )
@@ -324,7 +676,7 @@ int main( int argc, char* argv[] )
         {
           if( !bMultipleSlices )
           {
-            if( params.logLevel >= VVDEC_VERBOSE ) *logStream << "more data needed to tune in" << std::endl;
+            if( params.logLevel >= VVDEC_VERBOSE ) *logStream << "vvdecapp [verbose]: more data needed to tune in" << std::endl;
             if( bTunedIn )
             {
               // after the first frame is returned, the decoder must always return a frame
@@ -336,30 +688,46 @@ int main( int argc, char* argv[] )
             }
           }
         }
-        else if( iRet != VVDEC_OK )
+        else if( iRet == VVDEC_ERR_DEC_INPUT )
         {
-          std::string cErr = vvdec_get_last_error(dec);
-          std::string cAdditionalErr = vvdec_get_last_additional_error(dec);
+          std::string cErr           = vvdec_get_last_error( dec );
+          std::string cAdditionalErr = vvdec_get_last_additional_error( dec );
+          *logStream << "vvdecapp [warning]: " << cErr << " (" << vvdec_get_error_msg( iRet ) << ")";
           if( !cAdditionalErr.empty() )
           {
-            *logStream << "vvdecapp [error]: decoding failed: " << cErr << " (" <<vvdec_get_error_msg( iRet ) << ")"
-                       << " detail: " << vvdec_get_last_additional_error(dec) << std::endl;
+            *logStream << " detail: " << vvdec_get_last_additional_error( dec ) << std::endl;
           }
-          else
+          *logStream << std::endl;
+        }
+        else if( iRet != VVDEC_OK )
+        {
+          std::string cErr           = vvdec_get_last_error( dec );
+          std::string cAdditionalErr = vvdec_get_last_additional_error( dec );
+          *logStream << "vvdecapp [error]: decoding failed: " << cErr << " (" <<vvdec_get_error_msg( iRet ) << ")";
+          if( !cAdditionalErr.empty() )
           {
-            *logStream << "vvdecapp [error]: decoding failed: " << cErr << " (" <<vvdec_get_error_msg( iRet ) << ")" << std::endl;
+            *logStream << " detail: " << vvdec_get_last_additional_error( dec ) << std::endl;
           }
+          *logStream << std::endl;
+
           vvdec_accessUnit_free( accessUnit );
           return iRet;
         }
 
         if( NULL != pcFrame && pcFrame->ctsValid )
         {
+          if ( externAllocator ) refFrame( pcFrame ); 
+
           if (!bOutputInfoWritten)
           {
             if( params.logLevel >= VVDEC_INFO )
             {
               *logStream << "vvdecapp [info]: SizeInfo: " << pcFrame->width << "x" << pcFrame->height << " (" << pcFrame->bitDepth << "b)" << std::endl;
+              if( pcFrame->picAttributes && pcFrame->picAttributes->vui && pcFrame->picAttributes->vui->colourDescriptionPresentFlag )
+              {
+                *logStream << "vvdecapp [info]: VUI ColourDescription: colourPrim: " << pcFrame->picAttributes->vui->colourPrimaries << " transCharacteristics: " << pcFrame->picAttributes->vui->transferCharacteristics 
+                           << " matrixCoefficients: " << pcFrame->picAttributes->vui->matrixCoefficients << std::endl;
+              }
             }
             bOutputInfoWritten = true;
           }
@@ -375,16 +743,13 @@ int main( int argc, char* argv[] )
           if( pcFrame->picAttributes )
           {
             uiBitrate += pcFrame->picAttributes->bits;
+            (void)uiBitrate;
           }
 
-#if 0 // just sample code to retrieve sei messages
-          vvdecSEI *sei = vvdec_find_frame_sei( dec, VVDEC_CONTENT_LIGHT_LEVEL_INFO, pcFrame );
-          if( sei )
+          if( params.logLevel == VVDEC_DETAILS )
           {
-            vvdecSEIContentLightLevelInfo* p = reinterpret_cast<vvdecSEIContentLightLevelInfo *>(sei->payload);
-            *logStream << "vvdecapp [info]: CONTENT_LIGHT_LEVEL_INFO: " << p->maxContentLightLevel << "," << p->maxPicAverageLightLevel << std::endl;
+            printSEI( dec, pcFrame, logStream );
           }
-#endif
 
           if( pcFrame->frameFormat == VVDEC_FF_PROGRESSIVE )
           {
@@ -394,7 +759,7 @@ int main( int argc, char* argv[] )
             }
             if( cRecFile.is_open() || writeStdout )
             {
-              if( 0 != writeYUVToFile( outStream, pcFrame ) )
+              if( 0 != writeYUVToFile( outStream, pcFrame, y4mOutput ) )
               {
                 *logStream << "vvdecapp [error]: write of rec. yuv failed for picture seq. " <<  pcFrame->sequenceNumber << std::endl;
                 vvdec_accessUnit_free( accessUnit );
@@ -417,14 +782,17 @@ int main( int argc, char* argv[] )
               }
               if( cRecFile.is_open() || writeStdout )
               {
-                if( 0 != writeYUVToFileInterlaced( outStream, pcPrevField, pcFrame ) )
+                if( 0 != writeYUVToFileInterlaced( outStream, pcPrevField, pcFrame, y4mOutput ) )
                 {
                   *logStream << "vvdecapp [error]: write of rec. yuv failed for picture seq. " <<  pcFrame->sequenceNumber << std::endl;
                   vvdec_accessUnit_free( accessUnit );
                   return iRet;
                 }
               }
+              
+              if ( externAllocator ) unrefFrame( pcPrevField );
               vvdec_frame_unref( dec, pcPrevField );
+              
               pcPrevField = nullptr;
             }
           }
@@ -438,6 +806,7 @@ int main( int argc, char* argv[] )
           // free picture memory
           if( !pcPrevField || pcPrevField != pcFrame)
           {
+            if ( externAllocator ) unrefFrame( pcFrame );             
             vvdec_frame_unref( dec, pcFrame );
           }
         }
@@ -469,11 +838,22 @@ int main( int argc, char* argv[] )
       iRet = vvdec_flush( dec, &pcFrame );
       if( iRet != VVDEC_OK && iRet != VVDEC_EOF )
       {
-        *logStream << "vvdecapp [error]: decoding failed (" << iRet << ")" << std::endl;  return iRet;
+        std::string cErr           = vvdec_get_last_error(dec);
+        std::string cAdditionalErr = vvdec_get_last_additional_error(dec);
+        *logStream << "vvdecapp [error]: decoding failed: " << cErr << " (" <<vvdec_get_error_msg( iRet ) << ")";
+        if( !cAdditionalErr.empty() )
+        {
+          *logStream << " detail: " << cAdditionalErr;
+        }
+        *logStream << std::endl;
+        vvdec_accessUnit_free( accessUnit );
+        return iRet;
       }
 
       if( NULL != pcFrame  )
       {
+        if ( externAllocator ) refFrame( pcFrame );
+
         uiFrames++;
 
         if( pcFrame->frameFormat == VVDEC_FF_PROGRESSIVE )
@@ -484,7 +864,7 @@ int main( int argc, char* argv[] )
           }
           if( cRecFile.is_open() || writeStdout )
           {
-            if( 0 != writeYUVToFile( outStream, pcFrame ) )
+            if( 0 != writeYUVToFile( outStream, pcFrame, y4mOutput ) )
             {
               *logStream << "vvdecapp [error]: write of rec. yuv failed for picture seq. " << pcFrame->sequenceNumber << std::endl;
               vvdec_accessUnit_free( accessUnit );
@@ -507,13 +887,15 @@ int main( int argc, char* argv[] )
             }
             if( cRecFile.is_open() || writeStdout )
             {
-              if( 0 != writeYUVToFileInterlaced( outStream, pcPrevField, pcFrame ) )
+              if( 0 != writeYUVToFileInterlaced( outStream, pcPrevField, pcFrame, y4mOutput ) )
               {
                 *logStream << "vvdecapp [error]: write of rec. yuv failed for picture seq. " << pcFrame->sequenceNumber << std::endl;
                 vvdec_accessUnit_free( accessUnit );
                 return iRet;
               }
             }
+
+            if ( externAllocator ) unrefFrame( pcPrevField );
             vvdec_frame_unref( dec, pcPrevField );
             pcPrevField = nullptr;
           }
@@ -528,6 +910,7 @@ int main( int argc, char* argv[] )
         // free picture memory
         if( !pcPrevField || pcPrevField != pcFrame)
         {
+          if ( externAllocator ) unrefFrame( pcFrame );
           vvdec_frame_unref( dec, pcFrame );
         }
 
@@ -637,11 +1020,21 @@ int main( int argc, char* argv[] )
     {
       dFpsSum += f;
     }
-    if( params.logLevel > VVDEC_SILENT )
+    if( params.logLevel >= VVDEC_INFO )
     {
       *logStream <<"vvdecapp [info]: avg. fps for " << dFpsPerLoopVec.size() << " loops: " << dFpsSum/dFpsPerLoopVec.size() << " Hz " << std::endl;
     }
   }
 
+  if( externAllocator )
+  {
+   if ( (extAllocVec.size() != extFreeVec.size()) || ( extRefCount != extUnrefCount ) )
+   {
+      *logStream <<"vvdecapp [error]: extern alloc/free mismatch (allocated " << extAllocVec.size() << ", free " << extFreeVec.size() << ")" 
+                 << " refcount " << extRefCount << " unrefcount " << extUnrefCount << std::endl;
+      return -1;
+   }
+  }
+
   return 0;
 }
