@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -58,6 +54,11 @@ THE POSSIBILITY OF SUCH DAMAGE.
 #include "Picture.h"
 #include "Slice.h"
 
+#if ENABLE_SIMD_OPT_BUFFER && defined( TARGET_SIMD_X86 )
+#include "CommonDefX86.h"
+#include <simde/x86/sse.h>
+#endif
+
 namespace vvdec
 {
 
@@ -177,6 +178,7 @@ void copyBufferCore( const char *src, ptrdiff_t srcStride, char *dst, ptrdiff_t
   if( width == srcStride && width == dstStride )
   {
     memcpy( dst, src, width * height );
+    return;
   }
 
   for( int i = 0; i < height; i++ )
@@ -193,33 +195,6 @@ void copyBufferCore( const char *src, ptrdiff_t srcStride, char *dst, ptrdiff_t
   }
 }
 
-template<int padSize>
-void paddingCore(Pel *ptr, ptrdiff_t stride, int width, int height)
-{
-  /*left and right padding*/
-  Pel *ptrTemp1 = ptr;
-  Pel *ptrTemp2 = ptr + (width - 1);
-  ptrdiff_t offset = 0;
-  for (int i = 0; i < height; i++)
-  {
-    offset = stride * i;
-    for (int j = 1; j <= padSize; j++)
-    {
-      *(ptrTemp1 - j + offset) = *(ptrTemp1 + offset);
-      *(ptrTemp2 + j + offset) = *(ptrTemp2 + offset);
-    }
-  }
-  /*Top and Bottom padding*/
-  int numBytes = (width + padSize + padSize) * sizeof(Pel);
-  ptrTemp1 = (ptr - padSize);
-  ptrTemp2 = (ptr + (stride * (height - 1)) - padSize);
-  for (int i = 1; i <= padSize; i++)
-  {
-    memcpy(ptrTemp1 - (i * stride), (ptrTemp1), numBytes);
-    memcpy(ptrTemp2 + (i * stride), (ptrTemp2), numBytes);
-  }
-}
-
 void applyLutCore( Pel* ptr, ptrdiff_t ptrStride, int width, int height, const Pel* lut )
 {
   //    const auto rsp_sgnl_op  = [=, &dst]( int ADDR ){ dst[ADDR] = lut[dst[ADDR]]; };
@@ -236,7 +211,6 @@ void applyLutCore( Pel* ptr, ptrdiff_t ptrStride, int width, int height, const P
 #undef RSP_SGNL_INC
 }
 
-
 void fillN_CuCore( CodingUnit** ptr, ptrdiff_t ptrStride, int width, int height, CodingUnit* cuPtr )
 {
   if( width == ptrStride )
@@ -255,9 +229,9 @@ void fillN_CuCore( CodingUnit** ptr, ptrdiff_t ptrStride, int width, int height,
 }
 
 void sampleRateConvCore( const std::pair<int, int> scalingRatio, const std::pair<int, int> compScale,
-                         const Pel* orgSrc, const int orgStride, const int orgWidth, const int orgHeight,
+                         const Pel* orgSrc, const ptrdiff_t orgStride, const int orgWidth, const int orgHeight,
                          const int beforeScaleLeftOffset, const int beforeScaleTopOffset,
-                         Pel* scaledSrc, const int scaledStride, const int scaledWidth, const int scaledHeight,
+                         Pel* scaledSrc, const ptrdiff_t scaledStride, const int scaledWidth, const int scaledHeight,
                          const int afterScaleLeftOffset, const int afterScaleTopOffset,
                          const int bitDepth, const bool useLumaFilter,
                          const bool horCollocatedPositionFlag, const bool verCollocatedPositionFlag )
@@ -285,7 +259,7 @@ void sampleRateConvCore( const std::pair<int, int> scalingRatio, const std::pair
   int* buf = new int[orgHeight * scaledWidth];
   int maxVal = (1 << bitDepth) - 1;
 
-  CHECK( bitDepth > 17, "Overflow may happen!" );
+  CHECK_RECOVERABLE( bitDepth > 17, "Overflow may happen!" );
 
   for( int i = 0; i < scaledWidth; i++ )
   {
@@ -342,6 +316,25 @@ void sampleRateConvCore( const std::pair<int, int> scalingRatio, const std::pair
   delete[] buf;
 }
 
+void rspFwdCore( Pel* ptr, ptrdiff_t ptrStride, int width, int height, const int bd, const Pel OrgCW, const Pel* LmcsPivot, const Pel* ScaleCoeff, const Pel* InputPivot )
+{
+  int idxY;
+  int shift = getLog2( OrgCW );
+
+  //    const auto rsp_sgnl_op  = [=, &dst]( int ADDR ){ idxY = ( dst[ADDR] >> shift ); dst[ADDR] = static_cast<Pel>( ClipBD<int>( LmcsPivot[idxY] + ( ( ScaleCoeff[idxY] * ( dst[ADDR] - InputPivot[idxY] ) + ( 1 << 10 ) ) >> 11 ), bd ) ); };
+  //    const auto rsp_sgnl_inc = [=, &dst]            { dst += stride; };
+
+  //    size_aware_pel_op( rsp_sgnl_op, rsp_sgnl_inc, width, height );
+
+#define RSP_FWD_OP( ADDR ) { idxY = ( ptr[ADDR] >> shift ); ptr[ADDR] = static_cast<Pel>( ClipBD<int>( LmcsPivot[idxY] + ( ( ScaleCoeff[idxY] * ( ptr[ADDR] - InputPivot[idxY] ) + ( 1 << 10 ) ) >> 11 ), bd ) ); }
+#define RSP_FWD_INC        ptr      += ptrStride;
+
+  SIZE_AWARE_PER_EL_OP( RSP_FWD_OP, RSP_FWD_INC )
+
+#undef RSP_FWD_OP
+#undef RSP_FWD_INC
+}
+
 PelBufferOps::PelBufferOps()
 {
   addAvg4  = addAvgCore<Pel>;
@@ -358,13 +351,13 @@ PelBufferOps::PelBufferOps()
   wghtAvg8 = addWeightedAvgCore<Pel>;
 
   copyBuffer = copyBufferCore;
-  padding2 = paddingCore<2>;
-  padding1 = paddingCore<1>;
 
   transpose4x4 = transpose4x4Core<Pel>;
   transpose8x8 = transpose8x8Core<Pel>;
 
   applyLut = applyLutCore;
+  rspFwd   = rspFwdCore;
+  rspBcw   = nullptr;
 
   fillN_CU = fillN_CuCore;
 
@@ -430,12 +423,6 @@ void AreaBuf<Pel>::rescaleBuf( const AreaBuf<const Pel>& beforeScaling, Componen
                              isLuma( compID ) ? 1 : horCollocatedChromaFlag, isLuma( compID ) ? 1 : verCollocatedChromaFlag );
 }
 
-template<>
-void AreaBuf<Pel>::rspSignal( const Pel* lut )
-{
-  g_pelBufOP.applyLut( buf, stride, width, height, lut );
-}
-
 template<>
 void AreaBuf<Pel>::scaleSignal(const int scale, const ClpRng& clpRng)
 {
@@ -576,25 +563,27 @@ void AreaBuf<Pel>::linearTransform( const int scale, const int shift, const int
 template<>
 void AreaBuf<Pel>::transposedFrom( const AreaBuf<const Pel> &other )
 {
-  CHECK( width != other.height || height != other.width, "Incompatible size" );
+  CHECK_RECOVERABLE( width != other.height || height != other.width, "Incompatible size" );
 
-  if( ( width & 3 ) != 0 || ( height & 3 ) != 0 )
+  if( ( ( width | height ) & 7 ) == 0 )
   {
-          Pel* dst =       buf;
     const Pel* src = other.buf;
-    width          = other.height;
-    height         = other.width;
-    stride         = stride < width ? width : stride;
 
-    for( unsigned y = 0; y < other.height; y++ )
+    for( unsigned y = 0; y < other.height; y += 8 )
     {
-      for( unsigned x = 0; x < other.width; x++ )
+      Pel* dst = buf + y;
+
+      for( unsigned x = 0; x < other.width; x += 8 )
       {
-        dst[y + x*stride] = src[x + y * other.stride];
+        g_pelBufOP.transpose8x8( &src[x], other.stride, dst, stride );
+
+        dst += 8 * stride;
       }
+
+      src += 8 * other.stride;
     }
   }
-  else if( ( width & 7 ) != 0 || ( height & 7 ) != 0 )
+  else if( ( ( width | height ) & 3 ) == 0 )
   {
     const Pel* src = other.buf;
 
@@ -614,20 +603,18 @@ void AreaBuf<Pel>::transposedFrom( const AreaBuf<const Pel> &other )
   }
   else
   {
+          Pel* dst =       buf;
     const Pel* src = other.buf;
+    width          = other.height;
+    height         = other.width;
+    stride         = stride < width ? width : stride;
 
-    for( unsigned y = 0; y < other.height; y += 8 )
+    for( unsigned y = 0; y < other.height; y++ )
     {
-      Pel* dst = buf + y;
-
-      for( unsigned x = 0; x < other.width; x += 8 )
+      for( unsigned x = 0; x < other.width; x++ )
       {
-        g_pelBufOP.transpose8x8( &src[x], other.stride, dst, stride );
-
-        dst += 8 * stride;
+        dst[y + x*stride] = src[x + y * other.stride];
       }
-
-      src += 8 * other.stride;
     }
   }
 }
@@ -655,7 +642,8 @@ PelStorage::PelStorage()
 {
   for( uint32_t i = 0; i < MAX_NUM_COMPONENT; i++ )
   {
-    m_origin[i] = nullptr;
+    m_origin[i]        = nullptr;
+    m_allocator[i]     = nullptr;
   }
 }
 
@@ -669,9 +657,9 @@ void PelStorage::create( const UnitArea &_UnitArea )
   create( _UnitArea.chromaFormat, _UnitArea.blocks[0] );
 }
 
-void PelStorage::create( const ChromaFormat _chromaFormat, const Size& _size, const unsigned _maxCUSize, const unsigned _margin, const unsigned _alignmentByte, const bool _scaleChromaMargin )
+void PelStorage::create( const ChromaFormat _chromaFormat, const Size& _size, const unsigned _maxCUSize, const unsigned _margin, const unsigned _alignmentByte, const bool _scaleChromaMargin, const UserAllocator* userAlloc )
 {
-  CHECK( !bufs.empty(), "Trying to re-create an already initialized buffer" );
+  CHECK_RECOVERABLE( !bufs.empty(), "Trying to re-create an already initialized buffer" );
 
   chromaFormat = _chromaFormat;
 
@@ -712,15 +700,29 @@ void PelStorage::create( const ChromaFormat _chromaFormat, const Size& _size, co
     if( _alignment )
     {
       // make sure buffer lines are align
-      CHECK( _alignmentByte != MEMORY_ALIGN_DEF_SIZE, "Unsupported alignment" );
+      CHECK_RECOVERABLE( _alignmentByte != MEMORY_ALIGN_DEF_SIZE, "Unsupported alignment" );
       totalWidth = ( ( totalWidth + _alignment - 1 ) / _alignment ) * _alignment;
     }
 
+#if ENABLE_SIMD_OPT_INTER
+    uint32_t area = totalWidth * totalHeight + 1; // +1 for the extra Pel overread in prefetchPad_SSE, in case reading from the very bottom right of the picture
+#else
     uint32_t area = totalWidth * totalHeight;
-    CHECK( !area, "Trying to create a buffer with zero area" );
+#endif
+    CHECK_RECOVERABLE( !area, "Trying to create a buffer with zero area" );
 
     m_origSi[i] = Size{ totalWidth, totalHeight };
-    m_origin[i] = ( Pel* ) xMalloc( Pel, area );
+    if( userAlloc && userAlloc->enabled )
+    {
+      m_origin[i] = ( Pel* ) userAlloc->create( userAlloc->opaque, (vvdecComponentType)i, sizeof(Pel)*area, MEMORY_ALIGN_DEF_SIZE, &m_allocator[i] );
+      CHECK_RECOVERABLE( m_origin[i] == nullptr, "external allocator callback failed (returned NULL)." );
+      m_externAllocator = true;
+      m_userAlloc       = userAlloc;
+    }
+    else
+    {
+      m_origin[i] = ( Pel* ) xMalloc( Pel, area );
+    }
     Pel* topLeft = m_origin[i] + totalWidth * ymargin + xmargin;
     bufs.push_back( PelBuf( topLeft, totalWidth, _size.width >> scaleX, _size.height >> scaleY ) );
   }
@@ -748,13 +750,16 @@ void PelStorage::swap( PelStorage& other )
   for( uint32_t i = 0; i < numCh; i++ )
   {
     // check this otherwise it would turn out to get very weird
-    CHECK( chromaFormat                   != other.chromaFormat                  , "Incompatible formats" );
-    CHECK( get( ComponentID( i ) )        != other.get( ComponentID( i ) )       , "Incompatible formats" );
-    CHECK( get( ComponentID( i ) ).stride != other.get( ComponentID( i ) ).stride, "Incompatible formats" );
+    CHECK_RECOVERABLE( chromaFormat                   != other.chromaFormat                  , "Incompatible formats" );
+    CHECK_RECOVERABLE( get( ComponentID( i ) )        != other.get( ComponentID( i ) )       , "Incompatible formats" );
+    CHECK_RECOVERABLE( get( ComponentID( i ) ).stride != other.get( ComponentID( i ) ).stride, "Incompatible formats" );
 
     std::swap( bufs[i].buf,    other.bufs[i].buf );
     std::swap( bufs[i].stride, other.bufs[i].stride );
     std::swap( m_origin[i],    other.m_origin[i] );
+    std::swap( m_allocator[i], other.m_allocator[i] );
+    std::swap( m_externAllocator, other.m_externAllocator );
+    std::swap( m_userAlloc, other.m_userAlloc );
   }
 }
 
@@ -765,7 +770,15 @@ void PelStorage::destroy()
   {
     if( m_origin[i] )
     {
-      xFree( m_origin[i] );
+      if ( !m_externAllocator )
+      {
+        xFree( m_origin[i] );
+      }
+      else if( m_allocator[i])
+      {
+        CHECK_RECOVERABLE( m_userAlloc->unref == nullptr, "vvdecUnrefBufferCallback not valid, cannot unref picture buffer" )
+        m_userAlloc->unref( m_userAlloc->opaque, m_allocator[i] );
+      }
       m_origin[i] = nullptr;
     }
   }
@@ -862,6 +875,7 @@ template<typename T>
 void UnitBuf<T>::writeToFile( std::string filename ) const
 {
   FILE* f = fopen( filename.c_str(), "w" );
+  CHECK( f == nullptr, "writeToFile() cannot open file for writing" )
 
   for( auto& b: bufs )
   {
@@ -874,4 +888,4 @@ void UnitBuf<T>::writeToFile( std::string filename ) const
   fclose( f );
 }
 
-}
\ No newline at end of file
+}
