@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -64,12 +60,11 @@ THE POSSIBILITY OF SUCH DAMAGE.
 namespace vvdec
 {
 
-#if ENABLE_SIMD_OPT_BUFFER
 struct PelBufferOps
 {
   PelBufferOps();
 
-#ifdef TARGET_SIMD_X86
+#if defined( TARGET_SIMD_X86 ) && ENABLE_SIMD_OPT_BUFFER
   void initPelBufOpsX86();
   template<X86_VEXT vext>
   void _initPelBufOpsX86();
@@ -85,21 +80,20 @@ struct PelBufferOps
   void ( *wghtAvg4 )      ( const Pel* src0, ptrdiff_t src0Stride, const Pel* src1, ptrdiff_t src1Stride, Pel *dst, ptrdiff_t dstStride, int width, int height, int shift, int offset, int w0, int w1, const ClpRng& clpRng );
   void ( *wghtAvg8 )      ( const Pel* src0, ptrdiff_t src0Stride, const Pel* src1, ptrdiff_t src1Stride, Pel *dst, ptrdiff_t dstStride, int width, int height, int shift, int offset, int w0, int w1, const ClpRng& clpRng );
   void ( *copyBuffer )    ( const char*src,  ptrdiff_t srcStride,        char* dst, ptrdiff_t  dstStride,                                int width, int height );
-  void ( *padding1 )      (       Pel *dst,  ptrdiff_t stride,                                                                           int width, int height );
-  void ( *padding2 )      (       Pel *dst,  ptrdiff_t stride,                                                                           int width, int height );
   void ( *transpose4x4 )  ( const Pel* src,  ptrdiff_t srcStride, Pel* dst, ptrdiff_t dstStride );
   void ( *transpose8x8 )  ( const Pel* src,  ptrdiff_t srcStride, Pel* dst, ptrdiff_t dstStride );
   void ( *applyLut )      (       Pel* ptr,  ptrdiff_t ptrStride, int width, int height, const Pel* lut );
+  void ( *rspFwd )        (       Pel* ptr,  ptrdiff_t ptrStride, int width, int height, const int bd,                   const Pel OrgCW,  const Pel* LmcsPivot, const Pel* ScaleCoeff, const Pel* InputPivot );
+  void ( *rspBcw )        (       Pel* ptr,  ptrdiff_t ptrStride, int width, int height, const int bd, const int minBin, const int maxBin, const Pel* LmcsPivot, const Pel* InvScCoeff, const Pel* InputPivot );
   void ( *fillN_CU )      (       CodingUnit** ptr, ptrdiff_t ptrStride, int width, int height, CodingUnit* cuPtr );
 
   void (*sampleRateConv)  ( const std::pair<int, int> scalingRatio, const std::pair<int, int> compScale,
-                            const Pel* orgSrc, const int orgStride, const int orgWidth, const int orgHeight,
+                            const Pel* orgSrc, const ptrdiff_t orgStride, const int orgWidth, const int orgHeight,
                             const int beforeScaleLeftOffset, const int beforeScaleTopOffset,
-                            Pel* scaledSrc, const int scaledStride, const int scaledWidth, const int scaledHeight,
+                            Pel* scaledSrc, const ptrdiff_t scaledStride, const int scaledWidth, const int scaledHeight,
                             const int afterScaleLeftOffset, const int afterScaleTopOffset,
                             const int bitDepth, const bool useLumaFilter, const bool horCollocatedPositionFlag, const bool verCollocatedPositionFlag );
 };
-#endif
 
 extern PelBufferOps g_pelBufOP;
 
@@ -126,7 +120,15 @@ struct AreaBuf : public Size
   AreaBuf( T *_buf, const ptrdiff_t &_stride, const Size &size )                                : Size( size ),            buf( _buf ), stride( _stride )    { }
   AreaBuf( T *_buf, const SizeType &_width, const SizeType &_height )                           : Size( _width, _height ), buf( _buf ), stride( _width )     { }
   AreaBuf( T *_buf, const ptrdiff_t &_stride, const SizeType &_width, const SizeType &_height ) : Size( _width, _height ), buf( _buf ), stride( _stride )    { }
-  AreaBuf( const AreaBuf<typename std::remove_const<T>::type >& other )                         : Size( other ),           buf( other.buf ), stride( other.stride ) { }
+
+  AreaBuf( const AreaBuf& )  = default;
+  AreaBuf(       AreaBuf&& ) = default;
+  AreaBuf& operator=( const AreaBuf& )  = default;
+  AreaBuf& operator=(       AreaBuf&& ) = default;
+
+  // conversion from AreaBuf<const T> to AreaBuf<T>
+  template<bool T_IS_CONST = std::is_const<T>::value>
+  AreaBuf( const AreaBuf<typename std::remove_const_t<T>>& other, std::enable_if_t<T_IS_CONST>* = 0) : Size( other ), buf( other.buf ), stride( other.stride ) { }
 
   void fill                 ( const T &val );
   void memset               ( const int val );
@@ -146,8 +148,7 @@ struct AreaBuf : public Size
 
   void transposedFrom       ( const AreaBuf<const T> &other );
 
-  void rspSignal            ( const Pel *lut );
-  void scaleSignal          ( const int scale, const ClpRng& clpRng);
+  void scaleSignal          ( const int scale, const ClpRng& clpRng );
 
   void rescaleBuf           ( const AreaBuf<const T>& beforeScaling, const ComponentID compID, const std::pair<int, int> scalingRatio, const Window& confBefore, const Window& confAfter, const ChromaFormat chromaFormatIDC, const BitDepths& bitDepths, const bool horCollocatedChromaFlag = false, const bool verCollocatedChromaFlag = false );
 
@@ -430,8 +431,8 @@ void AreaBuf<T>::copyFrom( const AreaBuf<const T> &other ) const
 template<typename T>
 void AreaBuf<T>::subtract( const AreaBuf<const T> &other )
 {
-  CHECK( width  != other.width,  "Incompatible size" );
-  CHECK( height != other.height, "Incompatible size" );
+  CHECK_RECOVERABLE( width  != other.width,  "Incompatible size" );
+  CHECK_RECOVERABLE( height != other.height, "Incompatible size" );
 
         T* dest =       buf;
   const T* subs = other.buf;
@@ -494,7 +495,7 @@ void AreaBuf<T>::extendBorderPel( unsigned margin )
   int       w = width;
   ptrdiff_t s = stride;
 
-  CHECK( ( w + 2 * margin ) > s, "Size of buffer too small to extend" );
+  CHECK_RECOVERABLE( ( w + 2 * margin ) > s, "Size of buffer too small to extend" );
   // do left and right margins
   for( int y = 0; y < h; y++ )
   {
@@ -526,7 +527,7 @@ void AreaBuf<T>::extendBorderPel( unsigned margin )
 template<typename T>
 void AreaBuf<T>::extendBorderPel(unsigned margin, bool left, bool right, bool top, bool bottom)
 {
-  CHECK( ( width + left*margin + right*margin) > stride, "Size of buffer too small to extend" );
+  CHECK_RECOVERABLE( ( width + left*margin + right*margin) > stride, "Size of buffer too small to extend" );
   // do left and right margins
 
   if( left && right )
@@ -600,12 +601,12 @@ void AreaBuf<T>::extendBorderPel(unsigned margin, bool left, bool right, bool to
 template<typename T>
 void AreaBuf<T>::padBorderPel( unsigned marginX, unsigned marginY, int dir )
 {
-  T*  p = buf;
-  int s = stride;
-  int h = height;
-  int w = width;
+  T*   p = buf;
+  auto s = stride;
+  int  h = height;
+  int  w = width;
 
-  CHECK( w  > s, "Size of buffer too small to extend" );
+  CHECK_RECOVERABLE( w  > s, "Size of buffer too small to extend" );
 
   // top-left margin
   if ( dir == 1 )
@@ -643,7 +644,7 @@ template<> void AreaBuf<Pel>::transposedFrom( const AreaBuf<const Pel> &other );
 template<typename T>
 void AreaBuf<T>::transposedFrom( const AreaBuf<const T> &other )
 {
-  CHECK( width * height != other.width * other.height, "Incompatible size" );
+  CHECK_RECOVERABLE( width * height != other.width * other.height, "Incompatible size" );
 
         T* dst  =       buf;
   const T* src  = other.buf;
@@ -674,7 +675,6 @@ template<typename T>
 struct UnitBuf
 {
   typedef static_vector<AreaBuf<T>,       MAX_NUM_COMPONENT> UnitBufBuffers;
-  typedef static_vector<AreaBuf<const T>, MAX_NUM_COMPONENT> ConstUnitBufBuffers;
 
   ChromaFormat chromaFormat;
   UnitBufBuffers bufs;
@@ -686,14 +686,15 @@ struct UnitBuf
   UnitBuf( const ChromaFormat &_chromaFormat,       AreaBuf<T> &&blkY ) : chromaFormat( _chromaFormat ), bufs{ std::forward<AreaBuf<T> >(blkY) } { }
   UnitBuf( const ChromaFormat &_chromaFormat, const AreaBuf<T>  &blkY, const AreaBuf<T>  &blkCb, const AreaBuf<T>  &blkCr ) : chromaFormat( _chromaFormat ), bufs{ blkY, blkCb, blkCr } { if( chromaFormat == CHROMA_400 ) bufs.resize( 1 ); }
   UnitBuf( const ChromaFormat &_chromaFormat,       AreaBuf<T> &&blkY,       AreaBuf<T> &&blkCb,       AreaBuf<T> &&blkCr ) : chromaFormat( _chromaFormat ), bufs{ std::forward<AreaBuf<T> >(blkY), std::forward<AreaBuf<T> >(blkCb), std::forward<AreaBuf<T> >(blkCr) } { if( chromaFormat == CHROMA_400 ) bufs.resize( 1 ); }
-  UnitBuf( const UnitBuf<typename std::remove_const<T>::type>& other ) : chromaFormat( other.chromaFormat ), bufs{}
-  {
-    // TODO: delete to avoid unneccessary copying
-    for( auto &buf : other.bufs )
-    {
-      bufs.push_back( buf );
-    }
-  }
+
+  UnitBuf( const UnitBuf& other )  = default;
+  UnitBuf(       UnitBuf&& other ) = default;
+  UnitBuf& operator=( const UnitBuf& other )  = default;
+  UnitBuf& operator=(       UnitBuf&& other ) = default;
+
+  // conversion from UnitBuf<const T> to UnitBuf<T>
+  template<bool T_IS_COST = std::is_const<T>::value>
+  UnitBuf( const UnitBuf<typename std::remove_const_t<T>>& other, std::enable_if_t<T_IS_COST>* = 0 ) : chromaFormat( other.chromaFormat ), bufs( other.bufs ) { }
 
         AreaBuf<T>& get( const ComponentID comp )        { return bufs[comp]; }
   const AreaBuf<T>& get( const ComponentID comp )  const { return bufs[comp]; }
@@ -744,7 +745,7 @@ void UnitBuf<T>::fill( const T &val )
 template<typename T>
 void UnitBuf<T>::copyFrom( const UnitBuf<const T> &other ) const
 {
-  CHECK( chromaFormat != other.chromaFormat, "Incompatible formats" );
+  CHECK_RECOVERABLE( chromaFormat != other.chromaFormat, "Incompatible formats" );
 
   for( unsigned i = 0; i < bufs.size(); i++ )
   {
@@ -757,7 +758,7 @@ void UnitBuf<T>::copyFrom( const UnitBuf<const T> &other ) const
 template<typename T>
 void UnitBuf<T>::subtract( const UnitBuf<const T> &other )
 {
-  CHECK( chromaFormat != other.chromaFormat, "Incompatible formats" );
+  CHECK_RECOVERABLE( chromaFormat != other.chromaFormat, "Incompatible formats" );
 
   for( unsigned i = 0; i < bufs.size(); i++ )
   {
@@ -768,8 +769,8 @@ void UnitBuf<T>::subtract( const UnitBuf<const T> &other )
 template<typename T>
 void UnitBuf<T>::reconstruct(const UnitBuf<const T> &pred, const UnitBuf<const T> &resi, const ClpRngs& clpRngs)
 {
-  CHECK( chromaFormat != pred.chromaFormat, "Incompatible formats" );
-  CHECK( chromaFormat != resi.chromaFormat, "Incompatible formats" );
+  CHECK_RECOVERABLE( chromaFormat != pred.chromaFormat, "Incompatible formats" );
+  CHECK_RECOVERABLE( chromaFormat != resi.chromaFormat, "Incompatible formats" );
 
   for( unsigned i = 0; i < bufs.size(); i++ )
   {
@@ -783,7 +784,7 @@ void UnitBuf<T>::addWeightedAvg(const UnitBuf<T> &other1, const UnitBuf<T> &othe
   const size_t istart = chromaOnly ? 1 : 0;
   const size_t iend   = lumaOnly   ? 1 : bufs.size();
 
-  CHECK(lumaOnly && chromaOnly, "should not happen");
+  CHECK_RECOVERABLE(lumaOnly && chromaOnly, "should not happen");
 
   for(size_t i = istart; i < iend; i++)
   {
@@ -797,7 +798,7 @@ void UnitBuf<T>::addAvg(const UnitBuf<T> &other1, const UnitBuf<T> &other2, cons
   const size_t istart = chromaOnly ? 1 : 0;
   const size_t iend   = lumaOnly   ? 1 : bufs.size();
 
-  CHECK( lumaOnly && chromaOnly, "should not happen" );
+  CHECK_RECOVERABLE( lumaOnly && chromaOnly, "should not happen" );
 
   for( size_t i = istart; i < iend; i++)
   {
@@ -936,7 +937,7 @@ struct PelStorage : public PelUnitBuf
   void swap( PelStorage& other );
   void createFromBuf( PelUnitBuf buf );
   void create( const UnitArea &_unit );
-  void create( const ChromaFormat _chromaFormat, const Size& _size, const unsigned _maxCUSize = 0, const unsigned _margin = 0, const unsigned _alignment = 0, const bool _scaleChromaMargin = true );
+  void create( const ChromaFormat _chromaFormat, const Size& _size, const unsigned _maxCUSize = 0, const unsigned _margin = 0, const unsigned _alignment = 0, const bool _scaleChromaMargin = true, const UserAllocator* userAlloc = nullptr );
   void destroy();
 
          PelBuf getBuf( const CompArea &blk );
@@ -950,10 +951,19 @@ struct PelStorage : public PelUnitBuf
   Pel *getOrigin( const int id ) const { return m_origin[id]; }
   PelBuf getOriginBuf( const int id ) { return PelBuf( m_origin[id], m_origSi[id] ); }
 
+  Size  getBufSize( const int id )      const { return  m_origSi[id];     }
+  void *getBufAllocator( const int id ) const { return m_allocator[id];   }
+  bool  isExternAllocator()             const { return m_externAllocator; }
+
+  const UserAllocator* getUserAllocator() const { return m_userAlloc; }
+
 private:
 
-  Size m_origSi[MAX_NUM_COMPONENT];
-  Pel *m_origin[MAX_NUM_COMPONENT];
+  Size    m_origSi[MAX_NUM_COMPONENT];
+  Pel    *m_origin[MAX_NUM_COMPONENT];
+  void   *m_allocator[MAX_NUM_COMPONENT];
+  bool    m_externAllocator = false;
+  const UserAllocator* m_userAlloc  = nullptr;
 };
 
 }
