@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -59,8 +55,6 @@ THE POSSIBILITY OF SUCH DAMAGE.
 namespace vvdec
 {
 
-ThreadSafeCUCache g_globalUnitCache{};
-
 const UnitScale UnitScaleArray[NUM_CHROMA_FORMAT][MAX_NUM_COMPONENT] =
 {
   { {2,2}, {0,0}, {0,0} },  // 4:0:0
@@ -73,19 +67,20 @@ const UnitScale UnitScaleArray[NUM_CHROMA_FORMAT][MAX_NUM_COMPONENT] =
 // coding structure method definitions
 // ---------------------------------------------------------------------------
 
-CodingStructure::CodingStructure(std::shared_ptr<CUCache> cuCache, std::shared_ptr<TUCache> tuCache )
+CodingStructure::CodingStructure( CUChunkCache* cuChunkCache, TUChunkCache* tuChunkCache )
   : area      ()
   , picture   ( nullptr )
   , m_ctuData ( nullptr )
   , m_ctuDataSize( 0 )
-  , m_predBufSize ( 0 )
   , m_dmvrMvCache ( nullptr )
-  , m_dmvrMvCacheSize( 0 )
-  , m_cuCache ( cuCache )
-  , m_tuCache ( tuCache )
+  , m_cuCache ( cuChunkCache )
+  , m_tuCache ( tuChunkCache )
+  , m_cuMap   ( nullptr )
+  , m_cuMapSize( 0 )
+  , m_colMiMap ( nullptr )
+  , m_colMiMapSize ( 0 )
   , m_IBCBufferWidth( 0 )
 {
-  m_dmvrMvCacheOffset = 0;
 }
 
 void CodingStructure::destroy()
@@ -94,37 +89,38 @@ void CodingStructure::destroy()
 
   m_reco.destroy();
   m_rec_wrap.destroy();
-  m_pred.destroy();
 
-  if( m_dmvrMvCache )
-  {
-    free( m_dmvrMvCache );
-    m_dmvrMvCache = nullptr;
-    m_dmvrMvCacheSize = 0;
-  }
-
-  m_cuCache->defragment();
-  m_tuCache->defragment();
-  
   m_virtualIBCbuf.clear();
 
-  if( m_ctuData )
-  {
-    free( m_ctuData );
-    m_ctuData = nullptr;
-    m_ctuDataSize = 0;
-  }
+  deallocTempInternals();
 
-  if( m_predBuf )
-  {
-    m_predBuf.reset();
-    m_predBufSize = 0;
-  }
+  if( m_ctuData ) free( m_ctuData );
+  m_ctuData = nullptr;
+  m_ctuDataSize = 0;
+
+  if( m_colMiMap ) free( m_colMiMap );
+  m_colMiMap = nullptr;
+  m_colMiMapSize = 0;
+
+  if( m_cuMap ) free( m_cuMap );
+  m_cuMap = nullptr;
+  m_cuMapSize = 0;
+}
+
+void CodingStructure::resetForUse()
+{
+  vps.reset();
+  sps.reset();
+  pps.reset();
+  picHeader.reset();
+  std::fill( std::begin( alfApss ), std::end( alfApss ), nullptr );
+  lmcsAps.reset();
+  pcv = nullptr;
 }
 
 CodingUnit& CodingStructure::addCU( const UnitArea &unit, const ChannelType chType, const TreeType treeType, const ModeType modeType, const CodingUnit *cuLeft, const CodingUnit *cuAbove )
 {
-  CodingUnit *cu = m_cuCache->get();
+  CodingUnit *cu = m_cuCache.get();
 
   GCC_WARNING_DISABLE_class_memaccess
   memset( cu, 0, sizeof( CodingUnit ) );
@@ -135,25 +131,26 @@ CodingUnit& CodingStructure::addCU( const UnitArea &unit, const ChannelType chTy
   cu->setChType  ( chType );
   cu->setTreeType( treeType );
   cu->setModeType( modeType );
-  
-  CodingUnit *prevCU = m_lastCU;
 
   const int currRsAddr = ctuRsAddr( unit.blocks[chType].pos(), chType );
-  const int prevRsAddr = prevCU ? ctuRsAddr( prevCU->blocks[prevCU->chType()].pos(), prevCU->chType() ) : -1;
-
-  if( prevCU && currRsAddr == prevRsAddr )
-  {
-    prevCU->next = cu;
-  }
-
-  cu->idx = ++m_numCUs;
 
   uint32_t numCh = getNumberValidChannels( area.chromaFormat );
 
   CtuData& ctuData = getCtuData( currRsAddr );
-  cu->ctuData = &ctuData;
+  cu->ctuData      = &ctuData;
 
-  cu->predBuf = m_predBuf.get() + m_predBufOffset;
+  if( !ctuData.firstCU )
+  {
+    ctuData.firstCU = cu;
+  }
+
+  cu->idx = ++ctuData.numCUs;
+
+  CodingUnit* prevCU = cu;
+  std::swap( ctuData.lastCU, prevCU );
+  if( prevCU ) prevCU->next = cu;
+
+  cu->predBufOff = ctuData.predBufOffset;
 
   for( uint32_t i = 0; i < numCh; i++ )
   {
@@ -166,18 +163,22 @@ CodingUnit& CodingStructure::addCU( const UnitArea &unit, const ChannelType chTy
 
     if( i )
     {
-      m_predBufOffset += ( cuArea << 1 );
+      ctuData.predBufOffset += ( cuArea << 1 );
     }
     else
     {
-      m_predBufOffset += cuArea;
+      ctuData.predBufOffset += cuArea;
     }
 
     const ptrdiff_t  stride = ptrdiff_t( 1 ) << m_ctuWidthLog2[i];
-    const Area      &_blk   = cu-> blocks[i];
+    const Area&      _blk   = cu->blocks[i];
     const UnitScale  scale  = unitScale[i];
+    const int        sclX   = scale.scaleHor( _blk.x );
+    const int        sclY   = scale.scaleVer( _blk.y );
+    const int        sclW   = scale.scaleHor( _blk.width );
+    const int        sclH   = scale.scaleVer( _blk.height );
 
-    g_pelBufOP.fillN_CU( ctuData.cuPtr[i] + inCtuPos( _blk, ChannelType( i ) ), stride, scale.scaleHor( _blk.width ), scale.scaleVer( _blk.height ), cu );
+    g_pelBufOP.fillN_CU( ctuData.cuPtr[i] + ( sclX & m_ctuSizeMask[i] ) + ( ( sclY & m_ctuSizeMask[i] ) << m_ctuWidthLog2[i] ), stride, sclW, sclH, cu );
 
     if( i == chType )
     {
@@ -191,13 +192,10 @@ CodingUnit& CodingStructure::addCU( const UnitArea &unit, const ChannelType chTy
 
   if( isLuma( chType ) && unit.lheight() >= 8 && unit.lwidth()  >= 8 && unit.Y().area() >= 128 )
   {
-    CHECKD( m_dmvrMvCacheOffset >= m_dmvrMvCacheSize, "dmvr cache offset out of bounds" )
-    pu.mvdL0SubPu       = &m_dmvrMvCache[m_dmvrMvCacheOffset];
-    m_dmvrMvCacheOffset += std::max<int>( 1, unit.lwidth() >> DMVR_SUBCU_WIDTH_LOG2 ) * std::max<int>( 1, unit.lheight() >> DMVR_SUBCU_HEIGHT_LOG2 );
+    pu.mvdL0SubPuOff           = ctuData.dmvrMvCacheOffset;
+    ctuData.dmvrMvCacheOffset += std::max<int>( 1, unit.lwidth() >> DMVR_SUBCU_WIDTH_LOG2 ) * std::max<int>( 1, unit.lheight() >> DMVR_SUBCU_HEIGHT_LOG2 );
   }
 
-  m_lastCU = cu;
-
   return *cu;
 }
 
@@ -211,7 +209,7 @@ TransformUnit& CodingStructure::addTU( const UnitArea &unit, const ChannelType c
   }
   else
   {
-    tu = m_tuCache->get();
+    tu = m_tuCache.get();
 
     GCC_WARNING_DISABLE_class_memaccess
     memset( tu, 0, sizeof( TransformUnit ) );
@@ -221,7 +219,7 @@ TransformUnit& CodingStructure::addTU( const UnitArea &unit, const ChannelType c
     cu.lastTU       = tu;
   }
 
-  tu->idx               = ++m_numTUs;
+  tu->idx               = ++cu.ctuData->numTUs;
   tu->cu                =  &cu;
   tu->setChType         (   chType );
   tu->UnitArea::operator=(  unit );
@@ -250,30 +248,11 @@ void CodingStructure::addEmptyTUs( Partitioner &partitioner, CodingUnit& cu )
   }
 }
 
-CUTraverser CodingStructure::traverseCUs( const UnitArea& unit )
+CUTraverser CodingStructure::traverseCUs( const int ctuRsAddr )
 {
-  ChannelType lastChan  = unit.chromaFormat != CHROMA_400 ? CH_C : CH_L;
-  CodingUnit* firstCU   = getCU( unit.lumaPos(), CH_L );
-  CodingUnit* lastCU    = getCU( unit.block( getFirstComponentOfChannel( lastChan ) ).bottomRight(), lastChan );
+  CtuData& ctuData = m_ctuData[ctuRsAddr];
 
-  CHECKD( !firstCU || !lastCU || ctuRsAddr( firstCU->lumaPos(), CH_L ) != ctuRsAddr( lastCU->blocks[lastChan].pos(), lastChan ), "First CU and/or Last CU non-existent not in the same CTU!" );
-
-  if( lastCU ) lastCU = lastCU->next;
-
-  return CUTraverser( firstCU, lastCU );
-}
-
-cCUTraverser CodingStructure::traverseCUs( const UnitArea& unit ) const
-{
-  ChannelType lastChan      = unit.chromaFormat != CHROMA_400 ? CH_C : CH_L;
-  const CodingUnit* firstCU = getCU( unit.lumaPos(), CH_L );
-  const CodingUnit* lastCU  = getCU( unit.block( getFirstComponentOfChannel( lastChan ) ).bottomRight(), lastChan );
-
-  CHECKD( !firstCU || !lastCU || ctuRsAddr( firstCU->lumaPos(), CH_L ) != ctuRsAddr( lastCU->blocks[lastChan].pos(), lastChan ), "First CU and/or Last CU non-existent not in the same CTU!" );
-
-  if( lastCU ) lastCU = lastCU->next;
-
-  return cCUTraverser( firstCU, lastCU );
+  return CUTraverser( ctuData.firstCU, ctuData.lastCU->next );
 }
 
 // coding utilities
@@ -295,22 +274,6 @@ void CodingStructure::createInternals( const UnitArea& _unit )
   memcpy( unitScale, UnitScaleArray[area.chromaFormat], sizeof( unitScale ) );
 
   picture = nullptr;
-
-  // for the worst case of all PUs being 8x8 and using DMVR
-  const size_t _maxNumDmvrMvs = ( area.lwidth() >> 3 ) * ( area.lheight() >> 3 );
-  if( _maxNumDmvrMvs != m_dmvrMvCacheSize )
-  {
-    if( m_dmvrMvCache ) free( m_dmvrMvCache );
-    m_dmvrMvCacheSize = _maxNumDmvrMvs;
-    m_dmvrMvCache = ( Mv* ) malloc( sizeof( Mv ) * _maxNumDmvrMvs );
-  }
-
-  size_t predBufSize = area.Y().area() + ( isChromaEnabled( _unit.chromaFormat ) ? ( area.Cb().area() + area.Cr().area() ) : 0 );
-  if( predBufSize != m_predBufSize )
-  {
-    m_predBuf.reset( (Pel*) xMalloc( Pel, predBufSize ) );
-    m_predBufSize = predBufSize;
-  }
 }
 
 
@@ -322,14 +285,43 @@ void CodingStructure::rebindPicBufs()
   else                                                    m_rec_wrap.destroy();
 }
 
+void CodingStructure::allocTempInternals()
+{
+  const ptrdiff_t ctuCuMapSize    = pcv->num4x4CtuBlks;
+  const ptrdiff_t ctuColMiMapSize = pcv->num8x8CtuBlks;
+
+  if( m_cuMapSize != ctuCuMapSize * pcv->sizeInCtus * 2 )
+  {
+    if( m_cuMap ) free( m_cuMap );
+    m_cuMapSize = ctuCuMapSize * pcv->sizeInCtus * 2;
+    m_cuMap     = ( CodingUnit** ) malloc( sizeof( CodingUnit* ) * m_cuMapSize );
+  }
+
+  if( m_colMiMapSize != ctuColMiMapSize * pcv->sizeInCtus )
+  {
+    if( m_colMiMap ) free( m_colMiMap );
+    m_colMiMapSize = ctuColMiMapSize * pcv->sizeInCtus;
+    m_colMiMap     = ( ColocatedMotionInfo* ) malloc( sizeof( ColocatedMotionInfo ) * m_colMiMapSize );
+  }
+  
+  if( m_ctuDataSize != pcv->sizeInCtus )
+  {
+    m_ctuDataSize = pcv->sizeInCtus;
+    if( m_ctuData ) free( m_ctuData );
+    m_ctuData = ( CtuData* ) malloc( m_ctuDataSize * sizeof( CtuData ) );
+  }
+}
+
+void CodingStructure::deallocTempInternals()
+{
+  m_cuCache.releaseAll();
+  m_tuCache.releaseAll();
+}
+
 void CodingStructure::initStructData()
 {
-  m_numCUs = 0;
-  m_numTUs = 0;
-  m_lastCU = nullptr;
-
-  m_cuCache->defragment();
-  m_tuCache->defragment();
+  m_cuCache.releaseAll();
+  m_tuCache.releaseAll();
 
   m_widthInCtus = pcv->widthInCtus;
 
@@ -339,22 +331,41 @@ void CodingStructure::initStructData()
   m_ctuWidthLog2[0] = pcv->maxCUWidthLog2 - unitScale[CH_L].posx;
   m_ctuWidthLog2[1] = m_ctuWidthLog2[0]; // same for luma and chroma, because of the 2x2 blocks
 
-  if( m_ctuDataSize != pcv->sizeInCtus )
+  GCC_WARNING_DISABLE_class_memaccess
+  memset( m_ctuData,             0, sizeof( CtuData             ) * m_ctuDataSize );
+  memset( m_cuMap,               0, sizeof( CodingUnit*         ) * m_cuMapSize );
+  memset( m_colMiMap, CO_NOT_VALID, sizeof( ColocatedMotionInfo ) * m_colMiMapSize );
+  GCC_WARNING_RESET
+
+  const ptrdiff_t ctuSampleSizeL  = pcv->maxCUHeight * pcv->maxCUWidth;
+  const ptrdiff_t ctuSampleSizeC  = isChromaEnabled( pcv->chrFormat ) ? ( ctuSampleSizeL >> ( getChannelTypeScaleX( CH_C, pcv->chrFormat) + getChannelTypeScaleY( CH_C, pcv->chrFormat ) ) ) : 0;
+  const ptrdiff_t ctuSampleSize   = ctuSampleSizeL + 2 * ctuSampleSizeC;
+  const ptrdiff_t ctuCuMapSize    = pcv->num4x4CtuBlks;
+  const ptrdiff_t ctuColMiMapSize = pcv->num8x8CtuBlks;
+
+  hasIbcBlock.clear();
+  hasIbcBlock.resize( pcv->heightInCtus, 0 );
+
+  for( int y = 0; y < pcv->heightInCtus; y++ )
   {
-    m_ctuDataSize = pcv->sizeInCtus;
-    if( m_ctuData ) free( m_ctuData );
-    m_ctuData = ( CtuData* ) malloc( m_ctuDataSize * sizeof( CtuData ) );
+    for( int x = 0; x < pcv->widthInCtus; x++ )
+    {
+      int i = y * pcv->widthInCtus + x;
+
+      m_ctuData[i].lineIdx = y;
+      m_ctuData[i].colIdx  = x;
+      m_ctuData[i].ctuIdx  = i;
+
+      for( int j = 0; j < 2; j++ )
+      {
+        m_ctuData[i].cuPtr[j] = &m_cuMap[( 2 * i + j ) * ctuCuMapSize];
+      }
+
+      m_ctuData[i].colMotion         = &m_colMiMap[i * ctuColMiMapSize];
+      m_ctuData[i].predBufOffset     = i * ctuSampleSize;
+      m_ctuData[i].dmvrMvCacheOffset = i * pcv->num8x8CtuBlks;
+    }
   }
-
-  for( int i = 0; i < m_ctuDataSize; i++ )
-  {
-    CtuData& ctuData = m_ctuData[i];
-    memset( ctuData.cuPtr, 0, sizeof( ctuData.cuPtr ) );
-  }
-
-  m_dmvrMvCacheOffset = 0;
-
-  m_predBufOffset = 0;
 }
 
 MotionBuf CodingStructure::getMotionBuf( const Area& _area )
@@ -385,7 +396,7 @@ PelUnitBuf CodingStructure::getPredBuf(const CodingUnit &unit)
 
   if( unit.Y().valid() )
   {
-    ret.bufs[0].buf    = unit.predBuf;
+    ret.bufs[0].buf    = m_predBuf + unit.predBufOff;
     ret.bufs[0].stride = unit.blocks[0].width;
     ret.bufs[0].width  = unit.blocks[0].width;
     ret.bufs[0].height = unit.blocks[0].height;
@@ -395,7 +406,7 @@ PelUnitBuf CodingStructure::getPredBuf(const CodingUnit &unit)
   {
     if( unit.Cb().valid() )
     {
-      ret.bufs[1].buf    = unit.predBuf + unit.Y().area();
+      ret.bufs[1].buf    = m_predBuf + unit.predBufOff + unit.Y().area();
       ret.bufs[1].stride = unit.blocks[1].width;
       ret.bufs[1].width  = unit.blocks[1].width;
       ret.bufs[1].height = unit.blocks[1].height;
@@ -403,7 +414,7 @@ PelUnitBuf CodingStructure::getPredBuf(const CodingUnit &unit)
 
     if( unit.Cr().valid() )
     {
-      ret.bufs[2].buf    = unit.predBuf + unit.Y().area() + unit.Cb().area();
+      ret.bufs[2].buf    = m_predBuf + unit.predBufOff + unit.Y().area() + unit.Cb().area();
       ret.bufs[2].stride = unit.blocks[2].width;
       ret.bufs[2].width  = unit.blocks[2].width;
       ret.bufs[2].height = unit.blocks[2].height;
@@ -421,7 +432,7 @@ const CPelUnitBuf CodingStructure::getPredBuf(const CodingUnit &unit) const
 
   if( unit.Y().valid() )
   {
-    ret.bufs[0].buf    = unit.predBuf;
+    ret.bufs[0].buf    = m_predBuf + unit.predBufOff;
     ret.bufs[0].stride = unit.blocks[0].width;
     ret.bufs[0].width  = unit.blocks[0].width;
     ret.bufs[0].height = unit.blocks[0].height;
@@ -429,7 +440,7 @@ const CPelUnitBuf CodingStructure::getPredBuf(const CodingUnit &unit) const
 
   if( unit.Cb().valid() )
   {
-    ret.bufs[1].buf    = unit.predBuf + unit.Y().area();
+    ret.bufs[1].buf    = m_predBuf + unit.predBufOff + unit.Y().area();
     ret.bufs[1].stride = unit.blocks[1].width;
     ret.bufs[1].width  = unit.blocks[1].width;
     ret.bufs[1].height = unit.blocks[1].height;
@@ -437,7 +448,7 @@ const CPelUnitBuf CodingStructure::getPredBuf(const CodingUnit &unit) const
 
   if( unit.Cr().valid() )
   {
-    ret.bufs[2].buf    = unit.predBuf + unit.Y().area() + unit.Cb().area();
+    ret.bufs[2].buf    = m_predBuf + unit.predBufOff + unit.Y().area() + unit.Cb().area();
     ret.bufs[2].stride = unit.blocks[2].width;
     ret.bufs[2].width  = unit.blocks[2].width;
     ret.bufs[2].height = unit.blocks[2].height;
@@ -446,6 +457,17 @@ const CPelUnitBuf CodingStructure::getPredBuf(const CodingUnit &unit) const
   return ret;
 }
 
+const ColocatedMotionInfo& CodingStructure::getColInfo( const Position &pos, const Slice*& pColSlice ) const
+{
+  const CtuData& ctuData    = getCtuData( ctuRsAddr( pos, CH_L ) );
+  const ptrdiff_t rsPos     = colMotPos( pos );
+  const ColocatedMotionInfo&
+                 colMi      = ctuData.colMotion[rsPos];
+                 pColSlice  = ctuData.slice;
+  
+  return colMi;
+}
+
 const CodingUnit* CodingStructure::getCURestricted( const Position &pos, const CodingUnit& curCu, const ChannelType _chType, const CodingUnit* guess ) const
 {
   if( guess && guess->blocks[_chType].contains( pos ) ) return guess;
@@ -462,7 +484,7 @@ const CodingUnit* CodingStructure::getCURestricted( const Position &pos, const C
   {
     cu = curCu.ctuData->cuPtr[_chType][inCtuPos( pos, _chType )];
   }
-  else if( ydiff > 0 || xdiff > ( 1 - sps->getEntropyCodingSyncEnabledFlag() ) )
+  else if( ydiff > 0 || xdiff > ( 1 - sps->getEntropyCodingSyncEnabledFlag() ) || ( ydiff == 0 && xdiff > 0 ) )
   {
     return nullptr;
   }
@@ -471,7 +493,7 @@ const CodingUnit* CodingStructure::getCURestricted( const Position &pos, const C
     cu = getCU( pos, _chType );
   }
 
-  if( !cu || cu->idx > curCu.idx ) return nullptr;
+  if( !cu || ( sameCTU && cu->idx > curCu.idx ) ) return nullptr;
   else if( sameCTU ) return cu;
 
   if( cu->slice->getIndependentSliceIdx() == curCu.slice->getIndependentSliceIdx() && cu->tileIdx == curCu.tileIdx )
@@ -488,18 +510,27 @@ const CodingUnit* CodingStructure::getCURestricted( const Position &pos, const P
 {
   const int yshift     = pcv->maxCUWidthLog2 - getChannelTypeScaleY( _chType, area.chromaFormat );
   const int ydiff      = ( pos.y >> yshift ) - ( curPos.y >> yshift ); // ( a <= b ) ==> a - b <= 0
-  const CodingUnit* cu = ydiff <= 0 ? getCU( pos, _chType ) : nullptr;
   const int xshift     = pcv->maxCUWidthLog2 - getChannelTypeScaleX( _chType, area.chromaFormat );
   const int xdiff      = ( pos.x >> xshift ) - ( curPos.x >> xshift );
   const bool sameCTU   = !ydiff && !xdiff;
 
-  if( cu && ( sameCTU || ( cu->slice->getIndependentSliceIdx() == curSliceIdx && cu->tileIdx == curTileIdx ) ) )
+  const CodingUnit* cu = nullptr;
+  
+  if( sameCTU )
   {
-    if( xdiff > 0 && sps->getEntropyCodingSyncEnabledFlag() )
-    {
-      return nullptr;
-    }
+    return getCU( pos, _chType );
+  }
+  else if( ydiff > 0 || xdiff > ( 1 - sps->getEntropyCodingSyncEnabledFlag() ) )
+  {
+    return nullptr;
+  }
+  else
+  {
+    cu = getCU( pos, _chType );
+  }
 
+  if( cu && cu->slice->getIndependentSliceIdx() == curSliceIdx && cu->tileIdx == curTileIdx )
+  {
     return cu;
   }
   else
@@ -524,25 +555,22 @@ void CodingStructure::initVIbcBuf( int numCtuLines, ChromaFormat chromaFormatIDC
 
 void CodingStructure::fillIBCbuffer( CodingUnit &cu, int lineIdx )
 {
-  for( TransformUnit &tu : TUTraverser( &cu.firstTU, cu.lastTU->next ) )
+  for( const CompArea &area : cu.blocks )
   {
-    for( const CompArea &area : tu.blocks )
-    {
-      if (!area.valid())
-        continue;
+    if (!area.valid())
+      continue;
 
-      const unsigned int lcuWidth = sps->getMaxCUWidth();
-      const int shiftSampleHor = getComponentScaleX(area.compID(), cu.chromaFormat);
-      const int shiftSampleVer = getComponentScaleY(area.compID(), cu.chromaFormat);
-      const int ctuSizeVerLog2 = getLog2(lcuWidth) - shiftSampleVer;
-      const int pux = area.x & ((m_IBCBufferWidth >> shiftSampleHor) - 1);
-      const int puy = area.y & (( 1 << ctuSizeVerLog2 ) - 1);
-      const CompArea dstArea = CompArea(area.compID(), Position(pux, puy), Size(area.width, area.height));
-      CPelBuf srcBuf = getRecoBuf(area);
-      PelBuf dstBuf = m_virtualIBCbuf[lineIdx].getBuf(dstArea);
+    const unsigned int lcuWidth = sps->getMaxCUWidth();
+    const int shiftSampleHor = getComponentScaleX(area.compID(), cu.chromaFormat);
+    const int shiftSampleVer = getComponentScaleY(area.compID(), cu.chromaFormat);
+    const int ctuSizeVerLog2 = getLog2(lcuWidth) - shiftSampleVer;
+    const int pux = area.x & ((m_IBCBufferWidth >> shiftSampleHor) - 1);
+    const int puy = area.y & (( 1 << ctuSizeVerLog2 ) - 1);
+    const CompArea dstArea = CompArea(area.compID(), Position(pux, puy), Size(area.width, area.height));
+    CPelBuf srcBuf = getRecoBuf(area);
+    PelBuf dstBuf = m_virtualIBCbuf[lineIdx].getBuf(dstArea);
 
-      dstBuf.copyFrom(srcBuf);
-    }
+    dstBuf.copyFrom(srcBuf);
   }
 }
 
