@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -56,12 +52,26 @@ THE POSSIBILITY OF SUCH DAMAGE.
 #include <iostream>
 #include <iomanip>
 #include <limits>
-
+#include <cmath>   // needed for std::log2()
+#include <cstdarg>
 #include <functional>
 #include <mutex>
 
-namespace vvdec
-{
+#if defined( __x86_64__ ) || defined( _M_X64 ) || defined( __i386__ ) || defined( __i386 ) || defined( _M_IX86 )
+# define REAL_TARGET_X86 1
+#elif defined( __aarch64__ ) || defined( _M_ARM64 ) || defined( __arm__ ) || defined( _M_ARM )
+# define REAL_TARGET_ARM 1
+#elif defined( __wasm__ ) || defined( __wasm32__ )
+# define REAL_TARGET_WASM 1
+#endif
+
+#if defined( TARGET_SIMD_X86 )
+# ifdef _WIN32
+#  include <intrin.h>
+//# elif defined( __GNUC__ )
+//#  include <simde/x86/sse2.h>
+# endif
+#endif // TARGET_SIMD_X86
 
 #if defined( __INTEL_COMPILER )
 #pragma warning( disable : 1786 )
@@ -77,17 +87,12 @@ namespace vvdec
 
 #pragma warning( disable : 4996 )
 #endif // _MSC_VER > 1000
-}
 
-#include "TypeDef.h"
-#include "vvdec/version.h"
-
-namespace vvdec
-{
 
+// MS Visual Studio before 2017 does not support required C++14 features
 #ifdef _MSC_VER
-#if _MSC_VER <= 1500
-inline int64_t abs (int64_t x) { return _abs64(x); };
+#if _MSC_VER < 1910
+#error "MS Visual Studio version not supported. Please upgrade to Visual Studio 2017 or higher (or use other compilers)"
 #endif
 #endif
 
@@ -95,18 +100,16 @@ inline int64_t abs (int64_t x) { return _abs64(x); };
 
 // macros to selectively disable some usually useful warnings
 #if defined __GNUC__ && !defined __clang__
-# define GCC_WARNING_DISABLE_sequence_point      _Pragma("GCC diagnostic push"); _Pragma("GCC diagnostic ignored \"-Wsequence-point\"");
 # define GCC_WARNING_RESET                       _Pragma("GCC diagnostic pop");
 
 # define GCC_EXTRA_WARNING_switch_enum           _Pragma("GCC diagnostic push"); _Pragma("GCC diagnostic error \"-Wswitch-enum\"");
 #else
-# define GCC_WARNING_DISABLE_sequence_point
 # define GCC_WARNING_RESET
 
 # define GCC_EXTRA_WARNING_switch_enum
 #endif
 
-#if __GNUC__ >= 8
+#if __GNUC__ >= 8 && !defined __clang__
 # define GCC_WARNING_DISABLE_maybe_uninitialized _Pragma("GCC diagnostic push"); _Pragma("GCC diagnostic ignored \"-Wmaybe-uninitialized\"");
 # define GCC_WARNING_DISABLE_class_memaccess     _Pragma("GCC diagnostic push"); _Pragma("GCC diagnostic ignored \"-Wclass-memaccess\"");
 #else
@@ -115,52 +118,15 @@ inline int64_t abs (int64_t x) { return _abs64(x); };
 #endif
 
 
+#include "TypeDef.h"
+#include "vvdec/version.h"
+
+namespace vvdec
+{
+
 //! \ingroup CommonLib
 //! \{
 
-// ====================================================================================================================
-// Platform information
-// ====================================================================================================================
-
-#ifdef __GNUC__
-#define NVM_COMPILEDBY  "[GCC %d.%d.%d]", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__
-#ifdef __IA64__
-#define NVM_ONARCH    "[on 64-bit] "
-#else
-#define NVM_ONARCH    "[on 32-bit] "
-#endif
-#endif
-
-#ifdef __INTEL_COMPILER
-# ifndef NVM_COMPILEDBY
-#  define NVM_COMPILEDBY  "[ICC %d]", __INTEL_COMPILER
-# endif
-#elif  _MSC_VER
-#define NVM_COMPILEDBY  "[VS %d]", _MSC_VER
-#endif
-
-#ifndef NVM_COMPILEDBY
-#define NVM_COMPILEDBY "[Unk-CXX]"
-#endif
-
-#ifdef _WIN32
-#define NVM_ONOS        "[Windows]"
-#elif  __linux
-#define NVM_ONOS        "[Linux]"
-#elif  __CYGWIN__
-#define NVM_ONOS        "[Cygwin]"
-#elif __APPLE__
-#define NVM_ONOS        "[Mac OS X]"
-#else
-#define NVM_ONOS "[Unk-OS]"
-#endif
-
-#define NVM_BITS          "[%d bit] ", (sizeof(void*) == 8 ? 64 : 32) ///< used for checking 64-bit O/S
-
-#ifndef NULL
-#define NULL              0
-#endif
-
 // ====================================================================================================================
 // Common constants
 // ====================================================================================================================
@@ -179,6 +145,10 @@ static const int MAX_NUM_REF_PICS =                                16; ///< max.
 static const int MAX_NUM_REF =                                     16; ///< max. number of entries in picture reference list
 static const int MAX_QP =                                          63;
 static const int NOT_VALID =                                       -1;
+static const int MI_NOT_VALID =                                    -1;
+static const int MF_NOT_VALID =                                    -1;
+static const int MH_NOT_VALID =                                    -1;
+static const int CO_NOT_VALID =                                    -1;
 
 static const int AMVP_MAX_NUM_CANDS =                               2; ///< AMVP: advanced motion vector prediction - max number of final candidates
 static const int AMVP_MAX_NUM_CANDS_MEM =                           3; ///< AMVP: advanced motion vector prediction - max number of candidates
@@ -346,7 +316,7 @@ static const int BIO_TEMP_BUFFER_SIZE         =                     (MAX_BDOF_AP
 static const int PROF_BORDER_EXT_W            =                     1;
 static const int PROF_BORDER_EXT_H            =                     1;
 static const int BCW_NUM =                                          5; ///< the number of weight options
-static const int BCW_DEFAULT =                                      ((uint8_t)(BCW_NUM >> 1)); ///< Default weighting index representing for w=0.5
+static const int BCW_DEFAULT =                                      0; ///< Default weighting index representing for w=0.5, in the internal domain
 static const int BCW_SIZE_CONSTRAINT =                            256; ///< disabling Bcw if cu size is smaller than 256
 static const int MAX_NUM_HMVP_CANDS =                              (MRG_MAX_NUM_CANDS-1); ///< maximum number of HMVP candidates to be stored and used in merge list
 static const int MAX_NUM_HMVP_AVMPCANDS =                          4; ///< maximum number of HMVP candidates to be used in AMVP list
@@ -428,14 +398,11 @@ struct ClpRngTemplate
 };
 
 typedef ClpRngTemplate<Pel> ClpRng;
+typedef ClpRng ClpRngs;
 
-struct ClpRngs : ClpRng
-{
-};
-
-template <typename T> constexpr inline T Clip3  ( const T minVal, const T maxVal, const T a) { return std::min<T> (std::max<T> (minVal, a) , maxVal); }  ///< general min/max clip
-template <typename T> constexpr inline T ClipBD ( const T x, const int bitDepth )            { return Clip3( T( 0 ), T( ( 1 << bitDepth ) - 1 ), x ); }
-template <typename T> constexpr inline T ClipPel( const T a, const ClpRng& clpRng )          { return ClipBD( a, clpRng.bd ); }  ///< clip reconstruction
+template <typename T> constexpr static inline T Clip3  ( const T minVal, const T maxVal, const T a) { return std::min<T> (std::max<T> (minVal, a) , maxVal); }  ///< general min/max clip
+template <typename T> constexpr static inline T ClipBD ( const T x, const int bitDepth )            { return Clip3( T( 0 ), T( ( 1 << bitDepth ) - 1 ), x ); }
+template <typename T> constexpr static inline T ClipPel( const T a, const ClpRng& clpRng )          { return ClipBD( a, clpRng.bd ); }  ///< clip reconstruction
 
 static void default_msgFnc( void *, int level, const char* fmt, va_list args )
 {
@@ -446,14 +413,8 @@ extern MsgLevel g_verbosity;
 extern void    *g_context;
 extern std::function<void( void*, int, const char*, va_list )> g_msgFnc;
 
-}
 
-#include <stdarg.h>
-
-namespace vvdec
-{
-
-inline void msg( MsgLevel level, const char* fmt, ... )
+static inline void msg( MsgLevel level, const char* fmt, ... )
 {
   if ( vvdec::g_msgFnc && vvdec::g_verbosity >= level )
   {
@@ -481,7 +442,7 @@ inline void msg( MsgLevel level, const char* fmt, ... )
 #else
 namespace detail {
 template<typename T>
-T* aligned_malloc(size_t len, size_t alignement) {
+static inline T* aligned_malloc(size_t len, size_t alignement) {
   T* p = NULL;
   if( posix_memalign( (void**)&p, alignement, sizeof(T)*(len) ) )
   {
@@ -489,7 +450,7 @@ T* aligned_malloc(size_t len, size_t alignement) {
   }
   return p;
 }
-}
+}   // namespace detail
 #define xMalloc( type, len )        detail::aligned_malloc<type>( len, MEMORY_ALIGN_DEF_SIZE )
 #define xFree( ptr )                free( ptr )
 #endif
@@ -525,6 +486,16 @@ template<class T> struct AlignedDeleter
 #    define CLANG_VERSION_AT_LEAST(x,y) 0
 #endif
 
+#if defined( __GNUC__ )
+#if __has_attribute( no_sanitize )
+#    define NO_THREAD_SANITIZE __attribute__( ( no_sanitize( "thread" ) ) )
+#else
+#    define NO_THREAD_SANITIZE
+#endif
+#else
+#    define NO_THREAD_SANITIZE
+#endif
+
 #ifdef __GNUC__
 #    define ALWAYS_INLINE __attribute__((always_inline)) inline
 #elif defined _MSC_VER
@@ -535,8 +506,10 @@ template<class T> struct AlignedDeleter
 
 #if ENABLE_SIMD_OPT
 
-#ifdef TARGET_SIMD_X86
-typedef enum {
+#  ifdef TARGET_SIMD_X86
+typedef enum
+{
+  UNDEFINED = -1,
   SCALAR = 0,
   SSE41,
   SSE42,
@@ -544,56 +517,49 @@ typedef enum {
   AVX2,
   AVX512
 } X86_VEXT;
-#endif
 
-#ifdef TARGET_SIMD_X86
-X86_VEXT read_x86_extension_flags(const std::string &extStrId = std::string());
-const char* read_x86_extension(const std::string &extStrId);
-#endif
+#  endif
 
-#endif //ENABLE_SIMD_OPT
+#endif   // ENABLE_SIMD_OPT
 
-template <typename ValueType> inline ValueType leftShift       (const ValueType value, const int shift) { return (shift >= 0) ? ( value                                  << shift) : ( value                                   >> -shift); }
-template <typename ValueType> inline ValueType rightShift      (const ValueType value, const int shift) { return (shift >= 0) ? ( value                                  >> shift) : ( value                                   << -shift); }
-template <typename ValueType> inline ValueType leftShift_round (const ValueType value, const int shift) { return (shift >= 0) ? ( value                                  << shift) : ((value + (ValueType(1) << (-shift - 1))) >> -shift); }
-template <typename ValueType> inline ValueType rightShift_round(const ValueType value, const int shift) { return (shift >= 0) ? ((value + (ValueType(1) << (shift - 1))) >> shift) : ( value                                   << -shift); }
+template <typename ValueType> static inline ValueType rightShift      (const ValueType value, const int shift) { return (shift >= 0) ? ( value                                  >> shift) : ( value                                   << -shift); }
+template <typename ValueType> static inline ValueType rightShift_round(const ValueType value, const int shift) { return (shift >= 0) ? ((value + (ValueType(1) << (shift - 1))) >> shift) : ( value                                   << -shift); }
 
-#ifdef TARGET_SIMD_X86
-#ifdef _WIN32
-}
-# include <intrin.h>
-namespace vvdec {
-static inline unsigned long _bit_scan_reverse( long a )
+#if defined( _WIN32 ) && defined( TARGET_SIMD_X86 )
+static inline unsigned int bit_scan_reverse( int a )
 {
   unsigned long idx = 0;
   _BitScanReverse( &idx, a );
   return idx;
 }
-#else
+// disabled because it requires x86intrin.h which conflicts with simd-everywhere
+// #elif defined( __GNUC__ ) && defined( TARGET_SIMD_X86 ) && !defined( REAL_TARGET_WASM )
+// static inline unsigned int bit_scan_reverse( int a )
+// {
+//   return _bit_scan_reverse( a );
+// }
+#elif defined( __GNUC__ )
+static inline unsigned int bit_scan_reverse( int a )
+{
+  return __builtin_clz( a ) ^ ( 8 * sizeof( a ) - 1 );
 }
-# include <x86intrin.h>
-namespace vvdec {
 #endif
 
-#endif
-#if ENABLE_SIMD_LOG2 && defined( TARGET_SIMD_X86 )
-static inline int getLog2( long val )
+#if ENABLE_SIMD_LOG2
+static inline int getLog2( int val )
 {
-  return _bit_scan_reverse( val );
+  return bit_scan_reverse( val );
 }
 #else
-}
-#include <cmath>
-namespace vvdec {
 extern int8_t g_aucLog2[MAX_CU_SIZE + 1];
-static inline int getLog2( long val )
+static inline int getLog2( int val )
 {
   CHECKD( g_aucLog2[2] != 1, "g_aucLog2[] has not been initialized yet." );
-  if( val > 0 && val < (int)sizeof(g_aucLog2) )
+  if( val > 0 && val < (int) sizeof( g_aucLog2 ) )
   {
     return g_aucLog2[val];
   }
-  return std::log2(val);
+  return std::log2( val );
 }
 #endif
 
@@ -646,5 +612,4 @@ namespace vvdec {
 
 //! \}
 
-}
-
+}   // namespace vvdec
