@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -56,7 +52,9 @@ THE POSSIBILITY OF SUCH DAMAGE.
 namespace vvdec
 {
 
-CoeffCodingContext::CoeffCodingContext( const TransformUnit& tu, ComponentID component, bool signHide )
+static const int prefix_ctx[8] = { 0, 0, 0, 3, 6, 10, 15, 21 };
+
+CoeffCodingContext::CoeffCodingContext( const TransformUnit& tu, ComponentID component, bool signHide, CtxTpl* tplBuf )
   : m_chType                    (toChannelType(component))
   , m_width                     (tu.block(component).width)
   , m_height                    (tu.block(component).height)
@@ -77,10 +75,10 @@ CoeffCodingContext::CoeffCodingContext( const TransformUnit& tu, ComponentID com
   , m_CtxSetLastY               (Ctx::LastY[m_chType])
   , m_maxLastPosX               (g_uiGroupIdx[std::min<unsigned>(JVET_C0024_ZERO_OUT_TH, m_width)  - 1])
   , m_maxLastPosY               (g_uiGroupIdx[std::min<unsigned>(JVET_C0024_ZERO_OUT_TH, m_height) - 1])
-  , m_lastOffsetX               (0)
-  , m_lastOffsetY               (0)
-  , m_lastShiftX                (0)
-  , m_lastShiftY                (0)
+  , m_lastOffsetX               (isLuma( m_chType ) ? prefix_ctx[ m_log2BlockWidth  ] : 0)
+  , m_lastOffsetY               (isLuma( m_chType ) ? prefix_ctx[ m_log2BlockHeight ] : 0)
+  , m_lastShiftX                (isChroma( m_chType ) ? Clip3( 0, 2, int( m_width  >> 3 ) ) : (m_log2BlockWidth  + 1) >> 2 )
+  , m_lastShiftY                (isChroma( m_chType ) ? Clip3( 0, 2, int( m_height >> 3 ) ) : (m_log2BlockHeight + 1) >> 2 )
   , m_scanPosLast               (-1)
   , m_subSetId                  (-1)
   , m_subSetPos                 (-1)
@@ -104,20 +102,10 @@ CoeffCodingContext::CoeffCodingContext( const TransformUnit& tu, ComponentID com
   , m_bdpcm                     (isLuma(component) ? tu.cu->bdpcmMode() : tu.cu->bdpcmModeChroma())
   , m_regBinLimit               ( ( TU::getTbAreaAfterCoefZeroOut( tu, component ) * ( isLuma( component ) ? MAX_TU_LEVEL_CTX_CODED_BIN_CONSTRAINT_LUMA : MAX_TU_LEVEL_CTX_CODED_BIN_CONSTRAINT_CHROMA ) ) >> 4 )
   , m_ts                        (tu.mtsIdx( component ) == MTS_SKIP)
+  , m_tplBuf                    (tplBuf)
 {
-  if (m_chType == CHANNEL_TYPE_CHROMA)
-  {
-    const_cast<int&>(m_lastShiftX) = Clip3( 0, 2, int( m_width  >> 3) );
-    const_cast<int&>(m_lastShiftY) = Clip3( 0, 2, int( m_height >> 3) );
-  }
-  else
-  {
-    static const int prefix_ctx[8]  = { 0, 0, 0, 3, 6, 10, 15, 21 };
-    const_cast<int&>(m_lastOffsetX) = prefix_ctx[ m_log2BlockWidth  ];
-    const_cast<int&>(m_lastOffsetY) = prefix_ctx[ m_log2BlockHeight ];
-    const_cast<int&>(m_lastShiftX)  = (m_log2BlockWidth  + 1) >> 2;
-    const_cast<int&>(m_lastShiftY)  = (m_log2BlockHeight + 1) >> 2;
-  }
+  if( !m_ts || tu.cu->slice->getTSResidualCodingDisabledFlag() )
+    memset( tplBuf, 0, m_width * m_height * sizeof( CtxTpl ) );
 }
 
 void CoeffCodingContext::initSubblock( int SubsetId, bool sigGroupFlag )
@@ -130,7 +118,6 @@ void CoeffCodingContext::initSubblock( int SubsetId, bool sigGroupFlag )
   m_maxSubPos               = m_minSubPos + ( 1 << m_log2CGSize ) - 1;
   const bool lastHorGrp     = m_subSetPosX == m_widthInGroups  - 1;
   const bool lastVerGrp     = m_subSetPosY == m_heightInGroups - 1;
-  m_checkTplBnd             = lastHorGrp;
   if( sigGroupFlag )
   {
     m_sigCoeffGroupFlag.set ( m_subSetPos );
@@ -305,18 +292,18 @@ unsigned DeriveCtx::CtxIBCFlag( const CodingUnit& cu )
 
 void MergeCtx::setMergeInfo( PredictionUnit& pu, int candIdx )
 {
-  CHECK( candIdx >= numValidMergeCand, "Merge candidate does not exist" );
+  CHECK_RECOVERABLE( candIdx >= numValidMergeCand, "Merge candidate does not exist" );
 
-  pu.setMergeFlag            ( true );
-  pu.setMmvdFlag             ( false );
+  //pu.setMergeFlag            ( true );
+  //pu.setMmvdFlag             ( false );
   pu.setInterDir             ( interDirNeighbours[candIdx] );
   pu.setImv                  ( ( !pu.geoFlag() && useAltHpelIf[candIdx] ) ? IMV_HPEL : 0 );
-  pu.setMergeIdx             ( candIdx );
+  //pu.setMergeIdx             ( candIdx );
   pu.setMergeType            ( mrgTypeNeighbours[candIdx] );
   pu.mv  [REF_PIC_LIST_0][0] = mvFieldNeighbours[(candIdx << 1) + 0].mv;
   pu.mv  [REF_PIC_LIST_1][0] = mvFieldNeighbours[(candIdx << 1) + 1].mv;
-  pu.refIdx [REF_PIC_LIST_0] = mvFieldNeighbours[( candIdx << 1 ) + 0].refIdx;
-  pu.refIdx [REF_PIC_LIST_1] = mvFieldNeighbours[( candIdx << 1 ) + 1].refIdx;
+  pu.refIdx [REF_PIC_LIST_0] = mvFieldNeighbours[( candIdx << 1 ) + 0].mfRefIdx;
+  pu.refIdx [REF_PIC_LIST_1] = mvFieldNeighbours[( candIdx << 1 ) + 1].mfRefIdx;
   pu.mvpIdx [REF_PIC_LIST_0] = NOT_VALID;
   pu.mvpIdx [REF_PIC_LIST_1] = NOT_VALID;
   pu.setBcwIdx               ( ( interDirNeighbours[candIdx] == 3 ) ? BcwIdx[candIdx] : BCW_DEFAULT );
@@ -351,8 +338,8 @@ void MergeCtx::setMmvdMergeCandiInfo( PredictionUnit& pu, int candIdx )
   {
     offset <<= 2;
   }
-  const int refList0 = mmvdBaseMv[fPosBaseIdx][0].refIdx;
-  const int refList1 = mmvdBaseMv[fPosBaseIdx][1].refIdx;
+  const int refList0 = mmvdBaseMv[fPosBaseIdx][0].mfRefIdx;
+  const int refList1 = mmvdBaseMv[fPosBaseIdx][1].mfRefIdx;
 
   if( ( refList0 != -1 ) && ( refList1 != -1 ) )
   {
@@ -452,7 +439,7 @@ void MergeCtx::setMmvdMergeCandiInfo( PredictionUnit& pu, int candIdx )
     pu.mv    [L0][0] = mmvdBaseMv[fPosBaseIdx][0].mv + tempMv[0];
     pu.refIdx[L0]    = refList0;
     pu.mv    [L1][0] = Mv(0, 0);
-    pu.refIdx[L1]    = -1;
+    pu.refIdx[L1]    = NOT_VALID;
   }
   else if( refList1 != -1 )
   {
@@ -475,16 +462,16 @@ void MergeCtx::setMmvdMergeCandiInfo( PredictionUnit& pu, int candIdx )
 
     pu.setInterDir               ( 2 );
     pu.mv    [REF_PIC_LIST_0][0] = Mv(0, 0);
-    pu.refIdx[REF_PIC_LIST_0]    = -1;
+    pu.refIdx[REF_PIC_LIST_0]    = NOT_VALID;
     pu.mv    [REF_PIC_LIST_1][0] = mmvdBaseMv[fPosBaseIdx][1].mv + tempMv[1];
     pu.refIdx[REF_PIC_LIST_1]    = refList1;
   }
 
-  pu.setMmvdFlag        ( true );
+  //pu.setMmvdFlag        ( true );
   pu.mmvdIdx            = candIdx;
-  pu.setMergeFlag       ( true );
+  //pu.setMergeFlag       ( true );
   //pu.setMergeIdx        ( candIdx );
-  pu.setMergeType       ( MRG_TYPE_DEFAULT_N );
+  //pu.setMergeType       ( MRG_TYPE_DEFAULT_N );
   pu.mvpIdx [L0]        = NOT_VALID;
   pu.mvpIdx [L1]        = NOT_VALID;
   pu.setImv             ( mmvdUseAltHpelIf[fPosBaseIdx] ? IMV_HPEL : 0 );
