@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
-License, included below. No patent rights, trademark rights and/or
-other Intellectual Property Rights other than the copyrights concerning
+The copyright in this software is being made available under the Clear BSD
+License, included below. No patent rights, trademark rights and/or 
+other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software,
-especially patent licenses, a separate Agreement needs to be closed.
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V.
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -101,49 +97,88 @@ void paddPicBorderLeftRightCore(Pel *pi, ptrdiff_t stride,int width,int xmargin,
    }
 }
 
-void Picture::create(const ChromaFormat &_chromaFormat, const Size &size, const unsigned _maxCUSize, const unsigned _margin, const int _layerId)
+void Picture::create(const ChromaFormat &_chromaFormat, const Size &size, const unsigned _maxCUSize, const unsigned _margin, const int _layerId, UserAllocator* _userAllocator )
 {
   layerId = _layerId;
   UnitArea::operator=( UnitArea( _chromaFormat, Area( Position{ 0, 0 }, size ) ) );
   margin            = _margin;
-  m_bufs[PIC_RECONSTRUCTION].create( _chromaFormat, size, _maxCUSize, _margin, MEMORY_ALIGN_DEF_SIZE );
-  m_bufs[PIC_RECON_WRAP    ].create( _chromaFormat, size, _maxCUSize, _margin, MEMORY_ALIGN_DEF_SIZE );
+  m_bufs[PIC_RECONSTRUCTION].create( _chromaFormat, size, _maxCUSize, _margin, MEMORY_ALIGN_DEF_SIZE, true, _userAllocator );
 }
 
-void Picture::resetForUse()
+void Picture::createWrapAroundBuf( const bool isWrapAround, const unsigned _maxCUSize )
 {
-  CHECK( lockedByApplication, "the picture can not be re-used, because it has not been unlocked by the application." );
+  if( isWrapAround )
+    m_bufs[PIC_RECON_WRAP].create( chromaFormat, Y().size(), _maxCUSize, margin, MEMORY_ALIGN_DEF_SIZE );
+}
+
+void Picture::resetForUse( int _layerId )
+{
+  CHECK_RECOVERABLE( lockedByApplication, "the picture can not be re-used, because it has not been unlocked by the application." );
+
+  if( cs )
+  {
+    cs->resetForUse();
+  }
 
-  setPicHead( nullptr );
   m_subPicRefBufs.clear();
 
-  subPicExtStarted = false;
-  borderExtStarted = false;
-  wrapAroundValid  = false;
-  wrapAroundOffset = 0;
-  neededForOutput  = false;
-  reconstructed    = false;
-  inProgress       = false;
-  wasLost          = false;
-  skippedDecCount  = 0;
+  m_dProcessingTime       = 0;
+  subPicExtStarted        = false;
+  borderExtStarted        = false;
+  referenced              = false;
+  progress                = Picture::init;
+  neededForOutput         = false;
+  wasLost                 = false;
+  error                   = false;
+  exceptionThrownOut      = false;
+  longTerm                = false;
+  topField                = false;
+  fieldPic                = false;
+  nonReferencePictureFlag = false;
+  skippedDecCount         = 0;
 
   picCheckedDPH = false;
   subpicsCheckedDPH.clear();
 
+  lockedByApplication = false;
+
+  poc                 = 0;
+  cts                 = 0;
+  dts                 = 0;
+  tempLayer           = std::numeric_limits<uint32_t>::max();
+  depth               = 0;
+  layerId             = _layerId;
+  eNalUnitType        = NAL_UNIT_INVALID;
+  bits                = 0;
+  rap                 = 0;
+  decodingOrderNumber = 0;
+
+  sliceSubpicIdx.clear();
+  subPictures.clear();
+
+  subLayerNonReferencePictureDueToSTSA = 0;
+
+  m_divTasksCounter     .clearException();
   m_ctuTaskCounter      .clearException();
-  m_dmvrTaskCounter     .clearException();
+  m_motionTaskCounter   .clearException();
   m_borderExtTaskCounter.clearException();
   m_copyWrapBufDone     .clearException();
-  done                  .clearException();
+  reconDone             .clearException();
   parseDone             .clearException();
+#if RECO_WHILE_PARSE
   std::for_each( ctuParsedBarrier.begin(), ctuParsedBarrier.end(), []( auto& b ) { b.clearException(); } );
+#endif
 
-  done.lock();
+  clearSliceBuffer();
+
+  SEI_internal::deleteSEIs( seiMessageList );
+
+  reconDone.lock();
 }
 
 void Picture::destroy()
 {
-  CHECK( lockedByApplication, "the picture can not be destroyed, because it has not been unlocked by the application." );
+  CHECK_RECOVERABLE( lockedByApplication, "the picture can not be destroyed, because it has not been unlocked by the application." );
 
   for (uint32_t t = 0; t < NUM_PIC_TYPES; t++)
   {
@@ -161,23 +196,22 @@ void Picture::destroy()
   ctuParsedBarrier.clear();
 #endif
 
-  for( auto &ps : slices )
-  {
-    delete ps;
-  }
-  slices.clear();
+  clearSliceBuffer();
 
   SEI_internal::deleteSEIs( seiMessageList );
 
   subpicsCheckedDPH.clear();
 
+  m_divTasksCounter     .clearException();
   m_ctuTaskCounter      .clearException();
-  m_dmvrTaskCounter     .clearException();
+  m_motionTaskCounter   .clearException();
   m_borderExtTaskCounter.clearException();
   m_copyWrapBufDone     .clearException();
-  done                  .clearException();
+  reconDone             .clearException();
   parseDone             .clearException();
+#if RECO_WHILE_PARSE
   std::for_each( ctuParsedBarrier.begin(), ctuParsedBarrier.end(), []( auto& b ) { b.clearException(); } );
+#endif
 }
 
        PelBuf     Picture::getRecoBuf(const ComponentID compID, bool wrap)       { return getBuf(compID, wrap ? PIC_RECON_WRAP : PIC_RECONSTRUCTION); }
@@ -189,7 +223,7 @@ const CPelUnitBuf Picture::getRecoBuf(const UnitArea &unit, bool wrap)     const
        PelUnitBuf Picture::getRecoBuf( bool wrap )                               { return wrap ? m_bufs[PIC_RECON_WRAP] : m_bufs[PIC_RECONSTRUCTION]; }
 const CPelUnitBuf Picture::getRecoBuf( bool wrap )                         const { return wrap ? m_bufs[PIC_RECON_WRAP] : m_bufs[PIC_RECONSTRUCTION]; }
 
-void Picture::finalInit( const SPS *sps, const PPS *pps, PicHeader* picHeader, APS* alfApss[ALF_CTB_MAX_NUM_APS], APS* lmcsAps, APS* scalingListAps, bool phPSupdate )
+void Picture::finalInit( CUChunkCache* cuChunkCache, TUChunkCache* tuChunkCache, const SPS *sps, const PPS *pps, const std::shared_ptr<PicHeader>& ph, const APS* const alfApss[ALF_CTB_MAX_NUM_APS], const APS* lmcsAps, const APS* scalingListAps, bool phPSupdate )
 {
   SEI_internal::deleteSEIs( seiMessageList );
   clearSliceBuffer();
@@ -200,7 +234,7 @@ void Picture::finalInit( const SPS *sps, const PPS *pps, PicHeader* picHeader, A
 
   if( !cs )
   {
-    cs = new CodingStructure( g_globalUnitCache.getCuCache(), g_globalUnitCache.getTuCache() );
+    cs = new CodingStructure( cuChunkCache, tuChunkCache );
     cs->create( chromaFormatIDC, Area( 0, 0, iWidth, iHeight ) );
   }
 
@@ -212,18 +246,20 @@ void Picture::finalInit( const SPS *sps, const PPS *pps, PicHeader* picHeader, A
 #endif
 
   parseDone   . lock();
-  cs->picture = this;
-  cs->pps     = pps ? pps->getSharedPtr() : nullptr;
-  cs->sps     = sps ? sps->getSharedPtr() : nullptr;
+  cs->picture   = this;
+  cs->picHeader = ph;
+  cs->pps       = pps ? pps->getSharedPtr() : nullptr;
+  cs->sps       = sps ? sps->getSharedPtr() : nullptr;
 
   if( phPSupdate )
   {
-    picHeader->setSPSId         ( sps->getSPSId() );
-    picHeader->setPPSId         ( pps->getPPSId() );
-    picHeader->setLmcsAPS       ( lmcsAps        ? lmcsAps       ->getSharedPtr() : nullptr );
-    picHeader->setScalingListAPS( scalingListAps ? scalingListAps->getSharedPtr() : nullptr );
+    ph->setSPSId         ( sps->getSPSId() );
+    ph->setPPSId         ( pps->getPPSId() );
+    ph->setLmcsAPS       ( lmcsAps        ? lmcsAps       ->getSharedPtr() : nullptr );
+    ph->setScalingListAPS( scalingListAps ? scalingListAps->getSharedPtr() : nullptr );
   }
-  
+  nonReferencePictureFlag = ph->getNonReferencePictureFlag();
+
   for( int i = 0; i < ALF_CTB_MAX_NUM_APS; ++i )
   {
     cs->alfApss[i] = alfApss[i] ? alfApss[i]->getSharedPtr() : nullptr;
@@ -233,8 +269,8 @@ void Picture::finalInit( const SPS *sps, const PPS *pps, PicHeader* picHeader, A
     cs->lmcsAps = lmcsAps ? lmcsAps->getSharedPtr() : nullptr;
   }
 
-  cs->pcv     = pps->pcv.get();
-
+  cs->pcv = pps->pcv.get();
+  cs->allocTempInternals();
   cs->rebindPicBufs();
 
   resetProcessingTime();
@@ -279,31 +315,70 @@ Slice* Picture::allocateNewSlice( Slice** pilot )
   slice->setPPS( cs->pps.get() );
   slice->setSPS( cs->sps.get() );
   slice->setVPS( cs->vps.get() );
-  slice->setAlfAPSs( cs->alfApss );
+  slice->setAlfApss( cs->alfApss );
   slice->setPic( this );
 
   return slice;
 }
 
-void Picture::setPicHead( const std::shared_ptr<PicHeader>& ph )
-{
-  this->picHeader = ph;
-  if( cs )
-  {
-    cs->picHeader = ph.get();
-  }
-}
-
 void Picture::clearSliceBuffer()
 {
-  for( size_t i = 0; i < slices.size(); i++ )
+  for( auto &s: slices )
   {
-    delete slices[i];
+    delete s;
   }
-
   slices.clear();
 }
 
+bool Picture::lastSliceOfPicPresent() const
+{
+  if( slices.empty() )
+  {
+    return false;
+  }
+  const Slice*   lastSlice      = slices.back();
+  const unsigned lastCtuInSlice = lastSlice->getCtuAddrInSlice( lastSlice->getNumCtuInSlice() - 1 );
+  return lastCtuInSlice == lastSlice->getPPS()->pcv->sizeInCtus - 1;
+}
+
+void Picture::waitForAllTasks()
+{
+  m_motionTaskCounter.wait_nothrow();
+  m_ctuTaskCounter.wait_nothrow();
+  m_borderExtTaskCounter.wait_nothrow();
+  m_divTasksCounter.wait_nothrow();   // this waits for the slice parsing and the finishPic Task
+}
+
+void Picture::ensureUsableAsRef()
+{
+#if RECO_WHILE_PARSE
+  ctuParsedBarrier.clear();
+#endif
+  neededForOutput = false;
+
+  // set referenced to true, because we don't know if it has been set correctly, but that way it will be available as a reference pic
+  referenced = true;
+
+  // ensure cs->m_colMiMap is set to zero
+  cs->initStructData();
+
+  CHECK( m_motionTaskCounter.hasException(), "to be usable as reference the picture should not have an Exception on the dmvr task counter" );
+  CHECK( reconDone.hasException(), "to be usable as reference the picture should not have an Exception reconDone barrier" );
+}
+
+void Picture::fillGrey( const SPS* sps )
+{
+  // fill in grey buffer for missing reference pictures (GDR or broken bitstream)
+  const uint32_t yFill = 1 << ( sps->getBitDepth( CHANNEL_TYPE_LUMA ) - 1 );
+  const uint32_t cFill = 1 << ( sps->getBitDepth( CHANNEL_TYPE_CHROMA ) - 1 );
+  getRecoBuf().Y().fill( yFill );
+  getRecoBuf().Cb().fill( cFill );
+  getRecoBuf().Cr().fill( cFill );
+
+  progress = Picture::reconstructed;
+  reconDone.unlock();
+}
+
 void Picture::extendPicBorder( bool top, bool bottom, bool leftrightT, bool leftrightB, ChannelType chType )
 {
   if( cs->pps->getUseWrapAround() )
@@ -518,13 +593,61 @@ PelBuf Picture::getOriginBuf( const PictureType &type, const ComponentID compID
   return m_bufs[type].getOriginBuf( compID );
 }
 
+Size Picture::getBufSize( const PictureType &type, const ComponentID compID ) const
+{
+  return m_bufs[type].getBufSize( compID );
+}
+
+void* Picture::getBufAllocator( const ComponentID compID )
+{
+  return m_bufs[PIC_RECONSTRUCTION].getBufAllocator( compID );
+}
+
+bool Picture::isExternAllocator() const
+{
+  return m_bufs[PIC_RECONSTRUCTION].isExternAllocator();
+}
+
+const UserAllocator* Picture::getUserAllocator() const
+{
+  return m_bufs[PIC_RECONSTRUCTION].getUserAllocator();
+}
+
+std::vector<Picture*> Picture::buildAllRefPicsVec()
+{
+  std::vector<Picture*> refPics;
+  for( const Slice* slice : this->slices )
+  {
+    if( slice->isIntra() )
+    {
+      continue;
+    }
+
+    for( int iDir = REF_PIC_LIST_0; iDir < NUM_REF_PIC_LIST_01; ++iDir )
+    {
+      for( int iRefIdx = 0; iRefIdx < slice->getNumRefIdx( ( RefPicList ) iDir ); iRefIdx++ )
+      {
+        Picture* refPic = slice->getNoConstRefPic( ( RefPicList ) iDir, iRefIdx );
+        if( std::find( refPics.cbegin(), refPics.cend(), refPic ) == refPics.cend() )
+        {
+          refPics.push_back( refPic );
+        }
+      }
+    }
+  }
+
+  return refPics;
+}
+
 void Picture::startProcessingTimer()
 {
+  std::lock_guard<std::mutex> lock( m_timerMutex );
   m_processingStartTime = std::chrono::steady_clock::now();
 }
 
 void Picture::stopProcessingTimer()
 {
+  std::lock_guard<std::mutex> lock( m_timerMutex );
   auto endTime = std::chrono::steady_clock::now();
   m_dProcessingTime += std::chrono::duration<double>(endTime - m_processingStartTime).count();
 }
