@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -90,7 +86,7 @@ QpParam::QpParam( const TransformUnit& tu, const ComponentID& compIDX, const boo
     int
     chromaQpOffset  = pps.getQpOffset                    ( jCbCr );
     chromaQpOffset += tu.cu->slice->getSliceChromaQpDelta( jCbCr );
-    chromaQpOffset += pps.getChromaQpOffsetListEntry( tu.cu->chromaQpAdj ).u.offset[int( jCbCr ) - 1];
+    chromaQpOffset += pps.getChromaQpOffsetListEntry( tu.cu->chromaQpAdj ).get( jCbCr );
 
     int qpi = Clip3( -qpBdOffset, MAX_QP, qpy );
     baseQp  = sps.getMappedChromaQpValue( jCbCr, qpi );
@@ -172,7 +168,7 @@ static void DeQuantCore( const int              maxX,
         if( level )
         {
           const TCoeff           clipQCoef = TCoeff(Clip3<Intermediate_Int>(inputMinimum, inputMaximum, level));
-          const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale) << leftShift;
+          const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale) *(1<<leftShift);
 
           piCoef[n] = TCoeff(Clip3<Intermediate_Int>(transformMinimum,transformMaximum,iCoeffQ));
         }
@@ -227,7 +223,7 @@ static void DeQuantPCMCore( const int     maxX,
         if( level )
         {
           const TCoeff           clipQCoef = TCoeff(Clip3<Intermediate_Int>(inputMinimum, inputMaximum, level));
-          const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale) << leftShift;
+          const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale) *(1<< leftShift);
 
           piCoef[n] = TCoeff(Clip3<Intermediate_Int>(transformMinimum,transformMaximum,iCoeffQ));
         }
@@ -237,15 +233,16 @@ static void DeQuantPCMCore( const int     maxX,
   }
 }
 
-Quant::Quant()
+Quant::Quant( const Quant* other ) : m_dequantCoefBuf( nullptr ), m_ownDequantCoeff( false )
 {
-  xInitScalingList( nullptr );
-  DeQuant= DeQuantCore;
+  xInitScalingList( other );
+
+  DeQuant    = DeQuantCore;
   DeQuantPCM = DeQuantPCMCore;
-#if   ENABLE_SIMD_OPT_QUANT && defined( TARGET_SIMD_X86 )
+#if ENABLE_SIMD_OPT_QUANT && defined( TARGET_SIMD_X86 )
+
   initQuantX86();
 #endif
-
 }
 
 Quant::~Quant()
@@ -298,6 +295,16 @@ void invResDPCM( const TransformUnit &tu, const ComponentID &compID, CoeffBuf &d
   }
 }
 
+static inline int getTransformShift( const int channelBitDepth, const Size size, const int maxLog2TrDynamicRange )
+{
+  return maxLog2TrDynamicRange - channelBitDepth - ( ( getLog2( size.width ) + getLog2( size.height ) ) >> 1 );
+}
+
+static inline int getScalingListType( const PredMode predMode, const ComponentID compID )
+{
+  return ( ( predMode == MODE_INTRA ) ? 0 : MAX_NUM_COMPONENT ) + MAP_CHROMA( compID );
+}
+
 void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const ComponentID& compID, const QpParam& cQP )
 {
   const SPS*             sps                   = tu.cu->sps;
@@ -333,7 +340,7 @@ void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const Componen
     maxY = tu.maxScanPosY[compID];
   }
   
-  CHECK(scalingListType >= SCALING_LIST_NUM, "Invalid scaling list");
+  CHECK_RECOVERABLE(scalingListType >= SCALING_LIST_NUM, "Invalid scaling list");
 
   // Represents scaling through forward transform
   const bool bClipTransformShiftTo0 = false;// tu.mtsIdx[compID] != 1 && sps->getSpsRangeExtension().getExtendedPrecisionProcessingFlag();
@@ -377,7 +384,8 @@ void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const Componen
     }
     else
     {
-      int* piDequantCoef = getDequantCoeff( scalingListType, QP_rem, uiLog2TrWidth, uiLog2TrHeight );
+      int* piDequantCoef = getDequantCoeff( scalingListType, uiLog2TrWidth, uiLog2TrHeight );
+      int  scaleQP       = g_InvQuantScales[needSqrtAdjustment ? 1 : 0][QP_rem];
 
       if( ( tu.cu->bdpcmMode() && isLuma(compID) ) || ( tu.cu->bdpcmModeChroma() && isChroma(compID) ) )
       {
@@ -394,7 +402,7 @@ void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const Componen
 
               if( level )
               {
-                scale = piDequantCoef[n];
+                scale = piDequantCoef[n] * scaleQP;
 
                 const TCoeff           clipQCoef = TCoeff(Clip3<Intermediate_Int>(inputMinimum, inputMaximum, level));
                 const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale + iAdd) >> rightShift;
@@ -416,7 +424,7 @@ void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const Componen
 
               if( level )
               {
-                scale = piDequantCoef[n];
+                scale = piDequantCoef[n] * scaleQP;
 
                 const TCoeff           clipQCoef = TCoeff(Clip3<Intermediate_Int>(inputMinimum, inputMaximum, level));
                 const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale) << leftShift;
@@ -442,7 +450,7 @@ void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const Componen
 
               if( level )
               {
-                scale = piDequantCoef[n];
+                scale = piDequantCoef[n] * scaleQP;
 
                 const TCoeff           clipQCoef = TCoeff(Clip3<Intermediate_Int>(inputMinimum, inputMaximum, level));
                 Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale + iAdd) >> rightShift;
@@ -464,10 +472,10 @@ void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const Componen
 
               if( level )
               {
-                scale = piDequantCoef[n];
+                scale = piDequantCoef[n] * scaleQP;
 
                 const TCoeff           clipQCoef = TCoeff(Clip3<Intermediate_Int>(inputMinimum, inputMaximum, level));
-                const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale) << leftShift;
+                const Intermediate_Int iCoeffQ   = (Intermediate_Int(clipQCoef) * scale) *(1<< leftShift);
 
                 piCoef[n] = TCoeff(Clip3<Intermediate_Int>(transformMinimum,transformMaximum,iCoeffQ));
               }
@@ -484,13 +492,11 @@ void Quant::dequant( const TransformUnit& tu, CoeffBuf& dstCoeff, const Componen
  * \param scalingList quantized matrix address
  * \param format      chroma format
  */
-void Quant::setScalingListDec( ScalingList& scalingList )
+void Quant::setScalingListDec( const ScalingList& scalingList )
 {
-  const int minimumQp = 0;
-  const int maximumQp = SCALING_LIST_REM_NUM;
-
   int scalingListId    = 0;
   int recScalingListId = 0;
+  bool anyChange = false;
   for( uint32_t size = SCALING_LIST_FIRST_CODED; size <= SCALING_LIST_LAST_CODED; size++ )
   {
     for( uint32_t list = 0; list < SCALING_LIST_NUM; list++ )
@@ -500,12 +506,10 @@ void Quant::setScalingListDec( ScalingList& scalingList )
         continue;
       }
       scalingListId = g_scalingListId[size][list];
-      for( int qp = minimumQp; qp < maximumQp; qp++ )
-      {
-        xSetScalingListDec( scalingList, list, size, qp, scalingListId );
-      }
+      anyChange |= xSetScalingListDec( scalingList, list, size, scalingListId );
     }
   }
+  if( !anyChange ) return;
   // based on square result and apply downsample technology
   for( uint32_t sizew = 0; sizew <= SCALING_LIST_LAST_CODED; sizew++ )   // 7
   {
@@ -518,12 +522,9 @@ void Quant::setScalingListDec( ScalingList& scalingList )
       for( uint32_t list = 0; list < SCALING_LIST_NUM; list++ )   // 9
       {
         int largerSide = ( sizew > sizeh ) ? sizew : sizeh;
-        CHECK( largerSide < SCALING_LIST_4x4, "Rectangle Error!" );
+        CHECK_RECOVERABLE( largerSide < SCALING_LIST_4x4, "Rectangle Error!" );
         recScalingListId = g_scalingListId[largerSide][list];
-        for( int qp = minimumQp; qp < maximumQp; qp++ )
-        {
-          xSetRecScalingListDec( scalingList, list, sizew, sizeh, qp, recScalingListId );
-        }
+        xSetRecScalingListDec( scalingList, list, sizew, sizeh, recScalingListId );
       }
     }
   }
@@ -536,22 +537,18 @@ void Quant::setScalingListDec( ScalingList& scalingList )
  * \param qp Quantization parameter
  * \param format chroma format
  */
-void Quant::xSetScalingListDec(const ScalingList &scalingList, uint32_t listId, uint32_t sizeId, int qp, uint32_t scalingListId)
+bool Quant::xSetScalingListDec(const ScalingList &scalingList, uint32_t listId, uint32_t sizeId, uint32_t scalingListId)
 {
-  uint32_t width  = g_vvcScalingListSizeX[sizeId];
-  uint32_t height = g_vvcScalingListSizeX[sizeId];
-  uint32_t ratio  = g_vvcScalingListSizeX[sizeId]/std::min(MAX_MATRIX_SIZE_NUM,(int)g_vvcScalingListSizeX[sizeId]);
-  int *dequantcoeff;
+  const uint32_t width  = g_vvcScalingListSizeX[sizeId];
+  const uint32_t height = g_vvcScalingListSizeX[sizeId];
+  const uint32_t ratio  = g_vvcScalingListSizeX[sizeId]/std::min(MAX_MATRIX_SIZE_NUM,(int)g_vvcScalingListSizeX[sizeId]);
+
   const int *coeff = scalingList.getScalingListAddress(scalingListId);
+  int *dequantcoeff = getDequantCoeff(listId, sizeId, sizeId);
 
-  dequantcoeff = getDequantCoeff(listId, qp, sizeId, sizeId);
-
-  const int blockIsNotPowerOf4 = ( getLog2( width ) + getLog2( height ) ) & 1;
-  int invQuantScale = g_InvQuantScales[blockIsNotPowerOf4][qp];
-
+  return
   processScalingListDec(coeff,
                         dequantcoeff,
-                        invQuantScale,
                         height, width, ratio,
                         std::min(MAX_MATRIX_SIZE_NUM, (int)g_vvcScalingListSizeX[sizeId]),
                         scalingList.getScalingListDC(scalingListId));
@@ -564,21 +561,19 @@ void Quant::xSetScalingListDec(const ScalingList &scalingList, uint32_t listId,
 * \param qp Quantization parameter
 * \param format chroma format
 */
-void Quant::xSetRecScalingListDec(const ScalingList &scalingList, uint32_t listId, uint32_t sizeIdw, uint32_t sizeIdh, int qp, uint32_t scalingListId)
+void Quant::xSetRecScalingListDec(const ScalingList &scalingList, uint32_t listId, uint32_t sizeIdw, uint32_t sizeIdh, uint32_t scalingListId)
 {
   if (sizeIdw == sizeIdh) return;
-  uint32_t width = g_vvcScalingListSizeX[sizeIdw];
-  uint32_t height = g_vvcScalingListSizeX[sizeIdh];
-  uint32_t largeSideId = (sizeIdw > sizeIdh) ? sizeIdw : sizeIdh;  //16
+
+  const uint32_t width  = g_vvcScalingListSizeX[sizeIdw];
+  const uint32_t height = g_vvcScalingListSizeX[sizeIdh];
+  const uint32_t largeSideId = (sizeIdw > sizeIdh) ? sizeIdw : sizeIdh;  //16
 
   const int *coeff = scalingList.getScalingListAddress(scalingListId);
-  int *dequantcoeff;
-  dequantcoeff = getDequantCoeff(listId, qp, sizeIdw, sizeIdh);
-  const int blockIsNotPowerOf4 = ( getLog2( width ) + getLog2( height ) ) & 1;
-  int invQuantScale = g_InvQuantScales[blockIsNotPowerOf4][qp];
+  int *dequantcoeff = getDequantCoeff(listId, sizeIdw, sizeIdh);
+
   processScalingListDec(coeff,
                         dequantcoeff,
-                        invQuantScale,
                         height, width, (largeSideId>3) ? 2 : 1,
                         (largeSideId >= 3 ? 8 : 4),
                         scalingList.getScalingListDC(scalingListId));
@@ -594,80 +589,139 @@ void Quant::xSetRecScalingListDec(const ScalingList &scalingList, uint32_t listI
  * \param sizuNum matrix size
  * \param dc dc parameter
  */
-void Quant::processScalingListDec( const int *coeff, int *dequantcoeff, int invQuantScales, uint32_t height, uint32_t width, uint32_t ratio, int sizuNum, uint32_t dc)
+bool Quant::processScalingListDec( const int *coeff, int *dequantcoeff, uint32_t height, uint32_t width, uint32_t ratio, int sizuNum, uint32_t dc)
 {
   if (height != width)
   {
-    for (uint32_t j = 0; j<height; j++)
+    const int hl2 = getLog2( height );
+    const int wl2 = getLog2( width );
+    const int sl2 = getLog2( sizuNum );
+
+    const int loopH = std::min<int>( height, JVET_C0024_ZERO_OUT_TH );
+    const int loopW = std::min<int>( width,  JVET_C0024_ZERO_OUT_TH );
+    
+    const int ratioWH = height > width   ? hl2 - wl2 : wl2 - hl2;
+    const int ratioH  = height / sizuNum ? hl2 - sl2 : sl2 - hl2;
+    const int ratioW  = width / sizuNum  ? wl2 - sl2 : sl2 - wl2;
+
+    if( height > width )
     {
-      for (uint32_t i = 0; i<width; i++)
+      for( uint32_t j = 0; j < loopH; j += ( 1 << ratioH ) )
       {
-        if (i >= JVET_C0024_ZERO_OUT_TH || j >= JVET_C0024_ZERO_OUT_TH)
+        for( uint32_t i = 0; i < loopW; i++ )
         {
-          dequantcoeff[j*width + i] = 0;
-          continue;
+          dequantcoeff[j * width + i] = coeff[sizuNum * ( j >> ratioH ) + ( ( i << ratioWH ) >> ratioH )];
         }
-        int ratioWH = height > width   ? height / width   : width   / height;
-        int ratioH  = height / sizuNum ? height / sizuNum : sizuNum / height;
-        int ratioW  = width / sizuNum  ? width  / sizuNum : sizuNum / width;
-        //sizeNum = 8/4
-        if (height > width)
+
+        const int* src = &dequantcoeff[j * width];
+        for( int jj = 1; jj < ( 1 << ratioH ); jj++ )
         {
-          dequantcoeff[j*width + i] = invQuantScales * coeff[sizuNum * (j / ratioH) + ((i * ratioWH) / ratioH)];
-        }
-        else //ratioH < ratioW
-        {
-          dequantcoeff[j*width + i] = invQuantScales * coeff[sizuNum * ((j * ratioWH) / ratioW) + (i / ratioW)];
-        }
-        int largeOne = std::max( width, height );
-        if( largeOne > 8 )
-        {
-          dequantcoeff[0] = invQuantScales * dc;
+          memcpy( &dequantcoeff[( j + jj ) * width], src, loopW * sizeof( int ) );
         }
       }
     }
-    return;
-  }
-  for(uint32_t j=0;j<height;j++)
-  {
-    for(uint32_t i=0;i<width;i++)
+    else
     {
-      dequantcoeff[j*width + i] = invQuantScales * coeff[sizuNum * (j / ratio) + i / ratio];
+      for( uint32_t j = 0; j < loopH; j++ )
+      {
+        for( uint32_t i = 0; i < loopW; i += ( 1 << ratioW ) )
+        {
+          const int coeffi = coeff[sizuNum * ( ( j << ratioWH ) >> ratioW ) + ( i >> ratioW )];
+          for( uint32_t ii = 0; ii < ( 1 << ratioW ); ii++ )
+          {
+            dequantcoeff[j * width + i + ii] = coeffi;
+          }
+        }
+      }
+    }
+
+    const int largeOne = std::max( width, height );
+    if( largeOne > 8 )
+    {
+      dequantcoeff[0] = dc;
+    }
+    return true;
+  }
+
+  bool anyChange = false;
+
+  const int rl2   = getLog2( ratio );
+  const int loopH = std::min<int>( height, JVET_C0024_ZERO_OUT_TH );
+  const int loopW = std::min<int>( width, JVET_C0024_ZERO_OUT_TH );
+
+  for( uint32_t j = 0; j < loopH; j += ( 1 << rl2 ) )
+  {
+    for( uint32_t i = 0; i < loopW; i += ( 1 << rl2 ) )
+    {
+      const int coeffi = coeff[sizuNum * ( j >> rl2 ) + ( i >> rl2 )];
+      anyChange |= coeffi != dequantcoeff[j * width + i];
+      for( uint32_t ii = 0; anyChange && ii < ( 1 << rl2 ); ii++ )
+      {
+        dequantcoeff[j * width + i + ii] = coeffi;
+      }
+    }
+
+    const int* src = &dequantcoeff[j * width];
+    for( int jj = 1; jj < ( 1 << rl2 ); jj++ )
+    {
+      memcpy( &dequantcoeff[( j + jj ) * width], src, loopW * sizeof( int ) );
     }
   }
 
-  if(ratio > 1)
+  if( ratio > 1 )
   {
-    dequantcoeff[0] = invQuantScales * dc;
+    anyChange |= dequantcoeff[0] != dc;
+    dequantcoeff[0] = dc;
   }
+
+  return anyChange;
 }
 
+static constexpr int g_numScalingListCoeffs = 96774;
+
 /** initialization process of scaling list array
  */
 void Quant::xInitScalingList( const Quant* other )
 {
+  if( other )
+  {
+    m_dequantCoefBuf  = other->m_dequantCoefBuf;
+    m_ownDequantCoeff = false;
+  }
+  else
+  {
+    m_dequantCoefBuf  = new int[g_numScalingListCoeffs];
+    m_ownDequantCoeff = true;
+  }
+
   size_t numQuants = 0;
 
   for(uint32_t sizeIdX = 0; sizeIdX < SCALING_LIST_SIZE_NUM; sizeIdX++)
   {
     for(uint32_t sizeIdY = 0; sizeIdY < SCALING_LIST_SIZE_NUM; sizeIdY++)
     {
-      for(uint32_t qp = 0; qp < SCALING_LIST_REM_NUM; qp++)
+      for(uint32_t listId = 0; listId < SCALING_LIST_NUM; listId++)
       {
-        for(uint32_t listId = 0; listId < SCALING_LIST_NUM; listId++)
-        {
-          m_dequantCoef [sizeIdX][sizeIdY][listId][qp] = &m_dequantCoefBuf[numQuants];
-          numQuants += g_vvcScalingListSizeX[sizeIdX] * g_vvcScalingListSizeX[sizeIdY];
-        } // listID loop
-      }
+        m_dequantCoef [sizeIdX][sizeIdY][listId] = &m_dequantCoefBuf[numQuants];
+        numQuants += g_vvcScalingListSizeX[sizeIdX] * g_vvcScalingListSizeX[sizeIdY];
+      } // listID loop
     }
   }
+
+  CHECK_RECOVERABLE( numQuants != g_numScalingListCoeffs, "Incorrect size of scaling list entries number!" );
 }
 
 /** destroy quantization matrix array
  */
 void Quant::xDestroyScalingList()
 {
+  if( m_ownDequantCoeff )
+  {
+    delete[] m_dequantCoefBuf;
+  }
+
+  m_ownDequantCoeff = false;
+  m_dequantCoefBuf  = nullptr;
 }
 
 void Quant::init( const Picture *pic )
@@ -687,11 +741,11 @@ void Quant::init( const Picture *pic )
 
   if( slice && slice->getExplicitScalingListUsed() )
   {
-    std::shared_ptr<APS> scalingListAPS = slice->getPicHeader()->getScalingListAPS();
+    const std::shared_ptr<const APS> scalingListAPS = slice->getPicHeader()->getScalingListAPS();
     if( slice->getNalUnitLayerId() != scalingListAPS->getLayerId() )
     {
-      CHECK( scalingListAPS->getLayerId() > slice->getNalUnitLayerId(), "Layer Id of APS cannot be greater than layer Id of VCL NAL unit the refer to it" );
-      CHECK( slice->getSPS()->getVPSId() == 0, "VPSId of the referred SPS cannot be 0 when layer Id of APS and layer Id of current slice are different" );
+      CHECK_RECOVERABLE( scalingListAPS->getLayerId() > slice->getNalUnitLayerId(), "Layer Id of APS cannot be greater than layer Id of VCL NAL unit the refer to it" );
+      CHECK_RECOVERABLE( slice->getSPS()->getVPSId() == 0, "VPSId of the referred SPS cannot be 0 when layer Id of APS and layer Id of current slice are different" );
       for( int i = 0; i < slice->getVPS()->getNumOutputLayerSets(); i++ )
       {
         bool isCurrLayerInOls = false;
@@ -707,11 +761,15 @@ void Quant::init( const Picture *pic )
             isRefLayerInOls = true;
           }
         }
-        CHECK( isCurrLayerInOls && !isRefLayerInOls, "When VCL NAl unit in layer A refers to APS in layer B, all OLS that contains layer A shall also contains layer B" );
+        CHECK_RECOVERABLE( isCurrLayerInOls && !isRefLayerInOls, "When VCL NAl unit in layer A refers to APS in layer B, all OLS that contains layer A shall also contains layer B" );
       }
     }
-    ScalingList scalingList = scalingListAPS->getScalingList();
-    setScalingListDec(scalingList);
+    const ScalingList& scalingList = scalingListAPS->getScalingList();
+    if( m_ownDequantCoeff )
+    {
+      memset( m_dequantCoefBuf, 0, sizeof( int ) * g_numScalingListCoeffs );
+      setScalingListDec( scalingList );
+    }
     setUseScalingList(true);
   }
   else
