@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -348,24 +344,14 @@ SampleAdaptiveOffset::~SampleAdaptiveOffset()
   destroy();
 }
 
-void SampleAdaptiveOffset::create( int picWidth, int picHeight, ChromaFormat format, uint32_t maxCUWidth, uint32_t maxCUHeight, uint32_t maxCUDepth, uint32_t lumaBitShift, uint32_t chromaBitShift )
+void SampleAdaptiveOffset::create( int picWidth, int picHeight, ChromaFormat format, uint32_t maxCUWidth, uint32_t maxCUHeight, uint32_t maxCUDepth, uint32_t lumaBitShift, uint32_t chromaBitShift, PelUnitBuf& unitBuf )
 {
   offsetBlock = offsetBlock_core;
 #if ENABLE_SIMD_OPT_SAO && defined( TARGET_SIMD_X86 )
   initSampleAdaptiveOffsetX86();
 #endif
 
-  //temporary picture buffer
-  UnitArea picArea( format, Area( 0, 0, picWidth, picHeight ) );
-
-  if( !m_tempBuf.bufs.empty() && ( picWidth != m_tempBuf.Y().width || picHeight != m_tempBuf.Y().height ) )
-  {
-    m_tempBuf.destroy();
-  }
-  if( m_tempBuf.bufs.empty() )
-  {
-    m_tempBuf.create( picArea );
-  }
+  m_tempBuf = unitBuf;
 
   //bit-depth related
   for( int compIdx = 0; compIdx < MAX_NUM_COMPONENT; compIdx++ )
@@ -377,7 +363,6 @@ void SampleAdaptiveOffset::create( int picWidth, int picHeight, ChromaFormat for
 
 void SampleAdaptiveOffset::destroy()
 {
-  m_tempBuf.destroy();
 }
 
 #if 0
@@ -413,12 +398,13 @@ void SampleAdaptiveOffset::SAOPrepareCTULine( CodingStructure &cs, const UnitAre
   PelUnitBuf           rec = cs.getRecoBuf();
 
   const int height = lineArea.lumaSize().height;
-  const int y = lineArea.lumaPos().y;
+  const int y      = lineArea.lumaPos().y;
 
-  const int cpyY = y == 0 ? 0 : y + 2;
-  const int cpyHeight = std::min<int>( y == 0 ? height + 2 : height, pcv.lumaHeight - cpyY );
-
-  bool anySaoBlk = false;
+  const int cpyY       = y == 0 ? 0 : y + 2;
+        int cpyHeight  = height;
+            cpyHeight -= y == 0 ? 0 : 2;
+            cpyHeight -= cpyY + cpyHeight >= pcv.lumaHeight ? 0 : 2;
+            cpyHeight  = std::min<int>( cpyHeight, pcv.lumaHeight - cpyY );
 
   for( int x = 0; x < pcv.lumaWidth; x += pcv.maxCUWidth )
   {
@@ -428,34 +414,60 @@ void SampleAdaptiveOffset::SAOPrepareCTULine( CodingStructure &cs, const UnitAre
     getMergeList( cs, ctuRsAddr, mergeList );
 
     reconstructBlkSAOParam( cs.m_ctuData[ctuRsAddr].saoParam, mergeList );
+  }
 
-    for( int i = 0; i < MAX_NUM_COMPONENT; i++ )
+  const int numComp = getNumberValidComponents( pcv.chrFormat );
+
+  for( int i = 0; i < numComp; i++ )
+  {
+    const ComponentID compID = ComponentID( i );
+
+    const int currCpyY =   cpyY            >> getComponentScaleY( compID, pcv.chrFormat );
+    const int currCpyH =   cpyHeight       >> getComponentScaleY( compID, pcv.chrFormat );
+    const int currCtuW =   pcv.maxCUWidth  >> getComponentScaleX( compID, pcv.chrFormat );
+    const int currPicW =   pcv.lumaWidth   >> getComponentScaleX( compID, pcv.chrFormat );
+    
+    const ptrdiff_t src_stride =       rec.bufs[ compID ].stride;
+    const ptrdiff_t dst_stride = m_tempBuf.bufs[ compID ].stride;
+
+    const Pel* src =       rec.bufs[ compID ].bufAt( 0, currCpyY );
+          Pel* dst = m_tempBuf.bufs[ compID ].bufAt( 0, currCpyY );
+
+    for( int j = currCpyY; j < ( currCpyY + currCpyH ); j++, src += src_stride, dst += dst_stride )
     {
-      if( cs.m_ctuData[ctuRsAddr].saoParam[i].modeIdc != SAO_MODE_OFF )
+      int numCpy   = 0;
+      int cpyStart = 0;
+
+      int ctuRsAddr = getCtuAddr( Position( 0, y ), *cs.pcv );
+
+      for( int x = 0, currCpyX = 0; x <= pcv.widthInCtus; x++, currCpyX += currCtuW, ctuRsAddr++ )
       {
-        anySaoBlk = true;
+        if( x < pcv.widthInCtus && cs.m_ctuData[ctuRsAddr].saoParam[i].modeIdc != SAO_MODE_OFF )
+        {
+          if( !numCpy ) cpyStart = currCpyX - !!x;
+          numCpy++;
+        }
+        else if( numCpy )
+        {
+          // copy
+          int cpyLen = numCpy * currCtuW + !!cpyStart + 1;
+          if( cpyStart + cpyLen > currPicW ) cpyLen = currPicW - cpyStart;
+          memcpy( dst + cpyStart, src + cpyStart, cpyLen * sizeof( Pel ) );
+          numCpy = 0;
+        }
       }
     }
   }
 
-  if( !anySaoBlk )
+  const int cpyY2      = cpyY + cpyHeight;
+  const int cpyHeight2 = 4;
+  
+  if( cpyY2 < pcv.lumaHeight )
   {
-    const int cpyY2 = cpyY + cpyHeight - 4;
-    const int cpyHeight2 = 4;
-
-    if( cpyY2 < pcv.lumaHeight )
-    {
-      const UnitArea cpyLine2( pcv.chrFormat, Area( 0, cpyY2, pcv.lumaWidth, cpyHeight2 ) );
-
-      m_tempBuf.subBuf( cpyLine2 ).copyFrom( rec.subBuf( cpyLine2 ) );
-    }
-
-    return;
+    const UnitArea cpyLine2( pcv.chrFormat, Area( 0, cpyY2, pcv.lumaWidth, cpyHeight2 ) );
+  
+    m_tempBuf.subBuf( cpyLine2 ).copyFrom( rec.subBuf( cpyLine2 ) );
   }
-
-  const UnitArea cpyLine( pcv.chrFormat, Area( 0, cpyY, pcv.lumaWidth, cpyHeight ) );
-
-  m_tempBuf.subBuf( cpyLine ).copyFrom( rec.subBuf( cpyLine ) );
 }
 
 void SampleAdaptiveOffset::SAOProcessCTULine( CodingStructure &cs, const UnitArea &lineArea )
@@ -546,7 +558,7 @@ void SampleAdaptiveOffset::invertQuantOffsets(ComponentID compIdx, int typeIdc,
     {
       dstOffsets[i] = codedOffset[i] *(1<<m_offsetStepLog2[compIdx]);
     }
-    CHECK(dstOffsets[SAO_CLASS_EO_PLAIN] != 0, "EO offset is not '0'"); //keep EO plain offset as zero
+    CHECK_RECOVERABLE(dstOffsets[SAO_CLASS_EO_PLAIN] != 0, "EO offset is not '0'"); //keep EO plain offset as zero
   }
 
 }
@@ -631,7 +643,7 @@ void SampleAdaptiveOffset::reconstructBlkSAOParam(SAOBlkParam& recParam, SAOBlkP
     case SAO_MODE_MERGE:
       {
         const SAOBlkParam* mergeTarget = mergeList[offsetParam.typeIdc];
-        CHECK(mergeTarget == NULL, "Merge target does not exist");
+        CHECK_RECOVERABLE(mergeTarget == NULL, "Merge target does not exist");
 
         offsetParam = (*mergeTarget)[component];
       }
@@ -673,14 +685,14 @@ void SampleAdaptiveOffset::offsetCTU( const UnitArea& area, const CPelUnitBuf& s
   int  verVirBndryPos[]     = { -1, -1, -1 };
   int  horVirBndryPosComp[] = { -1, -1, -1 };
   int  verVirBndryPosComp[] = { -1, -1, -1 };
-  bool isCtuCrossedByVirtualBoundaries = isCrossedByVirtualBoundaries( cs.picHeader,
-                                                                             area.Y(),
-                                                                             numHorVirBndry, numVerVirBndry,
-                                                                             horVirBndryPos, verVirBndryPos );
+  bool isCtuCrossedByVirtualBoundaries = isCrossedByVirtualBoundaries( cs.picHeader.get(),
+                                                                       area.Y(),
+                                                                       numHorVirBndry, numVerVirBndry,
+                                                                       horVirBndryPos, verVirBndryPos );
   if( isCtuCrossedByVirtualBoundaries )
   {
-    CHECK( numHorVirBndry >= (int)( sizeof(horVirBndryPos) / sizeof(horVirBndryPos[0]) ), "Too many virtual boundaries" );
-    CHECK( numHorVirBndry >= (int)( sizeof(verVirBndryPos) / sizeof(verVirBndryPos[0]) ), "Too many virtual boundaries" );
+    CHECK_RECOVERABLE( numHorVirBndry >= (int)( sizeof(horVirBndryPos) / sizeof(horVirBndryPos[0]) ), "Too many virtual boundaries" );
+    CHECK_RECOVERABLE( numHorVirBndry >= (int)( sizeof(verVirBndryPos) / sizeof(verVirBndryPos[0]) ), "Too many virtual boundaries" );
   }
 
   for(int compIdx = 0; compIdx < numberOfComponents; compIdx++)
