@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -54,7 +50,9 @@ THE POSSIBILITY OF SUCH DAMAGE.
 #include "UnitTools.h"
 #include "ContextModelling.h"
 #include "CodingStructure.h"
-#include "CommonLib/TimeProfiler.h"
+#include "TimeProfiler.h"
+#include "Quant.h"
+#include "InterPrediction.h"
 
 
 #include "dtrace_buffer.h"
@@ -78,6 +76,33 @@ InvTrans *fastInvTrans[NUM_TRANS_TYPE][g_numTransformMatrixSizes] =
 //! \ingroup CommonLib
 //! \{
 
+static void invLfnstNxNCore( int* src, int* dst, const uint32_t mode, const uint32_t index, const uint32_t size, int zeroOutSize )
+{
+  int             maxLog2TrDynamicRange =  15;
+  const TCoeff    outputMinimum         = -( 1 << maxLog2TrDynamicRange );
+  const TCoeff    outputMaximum         =  ( 1 << maxLog2TrDynamicRange ) - 1;
+  const int8_t*   trMat                 =  ( size > 4 ) ? g_lfnst8x8[ mode ][ index ][ 0 ] : g_lfnst4x4[ mode ][ index ][ 0 ];
+  const int       trSize                =  ( size > 4 ) ? 48 : 16;
+  int             resi;
+  int*            out                   =  dst;
+
+  CHECK_RECOVERABLE( index > 2, "wrong" );
+
+  for( int j = 0; j < trSize; j++, trMat += 16 )
+  {
+    resi = 0;
+    const int8_t* trMatTmp = trMat;
+    int*          srcPtr   = src;
+
+    for( int i = 0; i < zeroOutSize; i++ )
+    {
+      resi += *srcPtr++ * *trMatTmp++;
+    }
+
+    *out++ = Clip3( outputMinimum, outputMaximum, ( int ) ( resi + 64 ) >> 7 );
+  }
+}
+
 static inline int64_t square( const int d ) { return d * (int64_t)d; }
 
 template<int signedMode> void invTransformCbCr( PelBuf &resCb, PelBuf &resCr )
@@ -101,7 +126,8 @@ template<int signedMode> void invTransformCbCr( PelBuf &resCb, PelBuf &resCr )
 // ====================================================================================================================
 // TrQuant class member functions
 // ====================================================================================================================
-TrQuant::TrQuant()
+
+TrQuant::TrQuant( class InterPrediction* ip, const TrQuant* other ) : Quant( other )
 {
   // allocate temporary buffers
   m_invICT      = m_invICTMem + maxAbsIctMode;
@@ -113,34 +139,25 @@ TrQuant::TrQuant()
   m_invICT[ 3]  = invTransformCbCr< 3>;
   m_invICT[-3]  = invTransformCbCr<-3>;
 
-  m_tmp  = ( TCoeff* ) xMalloc( TCoeff, MAX_TU_SIZE_FOR_PROFILE * MAX_TU_SIZE_FOR_PROFILE );
-  m_blk  = ( TCoeff* ) xMalloc( TCoeff, MAX_TU_SIZE_FOR_PROFILE * MAX_TU_SIZE_FOR_PROFILE );
-  m_dqnt = ( TCoeff* ) xMalloc( TCoeff, MAX_TU_SIZE_FOR_PROFILE * MAX_TU_SIZE_FOR_PROFILE );
+  m_invLfnstNxN = invLfnstNxNCore;
+
+  static_assert( sizeof( ip->m_acYuvPred[0] ) > sizeof( TCoeff ) * ( MAX_TU_SIZE_FOR_PROFILE * MAX_TU_SIZE_FOR_PROFILE + MEMORY_ALIGN_DEF_SIZE ), "Buffer to small to be reused!" );
+  static_assert( sizeof( ip->m_acYuvPred[1] ) > sizeof( TCoeff ) * ( MAX_TU_SIZE_FOR_PROFILE * MAX_TU_SIZE_FOR_PROFILE + MEMORY_ALIGN_DEF_SIZE ), "Buffer to small to be reused!" );
+  static_assert( sizeof( ip->m_acYuvPred[2] ) > sizeof( TCoeff ) * ( MAX_TU_SIZE_FOR_PROFILE * MAX_TU_SIZE_FOR_PROFILE + MEMORY_ALIGN_DEF_SIZE ), "Buffer to small to be reused!" );
+
+  char* tmp  = ( char* ) ip->m_acYuvPred[0];
+  char* blk  = ( char* ) ip->m_acYuvPred[1];
+  char* dqnt = ( char* ) ip->m_acYuvPred[2];
+
+  m_tmp  = ( TCoeff* ) ( ( ptrdiff_t ) tmp  + ( MEMORY_ALIGN_DEF_SIZE - ( ( ptrdiff_t ) tmp  & ( MEMORY_ALIGN_DEF_SIZE - 1 ) ) ) );
+  m_blk  = ( TCoeff* ) ( ( ptrdiff_t ) blk  + ( MEMORY_ALIGN_DEF_SIZE - ( ( ptrdiff_t ) blk  & ( MEMORY_ALIGN_DEF_SIZE - 1 ) ) ) );
+  m_dqnt = ( TCoeff* ) ( ( ptrdiff_t ) dqnt + ( MEMORY_ALIGN_DEF_SIZE - ( ( ptrdiff_t ) dqnt & ( MEMORY_ALIGN_DEF_SIZE - 1 ) ) ) );
+
+#if defined( TARGET_SIMD_X86 ) && ENABLE_SIMD_TCOEFF_OPS
+  initTrQuantX86();
+#endif
 }
 
-TrQuant::~TrQuant()
-{
-  // delete temporary buffers
-  if( m_tmp )
-  {
-    xFree( m_tmp );
-    m_tmp = nullptr;
-  }
-
-  if( m_blk )
-  {
-    xFree( m_blk );
-    m_blk = nullptr;
-  }
-
-  if( m_dqnt )
-  {
-    xFree( m_dqnt );
-    m_dqnt = nullptr;
-  }
-}
-
-
 void TrQuant::xDeQuant(const TransformUnit &tu,
                              CoeffBuf      &dstCoeff,
                        const ComponentID   &compID,
@@ -155,33 +172,6 @@ void TrQuant::init( const Picture *pic )
   Quant::init( pic );
 }
 
-void TrQuant::invLfnstNxN( int* src, int* dst, const uint32_t mode, const uint32_t index, const uint32_t size, int zeroOutSize )
-{
-  int             maxLog2TrDynamicRange =  15;
-  const TCoeff    outputMinimum         = -( 1 << maxLog2TrDynamicRange );
-  const TCoeff    outputMaximum         =  ( 1 << maxLog2TrDynamicRange ) - 1;
-  const int8_t*   trMat                 =  ( size > 4 ) ? g_lfnst8x8[ mode ][ index ][ 0 ] : g_lfnst4x4[ mode ][ index ][ 0 ];
-  const int       trSize                =  ( size > 4 ) ? 48 : 16;
-  int             resi;
-  int*            out                   =  dst;
-
-  CHECK( index > 2, "wrong" );
-
-  for( int j = 0; j < trSize; j++, trMat++ )
-  {
-    resi = 0;
-    const int8_t* trMatTmp = trMat;
-    int*          srcPtr   = src;
-
-    for( int i = 0; i < zeroOutSize; i++, trMatTmp += trSize )
-    {
-      resi += *srcPtr++ * *trMatTmp;
-    }
-
-    *out++ = Clip3( outputMinimum, outputMaximum, ( int ) ( resi + 64 ) >> 7 );
-  }
-}
-
 uint32_t TrQuant::getLFNSTIntraMode( int wideAngPredMode )
 {
   uint32_t intraMode;
@@ -249,7 +239,7 @@ void TrQuant::xInvLfnst( TransformUnit &tu, const ComponentID& compID )
       *dst++ = coeffTemp[ scan[y] ];
     }
 
-    invLfnstNxN( m_tempInMatrix, m_tempOutMatrix, g_lfnstLut[ intraMode ], lfnstIdx - 1, sbSize, ( tu4x4Flag || tu8x8Flag ) ? 8 : 16 );
+    m_invLfnstNxN( m_tempInMatrix, m_tempOutMatrix, g_lfnstLut[ intraMode ], lfnstIdx - 1, sbSize, ( tu4x4Flag || tu8x8Flag ) ? 8 : 16 );
 
     lfnstTemp = m_tempOutMatrix; // inverse spectral rearrangement
 
@@ -379,7 +369,7 @@ void TrQuant::getTrTypes( const TransformUnit& tu, const ComponentID compID, int
 
     if( sbtIdx == SBT_VER_HALF || sbtIdx == SBT_VER_QUAD )
     {
-      CHECK( lwidth > MTS_INTER_MAX_CU_SIZE, "wrong" );
+      CHECK_RECOVERABLE( lwidth > MTS_INTER_MAX_CU_SIZE, "wrong" );
       if( lheight > MTS_INTER_MAX_CU_SIZE )
       {
         trTypeHor = trTypeVer = DCT2;
@@ -392,7 +382,7 @@ void TrQuant::getTrTypes( const TransformUnit& tu, const ComponentID compID, int
     }
     else
     {
-      CHECK( lheight > MTS_INTER_MAX_CU_SIZE, "wrong" );
+      CHECK_RECOVERABLE( lheight > MTS_INTER_MAX_CU_SIZE, "wrong" );
       if( lwidth > MTS_INTER_MAX_CU_SIZE )
       {
         trTypeHor = trTypeVer = DCT2;
@@ -434,63 +424,64 @@ void TrQuant::xIT( const TransformUnit &tu, const ComponentID &compID, const CCo
 
   getTrTypes( tu, compID, trTypeHor, trTypeVer );
 
+  if( tu.maxScanPosX[compID] == 0 && tu.maxScanPosY[compID] == 0 && trTypeHor == DCT2 && trTypeVer == DCT2 )
+  {
+    int dcVal = 0;
+
+    if( width > 1 && height > 1 )
+    {
+      const int shift_1st = TRANSFORM_MATRIX_SHIFT + 1 + COM16_C806_TRANS_PREC;
+      const int shift_2nd = ( TRANSFORM_MATRIX_SHIFT + maxLog2TrDynamicRange - 1 ) - bitDepth + COM16_C806_TRANS_PREC;
+
+      dcVal = ( ( pCoeff.buf[0] << TRANSFORM_MATRIX_SHIFT ) + ( 1 << ( shift_1st - 1 ) ) ) >> shift_1st;
+      dcVal = ( ( dcVal << TRANSFORM_MATRIX_SHIFT ) + ( 1 << ( shift_2nd - 1 ) ) ) >> shift_2nd;
+    }
+    else
+    {
+      const int shift = ( TRANSFORM_MATRIX_SHIFT + maxLog2TrDynamicRange ) - bitDepth + COM16_C806_TRANS_PREC;
+      dcVal = ( ( pCoeff.buf[0] << TRANSFORM_MATRIX_SHIFT ) + ( 1 << ( shift - 1 ) ) ) >> shift;
+    }
+
+    pResidual.fill( dcVal );
+    return;
+  }
+
   const int skipWidth  = std::max<int>( ( trTypeHor != DCT2 && width  == 32 ) ? 16 : width  > JVET_C0024_ZERO_OUT_TH ? width  - JVET_C0024_ZERO_OUT_TH : 0, width  - tu.maxScanPosX[compID] - 1 );
   const int skipHeight = std::max<int>( ( trTypeVer != DCT2 && height == 32 ) ? 16 : height > JVET_C0024_ZERO_OUT_TH ? height - JVET_C0024_ZERO_OUT_TH : 0, height - tu.maxScanPosY[compID] - 1 );
 
   TCoeff *block = m_blk;
+  int shiftlast;
 
   if( width > 1 && height > 1 ) //2-D transform
   {
     const int      shift_1st              =   TRANSFORM_MATRIX_SHIFT + 1 + COM16_C806_TRANS_PREC; // 1 has been added to shift_1st at the expense of shift_2nd
     const int      shift_2nd              = ( TRANSFORM_MATRIX_SHIFT + maxLog2TrDynamicRange - 1 ) - bitDepth + COM16_C806_TRANS_PREC;
-    CHECK( shift_1st < 0, "Negative shift" );
-    CHECK( shift_2nd < 0, "Negative shift" );
+    CHECK_RECOVERABLE( shift_1st < 0, "Negative shift" );
+    CHECK_RECOVERABLE( shift_2nd < 0, "Negative shift" );
     TCoeff *tmp   = m_tmp;
-    fastInvTrans[trTypeVer][transformHeightIndex](pCoeff.buf, tmp, shift_1st, width, skipWidth, skipHeight, clipMinimum, clipMaximum);
-    fastInvTrans[trTypeHor][transformWidthIndex] (tmp,      block, shift_2nd, height,         0, skipWidth, clipMinimum, clipMaximum);
+    fastInvTrans[trTypeVer][transformHeightIndex](pCoeff.buf, tmp, shift_1st, width, skipWidth, skipHeight, true,  clipMinimum, clipMaximum);
+    fastInvTrans[trTypeHor][transformWidthIndex] (tmp,      block, shift_2nd, height,         0, skipWidth, false, clipMinimum, clipMaximum);
+    shiftlast = shift_2nd;
   }
   else if( width == 1 ) //1-D vertical transform
   {
     int shift = ( TRANSFORM_MATRIX_SHIFT + maxLog2TrDynamicRange - 1 ) - bitDepth + COM16_C806_TRANS_PREC;
-    CHECK( shift < 0, "Negative shift" );
-    CHECK( ( transformHeightIndex < 0 ), "There is a problem with the height." );
-    fastInvTrans[trTypeVer][transformHeightIndex]( pCoeff.buf, block, shift + 1, 1, 0, skipHeight, clipMinimum, clipMaximum );
+    CHECK_RECOVERABLE( shift < 0, "Negative shift" );
+    CHECK_RECOVERABLE( ( transformHeightIndex < 0 ), "There is a problem with the height." );
+    fastInvTrans[trTypeVer][transformHeightIndex]( pCoeff.buf, block, shift + 1, 1, 0, skipHeight, false, clipMinimum, clipMaximum );
+    shiftlast = shift + 1;
   }
   else //if(iHeight == 1) //1-D horizontal transform
   {
     const int      shift              = ( TRANSFORM_MATRIX_SHIFT + maxLog2TrDynamicRange - 1 ) - bitDepth + COM16_C806_TRANS_PREC;
-    CHECK( shift < 0, "Negative shift" );
-    CHECK( ( transformWidthIndex < 0 ), "There is a problem with the width." );
-    fastInvTrans[trTypeHor][transformWidthIndex]( pCoeff.buf, block, shift + 1, 1, 0, skipWidth, clipMinimum, clipMaximum );
+    CHECK_RECOVERABLE( shift < 0, "Negative shift" );
+    CHECK_RECOVERABLE( ( transformWidthIndex < 0 ), "There is a problem with the width." );
+    fastInvTrans[trTypeHor][transformWidthIndex]( pCoeff.buf, block, shift + 1, 1, 0, skipWidth, false, clipMinimum, clipMaximum );
+    shiftlast = shift + 1;
   }
 
-#if ENABLE_SIMD_TCOEFF_OPS
-  if( width & 3 )
-#endif //ENABLE_SIMD_TCOEFF_OPS
-  {
-    Pel       *dst    = pResidual.buf;
-    ptrdiff_t  stride = pResidual.stride;
-
-    for( int y = 0; y < height; y++ )
-    {
-      for( int x = 0; x < width; x++ )
-      {
-        dst[x] = ( Pel ) *block++;
-      }
-
-      dst += stride;
-    }
-  }
-#if ENABLE_SIMD_TCOEFF_OPS
-  else if( width & 7 )
-  {
-    g_tCoeffOps.cpyResi4( block, pResidual.buf, pResidual.stride, width, height );
-  }
-  else
-  {
-    g_tCoeffOps.cpyResi8( block, pResidual.buf, pResidual.stride, width, height );
-  }
-#endif //ENABLE_SIMD_TCOEFF_OPS
+  int round = 1 << ( shiftlast - 1 );
+  g_tCoeffOps.cpyResiClip[getLog2( width )]( block, pResidual.buf, pResidual.stride, width, height, clipMinimum, clipMaximum, round, shiftlast );
 }
 
 /** Wrapper function between HM interface and core NxN transform skipping
