@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -61,6 +57,7 @@ THE POSSIBILITY OF SUCH DAMAGE.
 #include <cstring>
 #include <assert.h>
 #include <cassert>
+#include <mutex>
 
 namespace vvdec
 {
@@ -72,6 +69,9 @@ namespace vvdec
 
 #define RECO_WHILE_PARSE                                  1
 #define ALLOW_MIDER_LF_DURING_PICEXT                      1
+#define MAX_THREADS                                       64 // maximum number of threads
+
+#define MAX_OUT_OF_ORDER_PICS                             3 // maximum number of pictures, that are reconstructed out of order
 
 #define JVET_O1170_CHECK_BV_AT_DECODER                    0 // For decoder to check if a BV is valid or not
 
@@ -87,6 +87,8 @@ namespace vvdec
 
 #define TBC                                               0
 
+#define ALF_PRE_TRANSPOSE                                 1
+
 #define JVET_R0270                                        TBC // JVET-S0270: Treating picture with mixed RASL and RADL slices as RASL picture
 #define JVET_S0155_EOS_NALU_CHECK                         TBC // JVET-S0155: Constraints on EOS NAL units
 
@@ -152,13 +154,17 @@ namespace vvdec
 #if ENABLE_SIMD_OPT_BUFFER
 #define ENABLE_SIMD_OPT_GBI                               1                                                 ///< SIMD optimization for GBi
 #endif
-#define ENABLE_SIMD_OPT_BIO                             ( 1 && ENABLE_SIMD_OPT )                            ///< SIMD optimization for BIO
+#define ENABLE_SIMD_OPT_INTER                           ( 1 && ENABLE_SIMD_OPT )                            ///< SIMD optimization for BIO
 #define ENABLE_SIMD_OPT_PICTURE                         ( 1 && ENABLE_SIMD_OPT )                            ///< SIMD optimization for Picture Padding
 #define ENABLE_SIMD_OPT_SAO                             ( 1 && ENABLE_SIMD_OPT )                            ///< SIMD optimization for BIO
 #define ENABLE_SIMD_TCOEFF_OPS                          ( 1 && ENABLE_SIMD_OPT )
 #define ENABLE_SIMD_LOG2                                ( 1 && ENABLE_SIMD_OPT )
 #define ENABLE_SIMD_DBLF                                ( 1 && ENABLE_SIMD_OPT )
 
+#if defined( TARGET_SIMD_X86 ) && !defined( REAL_TARGET_X86 )
+#  define SIMD_EVERYWHERE_EXTENSION_LEVEL                 SSE42
+#endif
+
 // End of SIMD optimizations
 
 #define LUMA_ADAPTIVE_DEBLOCKING_FILTER_QP_OFFSET         1 /// JVET-L0414 (CE11.2.2) with explicit signalling of num interval, threshold and qpOffset
@@ -204,7 +210,7 @@ typedef       int16_t         TFilterCoeff;      ///< filter coefficient
 typedef       int             Intermediate_Int;  ///< used as intermediate value in calculations
 #endif
 
-typedef       uint64_t        Distortion;        ///< distortion measurement
+typedef       uint32_t        Distortion;        ///< distortion measurement
 
 typedef       uint16_t        SplitSeries;       ///< used to encoded the splits that caused a particular CU size
 
@@ -349,16 +355,10 @@ enum RefPicList : uint8_t
 /// distortion function index
 enum DFunc
 {
-  DF_SAD             = 8,             ///< general size SAD
-  DF_SAD2            = DF_SAD+1,      ///<   2xM SAD
-  DF_SAD4            = DF_SAD+2,      ///<   4xM SAD
-  DF_SAD8            = DF_SAD+3,      ///<   8xM SAD
-  DF_SAD16           = DF_SAD+4,      ///<  16xM SAD
-  DF_SAD32           = DF_SAD+5,      ///<  32xM SAD
-  DF_SAD64           = DF_SAD+6,      ///<  64xM SAD
-  DF_SAD16N          = DF_SAD+7,      ///< 16NxM SAD
+  DF_SAD8            = 0,      ///<   8xM SAD
+  DF_SAD16           = 1,      ///<  16xM SAD
 
-  DF_TOTAL_FUNCTIONS = DF_SAD16N + 1
+  DF_TOTAL_FUNCTIONS = DF_SAD16 + 1
 };
 
 /// motion vector predictor direction used in AMVP
@@ -508,7 +508,8 @@ enum NalUnitType
   NAL_UNIT_CODED_SLICE_GDR,         // 10
 
   NAL_UNIT_RESERVED_IRAP_VCL_11,
-  NAL_UNIT_RESERVED_IRAP_VCL_12,
+
+  NAL_UNIT_OPI,                     // 12
   NAL_UNIT_DCI,                     // 13
   NAL_UNIT_VPS,                     // 14
   NAL_UNIT_SPS,                     // 15
@@ -735,6 +736,17 @@ enum MsgLevel
   DETAILS = 6
 };
 
+enum ErrHandlingFlags
+{
+  // Keep in sync with vvdecErrHandlingFlags from vvdec.h
+  ERR_HANDLING_OFF          = 0,   // no special internal error-handling besides tuning in to GDR streams
+  ERR_HANDLING_TRY_CONTINUE = 1,   // try to continue decoding after parsing errors or missing pictures
+#if 0
+  // NOT YET IMPLEMENTED
+  ERR_HANDLING_COPY_CLOSEST = 2    // replace missing reference pictures with the closest available picture (otherwise grey frame)
+#endif
+};
+
 // ---------------------------------------------------------------------------
 // exception class
 // ---------------------------------------------------------------------------
@@ -748,32 +760,59 @@ enum MsgLevel
 class Exception : public std::exception
 {
 public:
-  Exception( const std::string& _s ) : m_str( _s ) { }
-  Exception( const Exception& _e ) : std::exception( _e ), m_str( _e.m_str ) { }
-  virtual ~Exception() noexcept { };
-  virtual const char* what() const noexcept { return m_str.c_str(); }
-  Exception& operator=( const Exception& _e ) { std::exception::operator=( _e ); m_str = _e.m_str; return *this; }
-  template<typename T> Exception& operator<<( T t ) { std::ostringstream oss; oss << t; m_str += oss.str(); return *this; }
+  explicit Exception( const std::string& _s ) : m_str( _s ) {}
+  virtual ~Exception() noexcept = default;
+
+  virtual const char* what() const noexcept                 { return m_str.c_str(); }
+  template<typename T> Exception& operator<<( const T& t )  { std::ostringstream oss; oss << t; m_str += oss.str(); return *this; }
 private:
   std::string m_str;
 };
 
-// if a check fails with THROW or CHECK, please check if ported correctly from assert in revision 1196)
-#define THROW(x)            throw( Exception( "\nERROR: In function \"" ) << __FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x )
-//#define THROW(x)            { std::cerr << "\nERROR: In function \"" << __FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x << std::endl; abort(); }
-#define CHECK(c,x)          if(c){ THROW( x << "\nERROR CONDITION: " << #c ); }
-#define CHECK_WARN(c,x)     if(c){ std::cerr << "\nWARNING: In function \"" << __FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x << "\nERROR CONDITION: " << #c << std::endl; }
-#define EXIT(x)             throw( Exception( "\n" ) << x << "\n" )
-#define CHECK_NULLPTR(_ptr) CHECK( !( _ptr ), "Accessing an empty pointer!" )
+class RecoverableException : public Exception
+{
+public:
+  explicit RecoverableException( const std::string& _s ) : Exception( _s ) {}
+  virtual ~RecoverableException() noexcept = default;
 
-#if !NDEBUG  // for non MSVC compiler, define _DEBUG if in debug mode to have same behavior between MSVC and others in debug
-#ifndef _DEBUG
-#define _DEBUG 1
+  RecoverableException& operator=( const RecoverableException& _e )        { Exception::operator=( _e ); return *this; }
+  template<typename T> RecoverableException& operator<<( const T& t )      { Exception::operator<<( t ); return *this; }
+};
+
+class UnsupportedFeatureException : public Exception
+{
+public:
+  explicit UnsupportedFeatureException( const std::string& _s ) : Exception( _s ) {}
+  UnsupportedFeatureException( const UnsupportedFeatureException& _e )                   = default;
+  virtual ~UnsupportedFeatureException() noexcept                                 = default;
+  UnsupportedFeatureException& operator=( const UnsupportedFeatureException& _e )        { Exception::operator=( _e ); return *this; }
+  template<typename T> UnsupportedFeatureException& operator<<( const T& t )      { Exception::operator<<( t ); return *this; }
+};
+
+#if !defined( __PRETTY_FUNCTION__ ) && !defined( __GNUC__ )
+# define __PRETTY_FUNCTION__ __FUNCSIG__
 #endif
+
+#if !NDEBUG   // for non MSVC compiler, define _DEBUG if in debug mode to have same behavior between MSVC and others in debug
+#  ifndef _DEBUG
+#    define _DEBUG 1
+#  endif
 #endif
 
+// if a check fails with THROW or CHECK, please check if ported correctly from assert in revision 1196)
+#define THROW(x)               throw( Exception( "\nERROR: In function \"" ) << __PRETTY_FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x )
+#define ABORT(x)               { std::cerr << "\nERROR: In function \"" << __FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x << std::endl; abort(); }
+#define THROW_RECOVERABLE(x)   throw( RecoverableException( "\nERROR: In function \"" ) << __PRETTY_FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x )
+#define THROW_UNSUPPORTED(x)   throw( UnsupportedFeatureException( "\nERROR: In function \"" ) << __PRETTY_FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x )
+#define CHECK(c,x)             if(c){ THROW( x << "\nERROR CONDITION: " << #c ); }
+#define CHECK_RECOVERABLE(c,x) if(c){ THROW_RECOVERABLE( x << "\nERROR CONDITION: " << #c ); }
+#define CHECK_WARN(c,x)        if(c){ std::cerr << "\nWARNING: In function \"" << __PRETTY_FUNCTION__ << "\" in " << __FILE__ << ":" << __LINE__ << ": " << x << "\nERROR CONDITION: " << #c << std::endl; }
+#define CHECK_UNSUPPORTED(c,x) if(c){ THROW_UNSUPPORTED( x << "\nERROR CONDITION: " << #c ); }
+#define EXIT(x)                throw( Exception( "\n" ) << x << "\n" )
+#define CHECK_NULLPTR(_ptr)    CHECK( !( _ptr ), "Accessing an empty pointer!" )
+
 #if defined( _DEBUG )
-#define CHECKD(c,x)         if(c){ THROW(x); }
+#define CHECKD(c,x)            if(c){ ABORT( x << "\nERROR CONDITION: " << #c ); }
 #else
 #define CHECKD(c,x)
 #endif   // _DEBUG
@@ -788,7 +827,7 @@ template<typename T, size_t N>
 class static_vector
 {
   T _arr[ N ];
-  size_t _size;
+  size_t _size = 0;
 
 public:
 
@@ -804,11 +843,20 @@ public:
 
   static const size_type max_num_elements = N;
 
-  static_vector() : _size( 0 )                                 { }
+  static_vector()                                        = default;
+  static_vector( const static_vector<T, N>& )            = default;
+  static_vector( static_vector<T, N>&& )                 = default;
+  static_vector& operator=( const static_vector<T, N>& ) = default;
+  static_vector& operator=( static_vector<T, N>&& )      = default;
+
   static_vector( size_t N_ ) : _size( N_ )                     { CHECKD( _size > N, "capacity exceeded" ); }
   static_vector( size_t N_, const T& _val ) : _size( 0 )       { resize( N_, _val ); }
   template<typename It>
   static_vector( It _it1, It _it2 ) : _size( 0 )               { while( _it1 < _it2 ) _arr[ _size++ ] = *_it1++; }
+  template<typename Iterable,
+           typename IS_ITERABLE = decltype( std::cbegin( std::declval<Iterable>() ), std::cend( std::declval<Iterable>() ) )>
+  explicit static_vector( const Iterable& iterable ) : _size( 0 ) { for( auto& e: iterable ) { push_back( e ); } }
+
   static_vector( std::initializer_list<T> _il ) : _size( 0 )
   {
     typename std::initializer_list<T>::iterator _src1 = _il.begin();
@@ -864,7 +912,7 @@ public:
   size_type       max_size() const              { return N; }
   size_type       byte_capacity() const         { return sizeof(_arr); }
 
-  void            erase( const_iterator _pos )  { iterator it   = const_cast<iterator>( _pos ) - 1;
+  void            erase( const_iterator _pos )  { iterator it   = begin() + ( _pos - 1 - begin() );
                                                   iterator last = end() - 1;
                                                   while( ++it != last ) *it = *( it + 1 );
                                                   _size--; }
@@ -911,27 +959,47 @@ struct AlfSliceParam
   bool             lumaCoeffSummed = false;
   bool             lumaFinalDone   = false;
   bool             chrmFinalDone   = false;
+#if ALF_PRE_TRANSPOSE
   short            lumaCoeffFinal     [MAX_NUM_ALF_TRANSPOSE_ID * MAX_NUM_ALF_CLASSES * MAX_NUM_ALF_LUMA_COEFF];
   short            lumaClippFinal     [MAX_NUM_ALF_TRANSPOSE_ID * MAX_NUM_ALF_CLASSES * MAX_NUM_ALF_LUMA_COEFF];
+#else
+  short            lumaCoeffFinal     [MAX_NUM_ALF_CLASSES * MAX_NUM_ALF_LUMA_COEFF];
+  short            lumaClippFinal     [MAX_NUM_ALF_CLASSES * MAX_NUM_ALF_LUMA_COEFF];
+#endif
   short            chrmClippFinal     [MAX_NUM_ALF_ALTERNATIVES_CHROMA * MAX_NUM_ALF_CHROMA_COEFF];
   int              tLayer;
   bool             newFilterFlag      [MAX_NUM_CHANNEL_TYPE];
-  
+
+  mutable std::mutex recostructMutex;   // this must be the last member, so we can clear the rest of the struct using memset()
+
   AlfSliceParam()
   {
-    memset( this, 0, sizeof( *this ) );
+    static_assert( offsetof( AlfSliceParam, recostructMutex ) + sizeof( recostructMutex ) == sizeof( AlfSliceParam ), "recostructMutex must be last member" );
+    static_assert( std::is_standard_layout<AlfSliceParam>::value, "AlfSliceParam must be standard layout type for offsetof to work" );
+
+    GCC_WARNING_DISABLE_class_memaccess
+      memset( this, 0, offsetof( AlfSliceParam, recostructMutex ) );
+    GCC_WARNING_RESET
+
+    // GH: why are numLumaFilters and numAlternativesChroma not set to 1 as in reset()?
   }
 
   void reset()
   {
-    std::memset( this, 0, sizeof( AlfSliceParam ) );
+    GCC_WARNING_DISABLE_class_memaccess
+      memset( this, 0, offsetof( AlfSliceParam, recostructMutex ) );
+    GCC_WARNING_RESET
+
     numLumaFilters = 1;
     numAlternativesChroma = 1;
   }
 
   const AlfSliceParam& operator=( const AlfSliceParam& src )
   {
-    std::memcpy( this, &src, sizeof( AlfSliceParam ) );
+    GCC_WARNING_DISABLE_class_memaccess
+      std::memcpy( this, &src, offsetof( AlfSliceParam, recostructMutex ) );
+    GCC_WARNING_RESET
+
     return *this;
   }
 };
