@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -56,6 +52,8 @@ THE POSSIBILITY OF SUCH DAMAGE.
 #include "MotionInfo.h"
 #include "ChromaFormat.h"
 
+#include <mutex>
+
 namespace vvdec
 {
 
@@ -326,8 +324,8 @@ struct CodingUnit : public UnitArea
         CodingUnit *next;
   const CodingUnit *above;
   const CodingUnit *left;
-  Mv               *mvdL0SubPu; // 7 ptr (8 byte)
-  Pel              *predBuf;
+  ptrdiff_t         mvdL0SubPuOff; // 7 ptr (8 byte)
+  ptrdiff_t         predBufOff;
   uint32_t          idx;
   uint32_t          tileIdx;
   
@@ -344,6 +342,11 @@ struct CodingUnit : public UnitArea
   
   int8_t            chromaQpAdj;
   int8_t            qp;
+  uint8_t           _interDir       : 2;
+  uint8_t           _imv            : 2;                                  
+  uint8_t           _bcw            : 3;
+  uint8_t           _mergeType      : 2;
+  bool              _dmvrCond;      // avoid data races in a separate bit
 
   uint8_t           _sbtInfo;
   uint8_t           qtDepth         : 3;
@@ -363,29 +366,24 @@ struct CodingUnit : public UnitArea
   uint8_t           _lfnstIdx       : 2;
                                     
   uint8_t           _ispIdx         : 2;                                 
-  bool              _dmvrCond       : 1;
   bool              _ciipFlag       : 1;
   bool              _mergeFlag      : 1;
   bool              _mmvdFlag       : 1;
   bool              _affineFlag     : 1;
   bool              _geoFlag        : 1;
                                     
-  uint8_t           _mergeType      : 2;
   uint8_t           _mrgIdx         : 3;
   uint8_t           _geoMrgIdx0     : 3;
                                     
   uint8_t           _geoMrgIdx1     : 3;
   uint8_t           _affineType     : 2;
-  uint8_t           _interDir       : 2;
   bool              _mipFlag        : 1;
-                                    
-  uint8_t           _bcw            : 3;
+  
   uint8_t           _multiRefIdx    : 2;
   bool              planeCbfY       : 1;
   bool              planeCbfU       : 1;
   bool              planeCbfV       : 1;
   
-  uint8_t           _imv            : 2;
   uint8_t           _smvd           : 2;
 
   uint8_t            _geoDir0, _geoDir1;
@@ -561,102 +559,111 @@ namespace vvdec {
 static constexpr size_t DYN_CACHE_CHUNK_SIZE = 1024;
 
 template<typename T>
-class dynamic_cache
+class thread_safe_chunk_cache
 {
-  std::vector<T*> m_cache;
   std::vector<T*> m_cacheChunks;
+  std::mutex      m_mutex;
 
 public:
 
-  ~dynamic_cache()
+  ~thread_safe_chunk_cache()
   {
-    deleteEntries();
+    clear_chunks();
   }
 
-  void deleteEntries()
+  void clear_chunks()
   {
-    for( auto &chunk : m_cacheChunks )
+    for( auto& chunk : m_cacheChunks )
     {
       free( chunk );
     }
 
-    m_cache      .clear();
     m_cacheChunks.clear();
   }
 
+  T* get()
+  {
+    std::unique_lock<std::mutex> l( m_mutex );
+
+    if( m_cacheChunks.empty() )
+    {
+      l.unlock();
+      return ( T* ) malloc( DYN_CACHE_CHUNK_SIZE * sizeof( T ) );
+    }
+    else
+    {
+      T* chunk = m_cacheChunks.back();
+      m_cacheChunks.pop_back();
+      return chunk;
+    }
+  }
+
+  void cache( std::vector<T*>& chunks )
+  {
+    std::unique_lock<std::mutex> l( m_mutex );
+
+    m_cacheChunks.insert( m_cacheChunks.end(), chunks.begin(), chunks.end() );
+    chunks.clear();
+  }
+};
+
+template<typename T>
+class dynamic_cache
+{
+  ptrdiff_t       m_lastIdx;
+  std::vector<T*> m_cache;
+
+  thread_safe_chunk_cache<T> *m_chunkCache;
+
+  void clear_chunks()
+  {
+    m_chunkCache->cache( m_cache );
+  }
+
+public:
+
+  explicit dynamic_cache( thread_safe_chunk_cache<T>* chunkCache ) : m_lastIdx( DYN_CACHE_CHUNK_SIZE ), m_chunkCache( chunkCache )
+  {
+  }
+
+  ~dynamic_cache()
+  {
+    clear_chunks();
+  }
+
   T* get()
   {
     T* ret;
 
-    if( !m_cache.empty() )
+    if( m_lastIdx < DYN_CACHE_CHUNK_SIZE )
     {
-      ret = m_cache.back();
-      m_cache.pop_back();
+      ret = &m_cache.back()[m_lastIdx++];
     }
     else
     {
-      T* chunk = (T*) malloc( DYN_CACHE_CHUNK_SIZE * sizeof( T ) );
+      T* chunk = m_chunkCache->get();
 
-      //GCC_WARNING_DISABLE_class_memaccess
-      //memset( chunk, 0, DYN_CACHE_CHUNK_SIZE * sizeof( T ) );
-      //GCC_WARNING_RESET
+      m_cache.push_back( chunk );
 
-      m_cacheChunks.push_back( chunk );
-      m_cache      .reserve( m_cache.size() + DYN_CACHE_CHUNK_SIZE );
-
-      for( ptrdiff_t p = 0; p < DYN_CACHE_CHUNK_SIZE; p++ )
-      {
-        //m_cache.push_back( &chunk[DYN_CACHE_CHUNK_SIZE - p - 1] );
-        m_cache.push_back( &chunk[p] );
-      }
-
-      ret = m_cache.back();
-      m_cache.pop_back();
+      ret = &chunk[0];
+      m_lastIdx = 1;
     }
 
     return ret;
   }
 
-  void defragment()
+  void releaseAll()
   {
-    m_cache.clear();
-
-    for( T* chunk : m_cacheChunks )
-    {
-      //GCC_WARNING_DISABLE_class_memaccess
-      //memset( chunk, 0, DYN_CACHE_CHUNK_SIZE * sizeof( T ) );
-      //GCC_WARNING_RESET
-
-      for( ptrdiff_t p = 0; p < DYN_CACHE_CHUNK_SIZE; p++ )
-      {
-        //m_cache.push_back( &chunk[DYN_CACHE_CHUNK_SIZE - p - 1] );
-        m_cache.push_back( &chunk[p] );
-      }
-    }
-  }
-
-  void cache( std::vector<T*>& vel )
-  {
-    m_cache.insert( m_cache.end(), vel.begin(), vel.end() );
-    vel.clear();
+    clear_chunks();
+    m_lastIdx = DYN_CACHE_CHUNK_SIZE;
   }
 };
 
-typedef dynamic_cache<struct CodingUnit> CUCache;
+typedef dynamic_cache<struct CodingUnit>    CUCache;
 typedef dynamic_cache<struct TransformUnit> TUCache;
 
-struct ThreadSafeCUCache
-{
-  std::shared_ptr<CUCache> getCuCache();
-  std::shared_ptr<TUCache> getTuCache();
-
-private:
-  std::mutex m_mutex;
-
-  // we use the shared_ptr reference-count to track if a cache is in use or available
-  std::vector< std::shared_ptr<CUCache> > m_cuCaches;
-  std::vector< std::shared_ptr<TUCache> > m_tuCaches;
-};
+typedef thread_safe_chunk_cache<struct CodingUnit>    CUChunkCache;
+typedef thread_safe_chunk_cache<struct TransformUnit> TUChunkCache;
 
 }
 
