@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -106,7 +102,7 @@ void CABACReader::remaining_bytes( bool noTrailingBytesExpected )
 {
   if( noTrailingBytesExpected )
   {
-    CHECK( 0 != m_Bitstream->getNumBitsLeft(), "Bits left when not supposed" );
+//    CHECK_RECOVERABLE( 0 != m_Bitstream->getNumBitsLeft(), "Bits left when not supposed" );
   }
   else
   {
@@ -137,7 +133,7 @@ bool CABACReader::coding_tree_unit( CodingStructure& cs, Slice* slice, const Uni
   m_slice = slice;
 
   CUCtx cuCtx( qps[CH_L] );
-  Partitioner partitioner;
+  Partitioner &partitioner = m_partL;
 
   partitioner.initCtu( area, CH_L, cs, *m_slice );
   partitioner.treeType = TREE_D;
@@ -145,75 +141,14 @@ bool CABACReader::coding_tree_unit( CodingStructure& cs, Slice* slice, const Uni
 
   sao( cs, ctuRsAddr );
 
-  if( m_slice->getTileGroupAlfEnabledFlag( COMPONENT_Y ) )
-  {
-    const PreCalcValues& pcv                = *cs.pcv;
-    int                 frame_width_in_ctus = pcv.widthInCtus;
-    int                 ry                  = ctuRsAddr / frame_width_in_ctus;
-    int                 rx                  = ctuRsAddr - ry * frame_width_in_ctus;
-    const Position      pos( rx * cs.pcv->maxCUWidth, ry * cs.pcv->maxCUHeight );
-    bool                leftAvail  = cs.getCURestricted( pos.offset( -1, 0 ), pos, partitioner.currSliceIdx, partitioner.currTileIdx, CH_L ) ? true : false;
-    bool                aboveAvail = cs.getCURestricted( pos.offset( 0, -1 ), pos, partitioner.currSliceIdx, partitioner.currTileIdx, CH_L ) ? true : false;
-  
-    int leftCTUAddr  = leftAvail  ? ctuRsAddr - 1 : -1;
-    int aboveCTUAddr = aboveAvail ? ctuRsAddr - frame_width_in_ctus : -1;
-  
-    for( int compIdx = 0; compIdx < MAX_NUM_COMPONENT; compIdx++ )
-    {
-      if( m_slice->getTileGroupAlfEnabledFlag( ( ComponentID ) compIdx ) )
-      {
-        uint8_t* ctbAlfFlag = m_slice->getPic()->getAlfCtuEnableFlag( compIdx );
-        int ctx = 0;
-        ctx += leftCTUAddr  > -1 ? ( ctbAlfFlag[leftCTUAddr]  ? 1 : 0 ) : 0;
-        ctx += aboveCTUAddr > -1 ? ( ctbAlfFlag[aboveCTUAddr] ? 1 : 0 ) : 0;
-  
-        ctbAlfFlag[ctuRsAddr] = m_BinDecoder.decodeBin( Ctx::ctbAlfFlag( compIdx * 3 + ctx ) );
-  
-        if( isLuma( ( ComponentID ) compIdx ) && ctbAlfFlag[ctuRsAddr] )
-        {
-          readAlfCtuFilterIndex( cs, ctuRsAddr );
-        }
-
-        if( isChroma( ( ComponentID ) compIdx ) )
-        {
-          const int apsIdx                  = m_slice->getTileGroupApsIdChroma();
-          CHECK( m_slice->getAlfAPSs()[apsIdx] == nullptr, "APS not initialized" );
-          const AlfSliceParam& alfParam     = m_slice->getAlfAPSs()[apsIdx]->getAlfAPSParam();
-          const int numAlts                 = alfParam.numAlternativesChroma;
-          uint8_t* ctbAlfAlternative        = m_slice->getPic()->getAlfCtuAlternativeData( compIdx-1 );
-          ctbAlfAlternative[ctuRsAddr]      = 0;
-
-          if( ctbAlfFlag[ctuRsAddr] )
-          {
-            uint8_t decoded = 0;
-            while( decoded < numAlts - 1 && m_BinDecoder.decodeBin( Ctx::ctbAlfAlternative( compIdx - 1 ) ) )
-              ++decoded;
-            ctbAlfAlternative[ctuRsAddr] = decoded;
-          }
-        }
-      }
-    }
-  }
-
-  for( int compIdx = 1; compIdx < getNumberValidComponents( cs.pcv->chrFormat ); compIdx++ )
-  {
-    if( m_slice->getTileGroupCcAlfEnabledFlag( compIdx - 1 ) )
-    {
-      const int apsIdx        = compIdx == 1 ? m_slice->getTileGroupCcAlfCbApsId() : m_slice->getTileGroupCcAlfCrApsId();
-      const int filterCount   = m_slice->getAlfAPSs()[apsIdx]->getCcAlfAPSParam().ccAlfFilterCount[compIdx - 1];
-  
-      const Position lumaPos  = area.lumaPos();
-  
-      ccAlfFilterControlIdc( cs, ComponentID(compIdx), ctuRsAddr, m_slice->getPic()->getccAlfFilterControl( compIdx - 1 ), lumaPos, filterCount );
-    }
-  }
+  readAlf(cs, ctuRsAddr, partitioner);
 
   bool isLast = false;
 
   if( partitioner.isDualITree && cs.pcv->chrFormat != CHROMA_400 )
   {
     CUCtx cuCtxC( qps[CH_C] );
-    Partitioner partitionerC;
+    Partitioner &partitionerC = m_partC;
 
     partitionerC.initCtu( area, CH_C, cs, *m_slice );
     partitionerC.treeType = TREE_D;
@@ -282,10 +217,9 @@ bool CABACReader::dt_implicit_qt_split( CodingStructure& cs, Partitioner& partit
   return isLast;
 }
 
-void CABACReader::readAlfCtuFilterIndex( CodingStructure& cs, unsigned ctuRsAddr )
+short CABACReader::readAlfCtuFilterIndex( CodingStructure& cs, unsigned ctuRsAddr )
 {
-  short* alfCtbFilterSetIndex         = m_slice->getPic()->getAlfCtbFilterIndex();
-  const unsigned numAps               = m_slice->getTileGroupNumAps();
+  const unsigned numAps               = m_slice->getNumAlfAps();
   const unsigned numAvailableFiltSets = numAps + NUM_FIXED_FILTER_SETS;
   uint32_t filtIndex = 0;
 
@@ -306,41 +240,7 @@ void CABACReader::readAlfCtuFilterIndex( CodingStructure& cs, unsigned ctuRsAddr
     xReadTruncBinCode( filtIndex, NUM_FIXED_FILTER_SETS );
   }
 
-  alfCtbFilterSetIndex[ctuRsAddr] = filtIndex;
-}
-
-void CABACReader::ccAlfFilterControlIdc(CodingStructure &cs, const ComponentID compID, const int curIdx, uint8_t *filterControlIdc, Position lumaPos, int filterCount)
-{
-  Position       leftLumaPos    = lumaPos.offset(-(int)cs.pcv->maxCUWidth, 0);
-  Position       aboveLumaPos   = lumaPos.offset(0, -(int)cs.pcv->maxCUWidth);
-  const uint32_t curSliceIdx    = m_slice->getIndependentSliceIdx();
-  const uint32_t curTileIdx     = cs.pps->getTileIdx( lumaPos );
-  bool           leftAvail      = cs.getCURestricted( leftLumaPos,  lumaPos, curSliceIdx, curTileIdx, CH_L ) ? true : false;
-  bool           aboveAvail     = cs.getCURestricted( aboveLumaPos, lumaPos, curSliceIdx, curTileIdx, CH_L ) ? true : false;
-  int            ctxt           = 0;
-
-  if (leftAvail)
-  {
-    ctxt += ( filterControlIdc[curIdx - 1] ) ? 1 : 0;
-  }
-  if (aboveAvail)
-  {
-    ctxt += ( filterControlIdc[curIdx - cs.pcv->widthInCtus] ) ? 1 : 0;
-  }
-  ctxt += ( compID == COMPONENT_Cr ) ? 3 : 0;
-
-  int idcVal  = m_BinDecoder.decodeBin( Ctx::CcAlfFilterControlFlag( ctxt ) );
-  if ( idcVal )
-  {
-    while ( ( idcVal != filterCount ) && m_BinDecoder.decodeBinEP() )
-    {
-      idcVal++;
-    }
-  }
-  filterControlIdc[curIdx] = idcVal;
-
-  DTRACE(g_trace_ctx, D_SYNTAX, "ccAlfFilterControlIdc() compID=%d pos=(%d,%d) ctxt=%d, filterCount=%d, idcVal=%d\n",
-         compID, lumaPos.x, lumaPos.y, ctxt, filterCount, idcVal);
+  return filtIndex;
 }
 
 //================================================================================
@@ -488,8 +388,86 @@ void CABACReader::sao( CodingStructure& cs, unsigned ctuRsAddr )
   }
 }
 
+//================================================================================
+//    void  readAlf( cs, ctuRsAddr, partitioner )
+//================================================================================
+void CABACReader::readAlf( CodingStructure& cs, unsigned int ctuRsAddr, const Partitioner& partitioner )
+{
+  const PreCalcValues& pcv                = *cs.pcv;
+  int                 frame_width_in_ctus = pcv.widthInCtus;
+  int                 ry                  = ctuRsAddr / frame_width_in_ctus;
+  int                 rx                  = ctuRsAddr - ry * frame_width_in_ctus;
+  const Position      pos                 ( rx * cs.pcv->maxCUWidth, ry * cs.pcv->maxCUHeight );
+  bool                leftAvail           = cs.getCURestricted( pos.offset( -1, 0 ), pos, partitioner.currSliceIdx, partitioner.currTileIdx, CH_L ) ? true : false;
+  bool                aboveAvail          = cs.getCURestricted( pos.offset( 0, -1 ), pos, partitioner.currSliceIdx, partitioner.currTileIdx, CH_L ) ? true : false;
+  
+  CtuAlfData& currAlfData = cs.getCtuData( ctuRsAddr ).alfParam;
+  CtuAlfData  leftAlfData, aboveAlfData;
 
+  if( leftAvail )  leftAlfData  = cs.getCtuData( ctuRsAddr -                   1 ).alfParam;
+  if( aboveAvail ) aboveAlfData = cs.getCtuData( ctuRsAddr - frame_width_in_ctus ).alfParam;
 
+  if( m_slice->getAlfEnabledFlag( COMPONENT_Y ) )
+  {
+    for( int compIdx = 0; compIdx < MAX_NUM_COMPONENT; compIdx++ )
+    {
+      if( m_slice->getAlfEnabledFlag( ( ComponentID ) compIdx ) )
+      {
+        //uint8_t* ctbAlfFlag = m_slice->getPic()->getAlfCtuEnableFlag( compIdx );
+        int ctx = 0;
+          ctx += leftAlfData.alfCtuEnableFlag[compIdx];
+          ctx += aboveAlfData.alfCtuEnableFlag[compIdx];
+      
+        currAlfData.alfCtuEnableFlag[compIdx] = m_BinDecoder.decodeBin( Ctx::ctbAlfFlag( compIdx * 3 + ctx ) );
+      
+        if( isLuma( ( ComponentID ) compIdx ) && currAlfData.alfCtuEnableFlag[compIdx] )
+        {
+          currAlfData.alfCtbFilterIndex = readAlfCtuFilterIndex( cs, ctuRsAddr );
+        }
+
+        if( isChroma( ( ComponentID ) compIdx ) )
+        {
+          const int apsIdx                  = m_slice->getAlfApsIdChroma();
+          CHECK_RECOVERABLE( m_slice->getAlfAPSs()[apsIdx] == nullptr, "APS not initialized" );
+          const AlfSliceParam& alfParam     = m_slice->getAlfAPSs()[apsIdx]->getAlfAPSParam();
+          const int numAlts                 = alfParam.numAlternativesChroma;
+          currAlfData.alfCtuAlternative[compIdx - 1] = 0;
+
+          if( currAlfData.alfCtuEnableFlag[compIdx] )
+          {
+            uint8_t decoded = 0;
+            while( decoded < numAlts - 1 && m_BinDecoder.decodeBin( Ctx::ctbAlfAlternative( compIdx - 1 ) ) )
+              ++decoded;
+            currAlfData.alfCtuAlternative[compIdx - 1] = decoded;
+          }
+        }
+      }
+    }
+  }
+  for( int compIdx = 1; compIdx < getNumberValidComponents( cs.pcv->chrFormat ); compIdx++ )
+  {
+    if( m_slice->getCcAlfEnabledFlag( compIdx - 1 ) )
+    {
+      int ctxt = 0;
+      ctxt += ( leftAlfData.ccAlfFilterControl[compIdx - 1] ) ? 1 : 0;
+      ctxt += ( aboveAlfData.ccAlfFilterControl[compIdx - 1] ) ? 1 : 0;
+      ctxt += ( compIdx == COMPONENT_Cr ) ? 3 : 0;
+
+      int idcVal  = m_BinDecoder.decodeBin( Ctx::CcAlfFilterControlFlag( ctxt ) );
+
+      if ( idcVal )
+      {
+        const int apsIdx        = compIdx == 1 ? m_slice->getCcAlfCbApsId() : m_slice->getCcAlfCrApsId();
+        const int filterCount   = m_slice->getAlfAPSs()[apsIdx]->getCcAlfAPSParam().ccAlfFilterCount[compIdx - 1];
+        while ( ( idcVal != filterCount ) && m_BinDecoder.decodeBinEP() )
+        {
+          idcVal++;
+        }
+      }
+      currAlfData.ccAlfFilterControl[compIdx - 1] = idcVal;
+    }
+  }
+}
 
 //================================================================================
 //  clause 7.3.11.4
@@ -528,7 +506,7 @@ bool CABACReader::coding_tree( CodingStructure& cs, Partitioner& partitioner, CU
     //decide chroma split or not
     chromaNotSplit = modeTypeParent == MODE_TYPE_ALL && partitioner.modeType == MODE_TYPE_INTRA;
 
-    CHECK( chromaNotSplit && partitioner.chType != CHANNEL_TYPE_LUMA, "chType must be luma" );
+    CHECK_RECOVERABLE( chromaNotSplit && partitioner.chType != CHANNEL_TYPE_LUMA, "chType must be luma" );
 
     if( partitioner.treeType == TREE_D )
     {
@@ -611,7 +589,7 @@ bool CABACReader::coding_tree( CodingStructure& cs, Partitioner& partitioner, CU
     //therefore, after decoding the chroma CU, the cuCtx.qp shall be recovered to luma qp in order to decode next luma cu qp
 //    const CodingUnit* colLumaCu = cs.getLumaCU( lumaRefPos );
     const CodingUnit* colLumaCu = cs.getCU( lumaRefPos, CHANNEL_TYPE_LUMA );
-    CHECK( colLumaCu == nullptr, "colLumaCU shall exist" );
+    CHECK_RECOVERABLE( colLumaCu == nullptr, "colLumaCU shall exist" );
     lumaQPinLocalDualTree = cuCtx.qp;
 
     if( colLumaCu ) cuCtx.qp = colLumaCu->qp;
@@ -644,9 +622,6 @@ bool CABACReader::coding_tree( CodingStructure& cs, Partitioner& partitioner, CU
     }
   }
 
-  m_slice->incNumCu();
-  if( cu.predMode() != MODE_INTER || cu.ciipFlag() ) m_slice->incNumIntraCu();
-
 #if ENABLE_TRACING
   if( cu.chType() == CHANNEL_TYPE_CHROMA )
   {
@@ -956,6 +931,7 @@ void CABACReader::cu_skip_flag( CodingUnit& cu )
       //cu.setRootCbf    ( false );
       cu.setPredMode   ( MODE_IBC );
       //cu.setMmvdFlag   ( false );
+      cu.cs->hasIbcBlock[cu.ctuData->lineIdx] = 1;
     }
 
     return;
@@ -980,6 +956,7 @@ void CABACReader::cu_skip_flag( CodingUnit& cu )
         //cu.setRootCbf    ( false );
         cu.setPredMode   ( MODE_IBC );
         //cu.setMmvdFlag   ( false );
+        cu.cs->hasIbcBlock[cu.ctuData->lineIdx] = 1;
         return;
       }
 
@@ -992,6 +969,7 @@ void CABACReader::cu_skip_flag( CodingUnit& cu )
         cu.setPredMode           ( MODE_IBC );
         //cu.setMmvdFlag           ( false );
         //cu.setRegularMergeFlag   ( false );
+        cu.cs->hasIbcBlock[cu.ctuData->lineIdx] = 1;
       }
       else
       {
@@ -1114,6 +1092,7 @@ void CABACReader::pred_mode( CodingUnit& cu )
     if( m_BinDecoder.decodeBin( Ctx::IBCFlag( ctxidx ) ) )
     {
       cu.setPredMode( MODE_IBC );
+      cu.cs->hasIbcBlock[cu.ctuData->lineIdx] = 1;
     }
   }
 }
@@ -1186,6 +1165,7 @@ void CABACReader::cu_pred_data( CodingUnit &cu )
   if( !cu.Y().valid() ) // dual tree chroma CU
   {
     cu.setPredMode( MODE_IBC );
+    cu.cs->hasIbcBlock[cu.ctuData->lineIdx] = 1;
     return;
   }
 
@@ -1208,7 +1188,7 @@ void CABACReader::cu_bcw_flag(CodingUnit& cu)
     return;
   }
 
-  CHECK(!(BCW_NUM > 1 && (BCW_NUM == 2 || (BCW_NUM & 0x01) == 1)), " !( BCW_NUM > 1 && ( BCW_NUM == 2 || ( BCW_NUM & 0x01 ) == 1 ) ) ");
+  CHECK_RECOVERABLE(!(BCW_NUM > 1 && (BCW_NUM == 2 || (BCW_NUM & 0x01) == 1)), " !( BCW_NUM > 1 && ( BCW_NUM == 2 || ( BCW_NUM & 0x01 ) == 1 ) ) ");
 
   uint32_t idx    = 0;
   uint32_t symbol = m_BinDecoder.decodeBin( Ctx::BcwIdx( 0 ) );
@@ -1225,7 +1205,7 @@ void CABACReader::cu_bcw_flag(CodingUnit& cu)
   }
 
   uint8_t bcwIdx = ( uint8_t ) g_BcwParsingOrder[idx];
-  CU::setBcwIdx( cu, bcwIdx );
+  CU::setBcwIdx( cu, g_BcwInternFwd[bcwIdx] );
 
   DTRACE(g_trace_ctx, D_SYNTAX, "cu_bcw_flag() bcw_idx=%d\n", cu.BcwIdx() ? 1 : 0);
 }
@@ -1846,7 +1826,7 @@ void CABACReader::merge_idx( PredictionUnit& pu )
     const int maxNumGeoCand = pu.sps->getMaxNumGeoCand();
     const int numCandminus2 = maxNumGeoCand - 2;
 
-    CHECK( maxNumGeoCand < 2, "Incorrect max number of geo candidates" );
+    CHECK_RECOVERABLE( maxNumGeoCand < 2, "Incorrect max number of geo candidates" );
 
     int mergeCand0 = 0;
     int mergeCand1 = 0;
@@ -2323,7 +2303,7 @@ void CABACReader::transform_unit( TransformUnit& tu, CUCtx& cuCtx, Partitioner&
 
 void CABACReader::cu_qp_delta( CodingUnit& cu, int predQP, int8_t& qp )
 {
-  CHECK( predQP == std::numeric_limits<int>::max(), "Invalid predicted QP" );
+  CHECK_RECOVERABLE( predQP == std::numeric_limits<int>::max(), "Invalid predicted QP" );
   int qpY = predQP;
   int DQp = unary_max_symbol( Ctx::DeltaQP(), Ctx::DeltaQP(1), CU_DQP_TU_CMAX );
   if( DQp >= CU_DQP_TU_CMAX )
@@ -2409,7 +2389,7 @@ void CABACReader::residual_coding( TransformUnit& tu, ComponentID compID, CUCtx&
 
   // determine sign hiding
   bool signHiding = m_slice->getSignDataHidingEnabledFlag();
-  CoeffCodingContext  cctx    ( tu, compID, signHiding );
+  CoeffCodingContext  cctx( tu, compID, signHiding, m_tplBuf );
   // parse last coeff position
   cctx.setScanPosLast( last_sig_coeff( cctx, tu, compID ) );
   if (tu.mtsIdx( compID ) != MTS_SKIP && tu.blocks[compID].height >= 4 && tu.blocks[compID].width >= 4 )
@@ -2432,7 +2412,7 @@ void CABACReader::residual_coding( TransformUnit& tu, ComponentID compID, CUCtx&
   int       state         = 0;
 
   TCoeffSig *coeff = m_cffTmp;
-  ::memset( coeff, 0, ( cctx.maxNumCoeff() + 2 * cctx.width() ) * sizeof( TCoeffSig ) );
+  ::memset( coeff, 0, cctx.maxNumCoeff() * sizeof( TCoeffSig ) );
 
   int *sigPos      = m_blkPos;
   int  sigSubSetId = 0;
@@ -2453,9 +2433,7 @@ void CABACReader::residual_coding( TransformUnit& tu, ComponentID compID, CUCtx&
       }
     }
 
-    int numSigCoef = cctx.checkTplBnd()
-                   ? residual_coding_subblock<true >( cctx, coeff, stateTransTab, state, m_signVal[sigSubSetId], sigPos, m_sub1[sigSubSetId] )
-                   : residual_coding_subblock<false>( cctx, coeff, stateTransTab, state, m_signVal[sigSubSetId], sigPos, m_sub1[sigSubSetId] );
+    int numSigCoef = residual_coding_subblock( cctx, coeff, stateTransTab, state, m_signVal[sigSubSetId], sigPos, m_sub1[sigSubSetId] );
 
     if( numSigCoef > 0 )
     {
@@ -2509,8 +2487,16 @@ void CABACReader::residual_coding( TransformUnit& tu, ComponentID compID, CUCtx&
     }
   }
 
-  tu.maxScanPosX [compID] = maxX - 1;
-  tu.maxScanPosY [compID] = maxY - 1;
+  if( cctx.scanPosLast() == 0 )
+  {
+    tu.maxScanPosX[compID] = 0;
+    tu.maxScanPosY[compID] = 0;
+  }
+  else
+  {
+    tu.maxScanPosX[compID] = maxX - 1;
+    tu.maxScanPosY[compID] = maxY - 1;
+  }
 }
 
 void CABACReader::ts_flag( TransformUnit& tu, ComponentID compID )
@@ -2724,7 +2710,6 @@ int CABACReader::last_sig_coeff( CoeffCodingContext& cctx, TransformUnit& tu, Co
 }
 
 
-template<bool checkBnd>
 int CABACReader::residual_coding_subblock(CoeffCodingContext& cctx, TCoeffSig* coeff, const int stateTransTable, int& state, unsigned& signPattern, int *&sigPos, unsigned &stateVal)
 {
   // NOTE: All coefficients of the subblock must be set to zero before calling this function
@@ -2769,10 +2754,12 @@ int CABACReader::residual_coding_subblock(CoeffCodingContext& cctx, TCoeffSig* c
     int  blkPos  = cctx.blockPos(nextSigPos);
     bool sigFlag = (!numNonZero && nextSigPos == inferSigPos);
 
+    unsigned absVal = 0;
+
     if (!sigFlag)
     {
-      const unsigned sigCtxId = cctx.sigCtxIdAbs<checkBnd>(blkPos, coeff, state);
-      sigFlag = m_BinDecoder.decodeBin(sigCtxId);
+      const unsigned sigCtxId = cctx.sigCtxIdAbs( blkPos, state );
+      sigFlag = m_BinDecoder.decodeBin( sigCtxId );
       DTRACE(g_trace_ctx, D_SYNTAX_RESI, "sig_bin() bin=%d ctx=%d\n", sigFlag, sigCtxId);
       remRegBins--;
     }
@@ -2802,14 +2789,16 @@ int CABACReader::residual_coding_subblock(CoeffCodingContext& cctx, TCoeffSig* c
         DTRACE( g_trace_ctx, D_SYNTAX_RESI, "gt2_flag() bin=%d ctx=%d\n", gt2Flag, cctx.greater2CtxIdAbs( ctxOff ) );
         remRegBins--;
         *gt1PosPtr++    = blkPos;
-        coeff[blkPos]   = 2 + parFlag + (gt2Flag << 1);
+        absVal          = 2 + parFlag + (gt2Flag << 1);
         state           = ( stateTransTable >> ( ( state << 2 ) + ( parFlag << 1 ) ) ) & 3;
       }
       else
       {
-        coeff[blkPos] = 1;
-        state         = ( stateTransTable >> ( ( state << 2 ) + 2 ) ) & 3;
+        absVal = 1;
+        state  = ( stateTransTable >> ( ( state << 2 ) + 2 ) ) & 3;
       }
+
+      cctx.absVal1stPass( blkPos, coeff, absVal );
     }
     else
     {
@@ -2889,7 +2878,7 @@ void CABACReader::residual_codingTS( TransformUnit& tu, ComponentID compID )
   pb.memset( 0 );
 
   // init coeff coding context
-  CoeffCodingContext  cctx    ( tu, compID, false );
+  CoeffCodingContext  cctx( tu, compID, false, m_tplBuf );
   TCoeffSig *coeff = m_cffTmp;
   ::memset( coeff, 0, cctx.maxNumCoeff() * sizeof( TCoeffSig ) );
 
@@ -3165,4 +3154,4 @@ void CABACReader::mip_pred_mode( PredictionUnit &pu )
   DTRACE( g_trace_ctx, D_SYNTAX, "mip_pred_mode() pos=(%d,%d) mode=%d transposed=%d\n", pu.lumaPos().x, pu.lumaPos().y, pu.intraDir[CHANNEL_TYPE_LUMA], pu.mipTransposedFlag() ? 1 : 0 );
 }
 
-}
\ No newline at end of file
+}
