@@ -1,45 +1,41 @@
 /* -----------------------------------------------------------------------------
-The copyright in this software is being made available under the BSD
+The copyright in this software is being made available under the Clear BSD
 License, included below. No patent rights, trademark rights and/or 
 other Intellectual Property Rights other than the copyrights concerning 
 the Software are granted under this license.
 
-For any license concerning other Intellectual Property rights than the software, 
-especially patent licenses, a separate Agreement needs to be closed. 
-For more information please contact:
+The Clear BSD License
 
-Fraunhofer Heinrich Hertz Institute
-Einsteinufer 37
-10587 Berlin, Germany
-www.hhi.fraunhofer.de/vvc
-vvc@hhi.fraunhofer.de
-
-Copyright (c) 2018-2021, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. 
+Copyright (c) 2018-2023, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The VVdeC Authors.
 All rights reserved.
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are met:
+Redistribution and use in source and binary forms, with or without modification,
+are permitted (subject to the limitations in the disclaimer below) provided that
+the following conditions are met:
 
- * Redistributions of source code must retain the above copyright notice,
-   this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright notice,
-   this list of conditions and the following disclaimer in the documentation
-   and/or other materials provided with the distribution.
- * Neither the name of Fraunhofer nor the names of its contributors may
-   be used to endorse or promote products derived from this software without
-   specific prior written permission.
+     * Redistributions of source code must retain the above copyright notice,
+     this list of conditions and the following disclaimer.
 
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
-BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
-THE POSSIBILITY OF SUCH DAMAGE.
+     * Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+
+     * Neither the name of the copyright holder nor the names of its
+     contributors may be used to endorse or promote products derived from this
+     software without specific prior written permission.
+
+NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
+THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
+CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
+IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+POSSIBILITY OF SUCH DAMAGE.
 
 
 ------------------------------------------------------------------------------------------- */
@@ -54,8 +50,11 @@ THE POSSIBILITY OF SUCH DAMAGE.
 #include "CommonLib/dtrace_buffer.h"
 #include "CommonLib/TimeProfiler.h"
 
+#include "CommonLib/x86/CommonDefX86.h"
+
 #include "NALread.h"
 
+
 namespace vvdec
 {
 
@@ -113,9 +112,9 @@ DecLib::DecLib()
 }
 
 #if RPR_YUV_OUTPUT
-void DecLib::create(int numDecThreads, int parserFrameDelay, int upscaledOutput)
+void DecLib::create( int numDecThreads, int parserFrameDelay, const UserAllocator& userAllocator, ErrHandlingFlags errHandlingFlags, int upscaledOutput )
 #else
-void DecLib::create(int numDecThreads, int parserFrameDelay)
+void DecLib::create( int numDecThreads, int parserFrameDelay, const UserAllocator& userAllocator, ErrHandlingFlags errHandlingFlags )
 #endif
 {
   // run constructor again to ensure all variables, especially in DecLibParser have been reset
@@ -124,7 +123,7 @@ void DecLib::create(int numDecThreads, int parserFrameDelay)
 
   if( numDecThreads < 0 )
   {
-    numDecThreads = std::thread::hardware_concurrency();
+    numDecThreads = std::min<unsigned>( MAX_THREADS, std::thread::hardware_concurrency() );
   }
 
   m_decodeThreadPool.reset( new ThreadPool( numDecThreads, "DecThread" ) );
@@ -135,28 +134,30 @@ void DecLib::create(int numDecThreads, int parserFrameDelay)
     parserFrameDelay = numDecThreads;
   }
   m_parseFrameDelay = parserFrameDelay;
+
+  bool upscalingEnabled = false;
 #if RPR_YUV_OUTPUT
   m_upscaledOutput = upscaledOutput;
+  upscalingEnabled = upscaledOutput;
 #endif
-  m_picListManager.create( m_parseFrameDelay, ( int ) m_decLibRecon.size() );
-  m_decLibParser.create  ( m_decodeThreadPool.get(), m_parseFrameDelay, ( int ) m_decLibRecon.size(), numDecThreads );
-    
+  m_picListManager.create( m_parseFrameDelay, ( int ) m_decLibRecon.size(), upscalingEnabled, userAllocator );
+  m_decLibParser.create  ( m_decodeThreadPool.get(), m_parseFrameDelay, ( int ) m_decLibRecon.size(), numDecThreads, errHandlingFlags );
+
   int id=0;
   for( auto &dec: m_decLibRecon )
   {
-    dec.create( m_decodeThreadPool.get(), id++ );
+    dec.create( m_decodeThreadPool.get(), id++, upscalingEnabled );
   }
 
   std::stringstream cssCap;
   cssCap << "THREADS="     << numDecThreads << "; "
          << "PARSE_DELAY=" << parserFrameDelay << "; ";
 #if ENABLE_SIMD_OPT
-#if defined( TARGET_SIMD_X86 )
-  std::string cSIMD;
-  cssCap << "SIMD=" << read_x86_extension( cSIMD );
-#else
+#  if defined( TARGET_SIMD_X86 )
+  cssCap << "SIMD=" << read_x86_extension_name();
+#  else
   cssCap << "SIMD=SCALAR";
-#endif
+#  endif
 #else
   cssCap << "SIMD=NONE";
 #endif
@@ -186,30 +187,76 @@ void DecLib::destroy()
 Picture* DecLib::decode( InputNALUnit& nalu, int* pSkipFrame, int iTargetLayer )
 {
   PROFILER_SCOPE_AND_STAGE( 1, g_timeProfiler, P_NALU_SLICE_PIC_HL );
-  Picture * pcParsedPic = nullptr;
-  if( m_iMaxTemporalLayer >= 0 && nalu.m_temporalId > m_iMaxTemporalLayer )
+
+  bool newPic = false;
+  if( m_iMaxTemporalLayer < 0 || nalu.m_temporalId <= m_iMaxTemporalLayer )
   {
-    pcParsedPic = nullptr;
-  }
-  else
-  {
-    pcParsedPic = m_decLibParser.parse( nalu, pSkipFrame, iTargetLayer );
+    newPic = m_decLibParser.parse( nalu, pSkipFrame, iTargetLayer );
   }
 
-  if( pcParsedPic )
+  if( newPic )
   {
-    this->decompressPicture( pcParsedPic );
+    Picture* pcParsedPic = m_decLibParser.getNextDecodablePicture();
+    if( pcParsedPic )
+    {
+      while( pcParsedPic->error || pcParsedPic->wasLost || pcParsedPic->parseDone.hasException() )
+      {
+        CHECK( pcParsedPic->progress >= Picture::reconstructing, "The error picture shouldn't be in reconstructing state yet." );
+
+        std::exception_ptr parsing_exception = pcParsedPic->parseDone.hasException() ? pcParsedPic->parseDone.getException() : nullptr;
+        if( parsing_exception )
+        {   // the exception has not been thrown out of the library. Do that after preparing this picture for referencing
+          pcParsedPic->error = true;
+          pcParsedPic->waitForAllTasks();
+          pcParsedPic->parseDone.clearException();
+        }
+
+        pcParsedPic->waitForAllTasks();
+
+        if( pcParsedPic->progress < Picture::parsing )
+        {
+          // we don't know if all structures are there yet, so we init them
+          pcParsedPic->ensureUsableAsRef();
+        }
+        pcParsedPic->fillGrey( m_decLibParser.getParameterSetManager().getFirstSPS() );
+
+        // need to finish picture here, because it won't go through declibRecon
+        finishPicture( pcParsedPic );
+
+        // this exception has not been thrown outside (error must have happened in slice parsing task)
+        if( parsing_exception )
+        {
+          CHECK( pcParsedPic->exceptionThrownOut, "The exception shouldn't have been thrown out already." );
+          pcParsedPic->exceptionThrownOut = true;
+          std::rethrow_exception( parsing_exception );
+        }
+
+        // try again to get a picture, that we can reconstruct now
+        pcParsedPic = m_decLibParser.getNextDecodablePicture();
+      }
+
+      reconPicture( pcParsedPic );
+    }
   }
 
-  if( m_decLibParser.getParseNewPicture() &&
-      ( pcParsedPic || nalu.isSlice() || nalu.m_nalUnitType == NAL_UNIT_EOS ) )
+  if( newPic || nalu.m_nalUnitType == NAL_UNIT_EOS )
   {
     Picture* outPic = getNextOutputPic( false );
-    CHECK_WARN( m_checkMissingOutput && !outPic, "missing output picture" ); // we don't need this CHECK in flushPic(), because flushPic() is usually only called until the first nullptr is returned
     if( outPic )
     {
+      CHECK_WARN( outPic->progress < Picture::finished, "Picture should have been finished by now. Blocking and finishing..." );
+      if( outPic->progress < Picture::finished )
+      {
+        blockAndFinishPictures( outPic );
+
+        CHECK_RECOVERABLE( outPic->progress < Picture::finished, "Picture still not finished. Something is really broken." );
+      }
+
       m_checkMissingOutput = true;
     }
+
+    // warn if we don't produce an output picture for every incoming picture
+    CHECK_WARN( m_checkMissingOutput && !outPic, "missing output picture" ); // this CHECK is not needed in flushPic(), because in flushPic() the nullptr signals the end of the bitstream
     return outPic;
   }
 
@@ -218,26 +265,46 @@ Picture* DecLib::decode( InputNALUnit& nalu, int* pSkipFrame, int iTargetLayer )
 
 Picture* DecLib::flushPic()
 {
-  // at end of file, fill the decompression queue and decode pictures until we get one out
+  Picture* outPic = getNextOutputPic( false );
+  // at end of file, fill the decompression queue and decode pictures until the next output-picture is finished
   while( Picture* pcParsedPic = m_decLibParser.getNextDecodablePicture() )
   {
-    this->decompressPicture( pcParsedPic );
+    // reconPicture() blocks and finishes one picture on each call
+    reconPicture( pcParsedPic );
 
-    if( Picture* outPic = getNextOutputPic( false ) )
+    if( !outPic )
+    {
+      outPic = getNextOutputPic( false );
+    }
+    if( outPic && outPic->progress == Picture::finished )
     {
       return outPic;
     }
   }
 
-  // first try to get a picture without waiting for the decoder
-  if( Picture* outPic = getNextOutputPic( false ) )
+  if( outPic && outPic->progress == Picture::finished )
   {
     return outPic;
   }
 
-  // if no picture is done, actually block and wait
-  if( Picture* outPic = getNextOutputPic( true ) )
+  // if all pictures have been parsed, but not finished, iteratively wait for and finish next pictures
+  blockAndFinishPictures( outPic );
+  if( !outPic )
   {
+    outPic = getNextOutputPic( false );
+  }
+  if( outPic && outPic->progress == Picture::finished )
+  {
+    return outPic;
+  }
+
+  CHECK_RECOVERABLE( outPic, "we shouldn't be holding an output picture here" );
+  // flush remaining pictures without considering num reorder pics
+  outPic = getNextOutputPic( true );
+  if( outPic )
+  {
+    CHECK_RECOVERABLE( outPic->progress != Picture::finished, "all pictures should have been finished by now" );
+    outPic->referenced = false;
     return outPic;
   }
 
@@ -262,10 +329,32 @@ int DecLib::finishPicture( Picture* pcPic, MsgLevel msgl )
 #endif
 
   Slice*  pcSlice = pcPic->slices[0];
-  if( pcPic->wasLost )
+  if( pcPic->wasLost || pcPic->error || pcPic->reconDone.hasException() )
   {
-    msg( msgl, "POC %4d TId: %1d LOST\n", pcPic->poc, pcSlice->getTLayer() );
-    pcPic->reconstructed = true;
+    msg( msgl, "POC %4d LId: %2d TId: %1d %s\n", pcPic->poc, pcPic->layerId, pcSlice->getTLayer(), pcPic->wasLost ? "LOST" : "ERROR" );
+    pcPic->progress = Picture::finished;
+
+    // if the picture has an exception set (originating from thread-pool tasks), don't return here, but rethrow the exception
+    try
+    {
+      pcPic->parseDone.checkAndRethrowException();
+      pcPic->reconDone.checkAndRethrowException();
+    }
+    catch( ... )
+    {
+      pcPic->waitForAllTasks();
+
+      // need to clear exception so we can use it as reference picture
+      pcPic->reconDone.clearException();
+      pcPic->reconDone.unlock();
+      pcPic->error = true;
+      if( !pcPic->exceptionThrownOut )
+      {
+        pcPic->exceptionThrownOut = true;
+        throw;
+      }
+    }
+
     return pcPic->poc;
   }
 
@@ -296,18 +385,18 @@ int DecLib::finishPicture( Picture* pcPic, MsgLevel msgl )
       {
         if( scaleRatio.first != 1 << SCALE_RATIO_BITS || scaleRatio.second != 1 << SCALE_RATIO_BITS )
         {
-          msg( msgl, " %dc(%1.2lfx, %1.2lfx)", pcSlice->getRefPOC( RefPicList( iRefList ), iRefIndex ), double( scaleRatio.first ) / ( 1 << SCALE_RATIO_BITS ), double( scaleRatio.second ) / ( 1 << SCALE_RATIO_BITS ) );
+          msg( msgl, "%dc(%1.2lfx, %1.2lfx) ", pcSlice->getRefPOC( RefPicList( iRefList ), iRefIndex ), double( scaleRatio.first ) / ( 1 << SCALE_RATIO_BITS ), double( scaleRatio.second ) / ( 1 << SCALE_RATIO_BITS ) );
         }
         else
         {
-          msg( msgl, " %dc", pcSlice->getRefPOC( RefPicList( iRefList ), iRefIndex ) );
+          msg( msgl, "%dc ", pcSlice->getRefPOC( RefPicList( iRefList ), iRefIndex ) );
         }
       }
       else
       {
         if( scaleRatio.first != 1 << SCALE_RATIO_BITS || scaleRatio.second != 1 << SCALE_RATIO_BITS )
         {
-          msg( msgl, " %d(%1.2lfx, %1.2lfx)", pcSlice->getRefPOC( RefPicList( iRefList ), iRefIndex ), double( scaleRatio.first ) / ( 1 << SCALE_RATIO_BITS ), double( scaleRatio.second ) / ( 1 << SCALE_RATIO_BITS ) );
+          msg( msgl, "%d(%1.2lfx, %1.2lfx) ", pcSlice->getRefPOC( RefPicList( iRefList ), iRefIndex ), double( scaleRatio.first ) / ( 1 << SCALE_RATIO_BITS ), double( scaleRatio.second ) / ( 1 << SCALE_RATIO_BITS ) );
         }
         else
         {
@@ -358,6 +447,8 @@ int DecLib::finishPicture( Picture* pcPic, MsgLevel msgl )
 
   ITT_TASKEND( itt_domain_oth, itt_handle_finish );
 
+  pcPic->progress = Picture::finished;
+
   return pcSlice->getPOC();
 }
 
@@ -368,7 +459,7 @@ void DecLib::checkPictureHashSEI( Picture* pcPic )
     return;
   }
 
-  CHECK( !pcPic->reconstructed, "picture not reconstructed" );
+  CHECK_RECOVERABLE( pcPic->progress < Picture::reconstructed, "picture not reconstructed" );
 
   seiMessages pictureHashes = SEI_internal::getSeisByType( pcPic->seiMessageList, VVDEC_DECODED_PICTURE_HASH );
 
@@ -397,7 +488,7 @@ void DecLib::checkPictureHashSEI( Picture* pcPic )
     seiMessages scalableNestingSeis = SEI_internal::getSeisByType( pcPic->seiMessageList, VVDEC_SCALABLE_NESTING );
     for( auto* seiIt: scalableNestingSeis )
     {
-      CHECK( seiIt->payloadType != VVDEC_SCALABLE_NESTING, "expected nesting SEI" );
+      CHECK_RECOVERABLE( seiIt->payloadType != VVDEC_SCALABLE_NESTING, "expected nesting SEI" );
 
       const vvdecSEIScalableNesting* nestingSei = (vvdecSEIScalableNesting*)seiIt->payload;
       if( !nestingSei->snSubpicFlag )
@@ -408,7 +499,7 @@ void DecLib::checkPictureHashSEI( Picture* pcPic )
       for( int i = 0; i < nestingSei->snNumSEIs; ++i )
       {
         auto& nestedSei = nestingSei->nestedSEIs[i];
-        CHECK( nestedSei == nullptr, "missing nested sei" );
+        CHECK_RECOVERABLE( nestedSei == nullptr, "missing nested sei" );
         if( nestedSei && nestedSei->payloadType != VVDEC_DECODED_PICTURE_HASH )
           continue;
 
@@ -420,7 +511,7 @@ void DecLib::checkPictureHashSEI( Picture* pcPic )
         }
         else
         {
-          CHECK( pcPic->subpicsCheckedDPH.size() != pcPic->subPictures.size(), "Picture::subpicsCheckedDPH not properly initialized" );
+          CHECK_RECOVERABLE( pcPic->subpicsCheckedDPH.size() != pcPic->subPictures.size(), "Picture::subpicsCheckedDPH not properly initialized" );
         }
 
         for( int j = 0; j < nestingSei->snNumSubpics; ++j )
@@ -449,7 +540,7 @@ void DecLib::checkPictureHashSEI( Picture* pcPic )
     {
       // this warning is only enabled, when running with parse delay enabled, because otherwise we don't know here if the last DPH Suffix-SEI has already been
       // parsed
-      int checkedSubpicCount = std::count( pcPic->subpicsCheckedDPH.cbegin(), pcPic->subpicsCheckedDPH.cend(), true );
+      auto checkedSubpicCount = std::count( pcPic->subpicsCheckedDPH.cbegin(), pcPic->subpicsCheckedDPH.cend(), true );
       if( checkedSubpicCount != pcPic->subPictures.size() )
       {
         msg( WARNING, "Warning: missing decoded picture hash SEI message for SubPics (%u/%u).\n", checkedSubpicCount, pcPic->subPictures.size() );
@@ -460,18 +551,6 @@ void DecLib::checkPictureHashSEI( Picture* pcPic )
 
 Picture* DecLib::getNextOutputPic( bool bFlush )
 {
-  if( bFlush )
-  {
-    // wait for last pictures in bitstream
-    for( auto & dec: m_decLibRecon )
-    {
-      Picture* donePic = dec.waitForPrevDecompressedPic();
-      if( donePic )
-      {
-        finishPicture( donePic );
-      }
-    }
-  }
   if( m_picListManager.getFrontPic() == nullptr )
   {
     return nullptr;
@@ -493,41 +572,27 @@ Picture* DecLib::getNextOutputPic( bool bFlush )
     maxDecPicBufferingHighestTid = activeSPS->getMaxDecPicBuffering( m_iMaxTemporalLayer );
   }
 
-  Picture* outPic = m_picListManager.getNextOutputPic( numReorderPicsHighestTid, maxDecPicBufferingHighestTid, bFlush );
-  CHECK( outPic && outPic->done.isBlocked(), "next output-pic is not done yet." );
-  return outPic;
+  return m_picListManager.getNextOutputPic( numReorderPicsHighestTid, maxDecPicBufferingHighestTid, bFlush );
 }
 
-void DecLib::decompressPicture( Picture* pcPic )
+void DecLib::reconPicture( Picture* pcPic )
 {
   CHECK( std::any_of( m_decLibRecon.begin(), m_decLibRecon.end(), [=]( auto& rec ) { return rec.getCurrPic() == pcPic; } ),
          "(Reused) Picture structure is still in progress in decLibRecon." );
 
-  DecLibRecon* decLibInstance = &m_decLibRecon.front();
+  DecLibRecon* reconInstance = &m_decLibRecon.front();
   move_to_end( m_decLibRecon.begin(), m_decLibRecon );
 
-  while( pcPic->wasLost )
+  Picture* donePic = reconInstance->waitForPrevDecompressedPic();
+  try
   {
-    Picture* donePic = decLibInstance->waitForPrevDecompressedPic();
-    if( donePic )
-    {
-      finishPicture( donePic );
-    }
-
-    m_decLibParser.recreateLostPicture( pcPic );
-    finishPicture( pcPic );
-
-    pcPic = m_decLibParser.getNextDecodablePicture();
-    if( !pcPic )
-    {
-      msg(WARNING, "a lost picture was filled in, but no following picture is available for decoding.");
-      return;
-    }
+    reconInstance->decompressPicture( pcPic );
+  }
+  catch( ... )
+  {
+    pcPic->reconDone.setException( std::current_exception() );
+    pcPic->error = true;
   }
-
-  Picture* donePic = decLibInstance->waitForPrevDecompressedPic();
-
-  decLibInstance->decompressPicture( pcPic );
 
   if( donePic )
   {
@@ -535,6 +600,24 @@ void DecLib::decompressPicture( Picture* pcPic )
   }
 }
 
+void DecLib::blockAndFinishPictures( Picture* pcPic )
+{
+  // find Recon instance where current picture (if not null) is active and ensure the picture gets finished
+  // otherwise all pictures get finished
+  for( auto& recon: m_decLibRecon )
+  {
+    if( pcPic && recon.getCurrPic() != pcPic )
+    {
+      continue;
+    }
+
+    if( Picture* donePic = recon.waitForPrevDecompressedPic() )
+    {
+      finishPicture( donePic );
+    }
+  }
+}
+
 void DecLib::checkNalUnitConstraints( uint32_t naluType )
 {
   const ConstraintInfo *cInfo = NULL;
@@ -552,26 +635,26 @@ void DecLib::xCheckNalUnitConstraintFlags( const ConstraintInfo *cInfo, uint32_t
 {
   if( cInfo != NULL )
   {
-    CHECK( cInfo->getNoTrailConstraintFlag() && naluType == NAL_UNIT_CODED_SLICE_TRAIL,
-           "Non-conforming bitstream. no_trail_constraint_flag is equal to 1 but bitstream contains NAL unit of type TRAIL_NUT." );
-    CHECK( cInfo->getNoStsaConstraintFlag()  && naluType == NAL_UNIT_CODED_SLICE_STSA,
-           "Non-conforming bitstream. no_stsa_constraint_flag is equal to 1 but bitstream contains NAL unit of type STSA_NUT." );
-    CHECK( cInfo->getNoRaslConstraintFlag()  && naluType == NAL_UNIT_CODED_SLICE_RASL,
-           "Non-conforming bitstream. no_rasl_constraint_flag is equal to 1 but bitstream contains NAL unit of type RASL_NUT." );
-    CHECK( cInfo->getNoRadlConstraintFlag()  && naluType == NAL_UNIT_CODED_SLICE_RADL,
-           "Non-conforming bitstream. no_radl_constraint_flag is equal to 1 but bitstream contains NAL unit of type RADL_NUT." );
-    CHECK( cInfo->getNoIdrConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_IDR_W_RADL,
-           "Non-conforming bitstream. no_idr_constraint_flag is equal to 1 but bitstream contains NAL unit of type IDR_W_RADL." );
-    CHECK( cInfo->getNoIdrConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_IDR_N_LP,
-           "Non-conforming bitstream. no_idr_constraint_flag is equal to 1 but bitstream contains NAL unit of type IDR_N_LP." );
-    CHECK( cInfo->getNoCraConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_CRA,
-           "Non-conforming bitstream. no_cra_constraint_flag is equal to 1 but bitstream contains NAL unit of type CRA_NUT." );
-    CHECK( cInfo->getNoGdrConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_GDR,
-           "Non-conforming bitstream. no_gdr_constraint_flag is equal to 1 but bitstream contains NAL unit of type GDR_NUT." );
-    CHECK( cInfo->getNoApsConstraintFlag()   && naluType == NAL_UNIT_PREFIX_APS,
-           "Non-conforming bitstream. no_aps_constraint_flag is equal to 1 but bitstream contains NAL unit of type APS_PREFIX_NUT." );
-    CHECK( cInfo->getNoApsConstraintFlag()   && naluType == NAL_UNIT_SUFFIX_APS,
-           "Non-conforming bitstream. no_aps_constraint_flag is equal to 1 but bitstream contains NAL unit of type APS_SUFFIX_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoTrailConstraintFlag() && naluType == NAL_UNIT_CODED_SLICE_TRAIL,
+                       "Non-conforming bitstream. no_trail_constraint_flag is equal to 1 but bitstream contains NAL unit of type TRAIL_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoStsaConstraintFlag()  && naluType == NAL_UNIT_CODED_SLICE_STSA,
+                       "Non-conforming bitstream. no_stsa_constraint_flag is equal to 1 but bitstream contains NAL unit of type STSA_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoRaslConstraintFlag()  && naluType == NAL_UNIT_CODED_SLICE_RASL,
+                       "Non-conforming bitstream. no_rasl_constraint_flag is equal to 1 but bitstream contains NAL unit of type RASL_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoRadlConstraintFlag()  && naluType == NAL_UNIT_CODED_SLICE_RADL,
+                       "Non-conforming bitstream. no_radl_constraint_flag is equal to 1 but bitstream contains NAL unit of type RADL_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoIdrConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_IDR_W_RADL,
+                       "Non-conforming bitstream. no_idr_constraint_flag is equal to 1 but bitstream contains NAL unit of type IDR_W_RADL." );
+    CHECK_RECOVERABLE( cInfo->getNoIdrConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_IDR_N_LP,
+                       "Non-conforming bitstream. no_idr_constraint_flag is equal to 1 but bitstream contains NAL unit of type IDR_N_LP." );
+    CHECK_RECOVERABLE( cInfo->getNoCraConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_CRA,
+                       "Non-conforming bitstream. no_cra_constraint_flag is equal to 1 but bitstream contains NAL unit of type CRA_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoGdrConstraintFlag()   && naluType == NAL_UNIT_CODED_SLICE_GDR,
+                       "Non-conforming bitstream. no_gdr_constraint_flag is equal to 1 but bitstream contains NAL unit of type GDR_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoApsConstraintFlag()   && naluType == NAL_UNIT_PREFIX_APS,
+                       "Non-conforming bitstream. no_aps_constraint_flag is equal to 1 but bitstream contains NAL unit of type APS_PREFIX_NUT." );
+    CHECK_RECOVERABLE( cInfo->getNoApsConstraintFlag()   && naluType == NAL_UNIT_SUFFIX_APS,
+                       "Non-conforming bitstream. no_aps_constraint_flag is equal to 1 but bitstream contains NAL unit of type APS_SUFFIX_NUT." );
   }
 }
 
@@ -658,7 +741,7 @@ void DecLib::checkSeiInPictureUnit()
         }
       }
     }
-    CHECK(count > 4, "There shall be less than or equal to 4 identical sei_payload( ) syntax structures within a picture unit.");
+    CHECK_RECOVERABLE(count > 4, "There shall be less than or equal to 4 identical sei_payload( ) syntax structures within a picture unit.");
   }
 
   // free SEI message list memory
@@ -687,11 +770,11 @@ void DecLib::checkAPSInPictureUnit()
     if (NALUnit::isVclNalUnitType(nalu))
     {
       firstVCLFound = true;
-      CHECK( suffixAPSFound, "When any suffix APS NAL units are present in a PU, they shall follow the last VCL unit of the PU" );
+      CHECK_RECOVERABLE( suffixAPSFound, "When any suffix APS NAL units are present in a PU, they shall follow the last VCL unit of the PU" );
     }
     else if (nalu == NAL_UNIT_PREFIX_APS)
     {
-      CHECK( firstVCLFound, "When any prefix APS NAL units are present in a PU, they shall precede the first VCL unit of the PU");
+      CHECK_RECOVERABLE( firstVCLFound, "When any prefix APS NAL units are present in a PU, they shall precede the first VCL unit of the PU");
     }
     else if (nalu == NAL_UNIT_SUFFIX_APS)
     {
